/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: AbstractBasicComponent.java 601 2006-06-13 12:53:50Z wjoseph $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.component.common.basic;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.Bootstrap;
import javax.jbi.component.Component;
import javax.jbi.component.ComponentContext;
import javax.jbi.component.ComponentLifeCycle;
import javax.jbi.component.InstallationContext;
import javax.jbi.component.ServiceUnitManager;
import javax.jbi.management.DeploymentException;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.ObjectName;
import javax.xml.transform.Source;

import org.objectweb.petals.component.common.listener.IMessageExchangeProcessor;
import org.objectweb.petals.component.common.listener.MessageExchangeListener;
import org.objectweb.petals.component.common.util.ComponentLogger;
import org.objectweb.petals.component.common.util.SourceHelper;

import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;

/**
 * This class provides a basic component, allowing developper to create simply a
 * new JBI component
 * 
 * @author wjoseph - eBMWebSourcing
 * 
 */
public abstract class AbstractBasicComponent implements Component,
        ComponentLifeCycle, Bootstrap, IMessageExchangeProcessor {

    /**
     * The component's delivery channel
     */
    private DeliveryChannel channel;

    /**
     * the componentContext
     */
    private ComponentContext context;

    /**
     * The component's delivery logger
     */
    protected Logger log;

    /**
     * The component's service unit manager
     */
    protected AbstractServiceUnitManager serviceUnitManager;

    /**
     * The message exhange listener
     */
    private MessageExchangeListener listener;

    /**
     * Default contructor
     */
    public AbstractBasicComponent() {
    }

    /**
     * Create a new basic component, setting its logger and channel
     * 
     * @param channel
     *            the component's DeliveryChannel
     * @param log
     *            the component's logger
     */
    public AbstractBasicComponent(DeliveryChannel channel, Logger log) {
        this.channel = channel;
        this.log = log;
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Bootstrap#cleanUp()
     */
    public void cleanUp() throws JBIException {

    }

    /**
     * Returns the component DeliveryChannel
     * 
     * @return the DeliveryChannel
     * @throws JBIException
     *             if context retrieval fails
     * @throws MessagingException
     *             if {@link DeliveryChannel} retrieval fails
     */
    protected DeliveryChannel getDeliveryChannel() throws MessagingException,
        JBIException {
        if (channel == null) {
            channel = getComponentContext().getDeliveryChannel();
        }
        return channel;
    }

    /**
     * Returns the component context
     * 
     * @return the {@link ComponentContext}
     * @throws JBIException
     *             if the {@link Component} isn't correctly initilized
     */
    protected ComponentContext getComponentContext() throws JBIException {
        if (context == null) {
            throw new JBIException("Component not initilized.");
        }
        return context;
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ComponentLifeCycle#getExtensionMBeanName()
     */
    public ObjectName getExtensionMBeanName() {
        return null;
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#getLifeCycle()
     */
    public ComponentLifeCycle getLifeCycle() {
        return this;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#getServiceDescription(javax.jbi.servicedesc.ServiceEndpoint)
     */
    public Document getServiceDescription(ServiceEndpoint endpoint) {
        return serviceUnitManager.getServiceDescription(endpoint);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#getServiceUnitManager()
     */
    public ServiceUnitManager getServiceUnitManager() {
        return serviceUnitManager;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ComponentLifeCycle#init(javax.jbi.component.ComponentContext)
     */
    public void init(ComponentContext ctx) throws JBIException {
        this.context = ctx;
        Logger log = context.getLogger("", null);
        this.log = new ComponentLogger(log, log.getName(), log
                .getResourceBundleName(), context.getComponentName());
        channel = ctx.getDeliveryChannel();
        serviceUnitManager = createServiceUnitManager();
        log.log(Level.INFO, "init");
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Bootstrap#init(javax.jbi.component.InstallationContext)
     */
    public void init(InstallationContext installCtx) throws JBIException {

    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#isExchangeWithConsumerOkay(javax.jbi.servicedesc.ServiceEndpoint,
     *      javax.jbi.messaging.MessageExchange)
     */
    public boolean isExchangeWithConsumerOkay(ServiceEndpoint endpoint,
            MessageExchange exchange) {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#isExchangeWithProviderOkay(javax.jbi.servicedesc.ServiceEndpoint,
     *      javax.jbi.messaging.MessageExchange)
     */
    public boolean isExchangeWithProviderOkay(ServiceEndpoint endpoint,
            MessageExchange exchange) {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Bootstrap#onInstall()
     */
    public void onInstall() throws JBIException {

    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Bootstrap#onUninstall()
     */
    public void onUninstall() throws JBIException {

    }

    /**
     * Ignore DONE,ERROR status. Ignore "consumer" role.
     */
    abstract public boolean process(MessageExchange exchange) throws Exception;

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#resolveEndpointReference(org.w3c.dom.DocumentFragment)
     */
    public ServiceEndpoint resolveEndpointReference(DocumentFragment epr) {
        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ComponentLifeCycle#shutDown()
     */
    public void shutDown() throws JBIException {
        log.log(Level.INFO, "shutdown");
        listener = null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ComponentLifeCycle#start()
     */
    public void start() throws JBIException {
        log.log(Level.INFO, "start");
        listener = createMessageExchangeListener(this);
        listener.listen();
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ComponentLifeCycle#stop()
     */
    public void stop() throws JBIException {
        log.log(Level.INFO, "stop");
        this.listener.terminate();
    }

    /**
     * If the received exchange is a fault response from the consumer,
     * acknowledge it : status DONE
     * 
     * @param ex
     *            The processed {@link MessageExchange}
     * @return true if the ack. of the Fault is done; false if no fault is
     *         present in the exchange
     * @throws Exception
     */
    protected boolean ackFaultReception(MessageExchange ex) throws Exception {
        boolean result = false;

        if (ex.getFault() != null) {
            ex.setStatus(ExchangeStatus.DONE);

            result = true;
        }
        return result;
    }

    /**
     * Creates a Fault from an exception
     * 
     * @param e
     *            The {@link Exception} used to create a new {@link Fault}
     * @param exchange
     *            The {@link MessageExchange} that allows to create a new
     *            {@link Fault}
     * @return the created Fault
     * @throws MessagingException
     *             if an error occured during Fault creation or filling
     */
    protected Fault createFault(Exception e, MessageExchange exchange)
        throws MessagingException {
        Fault f = exchange.createFault();
        String faultString = SourceHelper.createSoapFault(e, null);
        Source content = SourceHelper.createSource(faultString);
        f.setContent(content);
        return f;
    }

    /**
     * Creates a new MessageExchangeListener
     * 
     * @param processor
     *            the messageExchange processor
     * @return the created {@link MessageExchangeListener}
     */
    protected MessageExchangeListener createMessageExchangeListener(
            AbstractBasicComponent processor) {
        return new MessageExchangeListener(channel, processor,
                MessageExchangeListener.IgnoredStatus.DONE_AND_ERROR_IGNORED,
                0, context.getComponentName());
    }

    /**
     * Creates a new NormalizedMessage from the message Exchange
     * 
     * @param exchange
     *            The MessageExchange to create the NormalizedMessage
     * @return a new NormalizedMessage
     * @throws MessagingException
     */
    protected NormalizedMessage createNormalizedMessage(MessageExchange exchange)
        throws MessagingException {
        return exchange.createMessage();
    }

    /**
     * Create a new service unit manager
     * 
     * @return new instance of service unit manager
     * @throws DeploymentException
     */
    protected AbstractServiceUnitManager createServiceUnitManager()
        throws DeploymentException {
        AbstractServiceUnitManager sum = new SimpleServiceUnitManager(context,
                log);
        return sum;
    }

    /**
     * Log errors in the system
     * 
     * @param e
     *            the error to log
     */
    protected void logError(Throwable e) {
        if (log != null)
            log.log(Level.SEVERE, e.getMessage(), e);
        else
            e.printStackTrace();
    }
}
