/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: AbstractComponent.java 601 2006-06-13 12:53:50Z wjoseph $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.component.common.basic;

import java.net.URI;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;

import org.objectweb.petals.component.common.MEPConstants;

/**
 * This class provides a basic component, allowing developper to create simply a
 * new JBI component
 * 
 * @author wjoseph - eBMWebSourcing
 * 
 */
public abstract class AbstractComponent extends AbstractBasicComponent {

    /**
     * Default contructor
     */
    public AbstractComponent() {
        super();
    }

    /**
     * Create a new AbstractComponent
     * 
     * @param channel
     * @param log
     */
    public AbstractComponent(DeliveryChannel channel, Logger log) {
        super(channel, log);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.component.common.basic.AbstractBasicComponent#process(javax.jbi.messaging.MessageExchange)
     */
    public boolean process(MessageExchange exchange) throws Exception {
        if (ExchangeStatus.ACTIVE.equals(exchange.getStatus())) {
            if (MessageExchange.Role.PROVIDER.equals(exchange.getRole())) {
                if (MEPConstants.IN_ONLY_PATTERN.equals(exchange.getPattern())) {
                    processInOnly(exchange);
                } else if (MEPConstants.ROBUST_IN_ONLY_PATTERN.equals(exchange
                        .getPattern())) {
                    processRobustInOnly(exchange);
                } else if (MEPConstants.IN_OUT_PATTERN.equals(exchange
                        .getPattern())
                        || MEPConstants.IN_OPTIONAL_OUT_PATTERN.equals(exchange
                                .getPattern())) {
                    processInOut(exchange);
                } else {
                    Exception e = new Exception(
                            "MessageExchangePattern not recognized :"
                                    + exchange.getPattern());
                    exchange.setError(e);
                    getDeliveryChannel().send(exchange);
                    throw e;
                }
                getDeliveryChannel().send(exchange);
            }
        }
        return true;
    }

    /**
     * Process "in-only" MessageExchange
     * 
     * @param exchange
     *            the MessageExchange
     * @throws Exception
     *             When errors occur during processing
     */
    protected void processInOnly(MessageExchange exchange) throws Exception {
        try {
            handleMessage(exchange.getEndpoint().getServiceName(), exchange
                    .getOperation(), exchange.getMessage("in"), exchange
                    .getMessage("out"), exchange.getPattern());
        } catch (Exception e) {
            // nothing, InOnly can not indicate a fault
            logError(new Exception(
                    "Error occured while processing, but InOnly message can not contains Fault.",
                    e));
        }
        exchange.setStatus(ExchangeStatus.DONE);
    }

    /**
     * Process "in-out" MessageExchange
     * 
     * @param exchange
     *            the MessageExchange
     * @throws Exception
     *             When errors occur during processing
     */
    protected void processInOut(MessageExchange exchange) throws Exception {
        try {
            if (!ackFaultReception(exchange)) {
                NormalizedMessage out = createNormalizedMessage(exchange);

                boolean outTreated = handleMessage(exchange.getEndpoint()
                        .getServiceName(), exchange.getOperation(), exchange
                        .getMessage("in"), out, exchange.getPattern());

                if (outTreated) {
                    exchange.setMessage(out, "out");
                } else {
                    exchange.setStatus(ExchangeStatus.DONE);
                }
            }
        } catch (Exception e) {
            try {
                exchange.setFault(createFault(e, exchange));
            } catch (Exception e1) {
                logError(new Exception(
                        "Error occured while processing, send a Fault.", e1));
            }
        }
    }

    /**
     * Process "robust-in-only" MessageExchange
     * 
     * @param exchange
     *            the MessageExchange
     * @throws Exception
     *             When errors occur during processing
     */
    protected void processRobustInOnly(MessageExchange exchange)
        throws Exception {
        try {
            handleMessage(exchange.getEndpoint().getServiceName(), exchange
                    .getOperation(), exchange.getMessage("in"), exchange
                    .getMessage("out"), exchange.getPattern());
            exchange.setStatus(ExchangeStatus.DONE);
        } catch (Exception e) {
            try {
                exchange.setFault(createFault(e, exchange));
            } catch (Exception e1) {
                logError(new Exception(
                        "Error occured while processing, send a Fault.", e1));
            }
        }
    }

    /**
     * Called to process a message
     * 
     * @param service
     *            The requested service QName
     * @param operation
     *            The requested operation QName
     * @param in
     *            The incoming {@link NormalizedMessage} extracted from the
     *            {@link MessageExchange}
     * @param out
     *            the outgoing {@link NormalizedMessage} that have to be set if
     *            MessageExchangePattern (MEP) is "InOut" or optionaly set if
     *            MEP is "InOptionalOut". This parameter is null if MEP is
     *            "InOnly" or "RobustInOnly". This parameter is an empty
     *            NormalizedMessage if MEP is "InOut" or "InOptionalOut"
     * @param messageExhangePattern
     *            The MessageExchange Pattern 
     * @see MEPConstants
     * @return true if the out response has been set, false otherwise
     * @throws Exception
     *             a Fault will be generated with the exception, the "out" will
     *             not be treated
     */
    protected abstract boolean handleMessage(QName service, QName operation,
            NormalizedMessage in, NormalizedMessage out,
            URI messageExhangePattern) throws Exception;

}
