/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: AbstractBindingComponent.java 601 2006-06-13 12:53:50Z wjoseph $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.component.common.basic.binding;

import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.management.DeploymentException;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;

import org.objectweb.petals.component.common.MEPConstants;
import org.objectweb.petals.component.common.basic.AbstractBasicComponent;
import org.objectweb.petals.component.common.basic.AbstractServiceUnitManager;

/**
 * This class provides a basic binding component, allowing developper to create
 * simply a new JBI binding component
 * 
 * @author wjoseph - eBMWebSourcing
 * 
 */
public abstract class AbstractBindingComponent extends AbstractBasicComponent {

    protected JBIListener jbiListener;

    protected ExternalListenerManager externalListenerManager;

    /**
     * Default contructor
     */
    public AbstractBindingComponent() {
        super();
    }

    /**
     * Create a new binding component, setting its logger and channel
     * 
     * @param channel
     *            the component's DeliveryChannel
     * @param log
     *            the component's logger
     */
    public AbstractBindingComponent(DeliveryChannel channel, Logger log) {
        super(channel, log);
    }

    /**
     * If this binding component must handle jbi messages, you must implement
     * this method to return your custom {@link JBIListener}.
     * 
     * @return
     */
    public abstract JBIListener getJBIListener();

    /**
     * If this binding component must handle messages from external service to
     * internal jbi components, you must implement this method to return your
     * custom {@link ExternalListenerManager}.
     * 
     * @return
     */
    public abstract ExternalListenerManager getExternalListenerManager();

    @Override
    protected AbstractServiceUnitManager createServiceUnitManager()
        throws DeploymentException {
        AbstractServiceUnitManager serviceUnitManager = null;
        try {
            serviceUnitManager = new SimpleBindingComponentServiceUnitManager(
                getComponentContext(), log);
        } catch (JBIException e) {
            throw new DeploymentException(
                "Error when creating the Service Unit Manager", e);
        }

        return serviceUnitManager;
    }

    @Override
    public void start() throws JBIException {
        jbiListener = getJBIListener();
        externalListenerManager = getExternalListenerManager();
        ((SimpleBindingComponentServiceUnitManager) serviceUnitManager)
            .setExternalListenerManager(externalListenerManager);
        super.start();

    }

    /**
     * Process incoming messages for the component. Ignore DONE,ERROR status.
     * Ignore "consumer" role.
     */
    public boolean process(MessageExchange exchange) throws Exception {

        if (ExchangeStatus.ACTIVE.equals(exchange.getStatus())
            && MessageExchange.Role.PROVIDER.equals(exchange.getRole())) {

            String address = ((SimpleBindingComponentServiceUnitManager) serviceUnitManager)
                .getAddressForEndpoint(exchange.getEndpoint());

            if (MEPConstants.IN_ONLY_PATTERN.equals(exchange.getPattern())) {
                processInOnly(exchange, address);
                
            } else if (MEPConstants.ROBUST_IN_ONLY_PATTERN.equals(exchange
                .getPattern())) {
                processRobustInOnly(exchange, address);
                
            } else if (MEPConstants.IN_OUT_PATTERN
                .equals(exchange.getPattern())
                || MEPConstants.IN_OPTIONAL_OUT_PATTERN.equals(exchange
                    .getPattern())) {
                processInOut(exchange, address);
                
            } else {
                Exception e = new Exception(
                    "MessageExchangePattern not recognized :"
                        + exchange.getPattern());
                exchange.setError(e);
                getDeliveryChannel().send(exchange);
                throw e;
            }
            getDeliveryChannel().send(exchange);
        }
        return true;
    }

    /**
     * Call listener's 'onJBIMessage()' for an InOnly exchange.
     * The Exchange status is set to done. 
     * @param exchange
     * @param address
     * @throws MessagingException
     */
    protected void processInOnly(MessageExchange exchange, String address)
        throws MessagingException {
        NormalizedMessage in = exchange.getMessage("in");
        try{
        jbiListener.onJBIMessage(address, in, null, exchange.getOperation(),
            exchange.getPattern());
        }catch(MessagingException e){
            // even if there was a problem, the exchange is set to done
        }
        exchange.setStatus(ExchangeStatus.DONE);
    }

    /**
     * Call listener's 'onJBIMessage()' for an RobustInOnly exchange.
     * The Exchange status is set to done or a fault is returned on the exchange. 
     * @param exchange
     * @param address
     * @throws MessagingException
     */
    protected void processRobustInOnly(MessageExchange exchange,String address) throws MessagingException{
        NormalizedMessage in = exchange.getMessage("in");
        try {
            jbiListener.onJBIMessage(address, in, null, exchange.getOperation(), exchange
                .getPattern());
            // exchange is done
            exchange.setStatus(ExchangeStatus.DONE);
        } catch (MessagingException e) {
            // a fault has to be returned
            exchange.setFault(createFault(new Exception(
                    "Error sending the message", e), exchange));
        }
    }

    /**
     * Call listener's 'onJBIMessage()' for an InOut or InOptionalOut exchange.
     * An out or a fault is returned on the exchange
     * @param exchange
     * @param address
     * @throws MessagingException
     */
    protected void processInOut(MessageExchange exchange, String address)
        throws MessagingException {
        
        NormalizedMessage out = exchange.createMessage();
        NormalizedMessage in = exchange.getMessage("in");
        
        try {
            boolean responseSet = jbiListener.onJBIMessage(address, in, out, exchange
                .getOperation(), exchange.getPattern());
            
            if (responseSet || MEPConstants.IN_OUT_PATTERN.equals(exchange
                    .getPattern())) {
                // response has to be returned
                exchange.setMessage(out, "out");
            } else {
                // exchange is done
                exchange.setStatus(ExchangeStatus.DONE);
            }
        } catch (MessagingException e) {
            // a fault has to be returned
            exchange.setFault(createFault(new Exception(
                "Error sending the message", e), exchange));
        }

    }

    /**
     * Send a message to the endpoint associated to the distant address. This
     * method is used by a listener, that listen to messages from external
     * services (linked to the given address), in order to send these messages
     * to internal JBI Components.
     * 
     * @param address
     *            the distant address of the external consumer service
     * @param messageExchange
     *            the message exchange to be sent to the internal JBI Component
     * @throws MessagingException
     *             if the address has no endpoint associated to it, or if an
     *             error occurs while sending the message on the container
     */
    public void sendMessage(String address, MessageExchange messageExchange)
        throws MessagingException {
        ServiceEndpoint ep = ((SimpleBindingComponentServiceUnitManager) serviceUnitManager)
            .getEndpointForAddress(address);
        if (ep != null) {
            messageExchange.setEndpoint(ep);
            messageExchange.setService(ep.getServiceName());
            try {
                getDeliveryChannel().send(messageExchange);
            } catch (JBIException e) {
                throw new MessagingException(
                    "Failed to send the mapped message to the delivery channel.");
            }
        } else {
            throw new MessagingException(
                "No endpoint associated with this address");
        }
    }
}
