/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.component.common.listener;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.objectweb.petals.component.common.MEPConstants;
import org.objectweb.petals.component.common.util.SourceHelper;
import org.objectweb.petals.component.common.util.WSDLHelper;

/**
 * This abstract class helps the processing of MessageExchanges for a Binding
 * Component that have to transfer this call to an external service. Only two
 * methods have to be implemented: - externalSend(in), for sending an In message
 * to an external service, - externalSend(in,out), same as above, and wait for
 * the response.
 * 
 * DeliveryChannel.send() is treated, Reception or send of DONE-status is
 * treated, Fault message is generated if one of the method throws an exception
 * (and if the MEP allow Fault message, otherwise, nothing is done).
 * 
 * @version $Revision: $ $Date: $
 * @since Petals 1.0
 * @author alouis
 */
public abstract class AbstractExternalMEProcessor implements
        IMessageExchangeProcessor {

    protected ComponentContext ctx;

    protected Logger l;

    /**
     * Constructor. The componentcontext is used to retrieve the
     * wsdl-information of the endpoint to contact. TODO use the SUMan instead
     * of the CContext
     * 
     * @param ctx,
     *            must be non null
     * @param l
     *            logger, may be null
     */
    public AbstractExternalMEProcessor(ComponentContext ctx, Logger l) {
        this.ctx = ctx;
        this.l = l;
    }

    /**
     * Ignore DONE,ERROR status. Ignore "consumer" role.
     */
    public boolean process(MessageExchange exchange) throws Exception {
        String address = getExternalAddress(exchange.getEndpoint(), exchange
                .getService());

        if (ExchangeStatus.ACTIVE.equals(exchange.getStatus())) {
            if (MessageExchange.Role.PROVIDER.equals(exchange.getRole())) {
                if (MEPConstants.IN_ONLY_PATTERN.equals(exchange.getPattern())) {
                    processInOnly(address, exchange);
                } else if (MEPConstants.ROBUST_IN_ONLY_PATTERN.equals(exchange
                        .getPattern())) {
                    processRobustInOnly(address, exchange);
                } else if (MEPConstants.IN_OUT_PATTERN.equals(exchange
                        .getPattern())) {
                    processInOut(address, exchange, false);
                } else if (MEPConstants.IN_OPTIONAL_OUT_PATTERN.equals(exchange
                        .getPattern())) {
                    processInOut(address, exchange, true);
                } else {
                    Exception e = new Exception(
                            "MessageExchangePattern not recognized :"
                                    + exchange.getPattern());
                    exchange.setError(e);
                    ctx.getDeliveryChannel().send(exchange);
                    throw e;
                }
                ctx.getDeliveryChannel().send(exchange);
            }
        }
        return true;
    }

    /**
     * If the received exchange is a fault response from the consumer,
     * acknowledge it : status DONE
     * 
     * @param ex
     * @return true if the ack. of the Fault is done; false if no fault is
     *         present in the exchange
     * @throws Exception
     */
    protected boolean ackFaultReception(MessageExchange ex) throws Exception {
        boolean result = false;

        if (ex.getFault() != null) {
            ex.setStatus(ExchangeStatus.DONE);

            result = true;
        }
        return result;
    }

    /**
     * Creates a Fault from an exception
     * 
     * @param e
     * @return
     */
    protected Fault createFault(Exception e, MessageExchange exchange)
        throws MessagingException {
        Fault f = exchange.createFault();
        String faultString = SourceHelper.createSoapFault(e, null);
        Source content = SourceHelper.createSource(faultString);
        f.setContent(content);
        return f;
    }

    /**
     * Creates a NormalizedMessage for Out
     * 
     * @return
     */
    protected NormalizedMessage createNormalizedMessage(MessageExchange exchange)
        throws MessagingException {
        return exchange.createMessage();
    }

    /**
     * Return the physical address of the external endpoint to contact. The
     * address is retrieved from the WSDL description attached with the
     * specified ServiceEndpoint.
     * 
     * @param endpoint
     * @param serviceName
     * @return External service address
     * 
     * @throws JBIException
     */
    protected String getExternalAddress(ServiceEndpoint endpoint,
            QName serviceName) throws JBIException {
        String result = null;
        try {
            result = WSDLHelper.getServiceAddressFromWSDLDocument(ctx
                    .getEndpointDescriptor(endpoint), serviceName);
        } catch (Exception e) {
            throw new JBIException("Could not get service address", e);
        }
        return result;
    }

    protected void logError(Throwable e) {
        if (l != null)
            l.log(Level.SEVERE, e.getMessage(), e);
        else
            e.printStackTrace();
    }

    protected void processInOnly(String address, MessageExchange exchange)
        throws Exception {
        try {
            sendInMessage(address, exchange.getEndpoint().getServiceName(),
                    exchange.getOperation(), exchange.getMessage("in"));
        } catch (Exception e) {
            // nothing, InOnly can not indicate a fault
            logError(new Exception(
                    "Error occured while processing, but InOnly message can not contains Fault.",
                    e));
        }
        exchange.setStatus(ExchangeStatus.DONE);
    }

    protected void processInOut(String address, MessageExchange exchange,
            boolean optional) throws Exception {
        try {
            if (!ackFaultReception(exchange)) {
                NormalizedMessage out = createNormalizedMessage(exchange);

                boolean outTreated = sendInOutMessage(address, exchange
                        .getEndpoint().getServiceName(), exchange
                        .getOperation(), exchange.getMessage("in"), out,
                        optional);

                if (outTreated) {
                    exchange.setMessage(out, "out");
                } else {
                    exchange.setStatus(ExchangeStatus.DONE);
                }
            }
        } catch (Exception e) {
            logError(new Exception(
                    "Error occured while processing, send a Fault.", e));
            exchange.setFault(createFault(e, exchange));
        }
    }

    protected void processRobustInOnly(String address, MessageExchange exchange)
        throws Exception {
        try {
            sendInMessage(address, exchange.getEndpoint().getServiceName(),
                    exchange.getOperation(), exchange.getMessage("in"));

            exchange.setStatus(ExchangeStatus.DONE);
        } catch (Exception e) {
            logError(new Exception(
                    "Error occured while processing, send a Fault.", e));
            exchange.setFault(createFault(e, exchange));
        }
    }

    /**
     * Called to send an In message (InOnly or RobustInOnly). The given address
     * is resolved by a ctx.getEndpointDescriptor(endpoint), which in fact
     * search the address specified in the WSDL description related to the
     * endpoint.
     * 
     * @param address
     *            the address mapped with the ServiceEndpoint
     * @param service
     * @param operation
     * @param in
     * @throws Exception
     *             if RobustInOnly, a Fault will be generated with the exception
     */
    protected abstract void sendInMessage(String address, QName service,
            QName operation, NormalizedMessage in) throws Exception;

    /**
     * Called to send an InOut message (InOut or InOptionalOut). The given
     * address is resolved by a ctx.getEndpointDescriptor(endpoint), which in
     * fact search the address specified in the WSDL description related to the
     * endpoint. when the response is received, the 'out' object has to be
     * updated with it.
     * 
     * @param address
     *            the address mapped with the ServiceEndpoint
     * @param service
     * @param operation
     * @param in
     * @param out
     *            the out message on which the response has to be set
     * @param optionalOut
     *            the response is optional or not
     * @return true if the out response has been set (correspond to the
     *         'optionalOut' notion)
     * @throws Exception
     *             a Fault will be generated with the exception, the "out" will
     *             not be treated
     */
    protected abstract boolean sendInOutMessage(String address, QName service,
            QName operation, NormalizedMessage in, NormalizedMessage out,
            boolean optionalOut) throws Exception;
}
