/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.component.common.listener;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.objectweb.petals.component.common.MEPConstants;
import org.objectweb.petals.component.common.util.SourceHelper;

/**
 * This abstract class helps the processing of MessageExchanges for a Service
 * Engine that have to transfer this call to an internal service. Only two
 * methods have to be implemented: - sendInMessage(in), for sending an In
 * message to an internal service, - sendInOutMessage(in,out), same as above,
 * and wait for the response.
 * 
 * DeliveryChannel.send() is treated, Reception or send of DONE-status is
 * treated, Fault message is generated if one of the method throws an exception
 * (and if the MEP allow Fault message, otherwise, nothing is done).
 * 
 * @version $Revision: $ $Date: $
 * @since Petals 1.0
 * @author alouis
 */
public abstract class AbstractInternalMEProcessor implements
    IMessageExchangeProcessor {

	/**
	 * The component's delivery channel
	 */
    protected DeliveryChannel channel;

    protected Logger l;

    /**
     * Constructor. The componentcontext is used to retrieve the
     * wsdl-information of the endpoint to contact. TODO use the SUMan instead
     * of the CContext
     * 
     * @param ctx,
     *            must be non null
     * @param l
     *            logger, may be null
     */
    public AbstractInternalMEProcessor(DeliveryChannel channel, Logger l) {
        this.channel = channel;
        this.l = l;
    }
    
    /**
     * Default constructor
     */
    public AbstractInternalMEProcessor() {
    }

    /**
     * Ignore DONE,ERROR status. Ignore "consumer" role.
     */
    public boolean process(MessageExchange exchange) throws Exception {
        if (ExchangeStatus.ACTIVE.equals(exchange.getStatus())) {
            if (MessageExchange.Role.PROVIDER.equals(exchange.getRole())) {
                if (MEPConstants.IN_ONLY_PATTERN.equals(exchange.getPattern())) {
                    processInOnly(exchange);
                } else if (MEPConstants.ROBUST_IN_ONLY_PATTERN.equals(exchange
                    .getPattern())) {
                    processRobustInOnly(exchange);
                } else if (MEPConstants.IN_OUT_PATTERN.equals(exchange
                    .getPattern())) {
                    processInOut(exchange, false);
                } else if (MEPConstants.IN_OPTIONAL_OUT_PATTERN.equals(exchange
                    .getPattern())) {
                    processInOut(exchange, true);
                } else {
                    Exception e = new Exception(
                        "MessageExchangePattern not recognized :"
                            + exchange.getPattern());
                    exchange.setError(e);
                    channel.send(exchange);
                    throw e;
                }
                channel.send(exchange);
            }
        }
        return true;
    }

    /**
     * If the received exchange is a fault response from the consumer,
     * acknowledge it : status DONE
     * 
     * @param ex
     * @return true if the ack. of the Fault is done; false if no fault is
     *         present in the exchange
     * @throws Exception
     */
    protected boolean ackFaultReception(MessageExchange ex) throws Exception {
        boolean result = false;

        if (ex.getFault() != null) {
            ex.setStatus(ExchangeStatus.DONE);

            result = true;
        }
        return result;
    }

    /**
     * Creates a Fault from an exception
     * 
     * @param e
     * @return
     */
    protected Fault createFault(Exception e, MessageExchange exchange)
        throws MessagingException {
        Fault f = exchange.createFault();
        String faultString = SourceHelper.createSoapFault(e, null);
        Source content = SourceHelper.createSource(faultString);
        f.setContent(content);
        return f;
    }

    /**
     * Creates a NormalizedMessage for Out
     * 
     * @return
     */
    protected NormalizedMessage createNormalizedMessage(MessageExchange exchange)
        throws MessagingException {
        return exchange.createMessage();
    }

    protected void logError(Throwable e) {
        if (l != null)
            l.log(Level.SEVERE, e.getMessage(), e);
        else
            e.printStackTrace();
    }

    protected void processInOnly(MessageExchange exchange) throws Exception {
        try {
            sendInMessage(exchange.getEndpoint().getServiceName(), exchange
                .getOperation(), exchange.getMessage("in"));
        } catch (Exception e) {
            // nothing, InOnly can not indicate a fault
            logError(new Exception(
                "Error occured while processing, but InOnly message can not contains Fault.",
                e));
        }
        exchange.setStatus(ExchangeStatus.DONE);
    }

    protected void processInOut(MessageExchange exchange, boolean optional)
        throws Exception {
        try {
            if (!ackFaultReception(exchange)) {
                NormalizedMessage out = createNormalizedMessage(exchange);

                boolean outTreated = sendInOutMessage(exchange.getEndpoint()
                    .getServiceName(), exchange.getOperation(), exchange
                    .getMessage("in"), out, optional);

                if (outTreated) {
                    exchange.setMessage(out, "out");
                } else {
                    exchange.setStatus(ExchangeStatus.DONE);
                }
            }
        } catch (Exception e) {
            try {
                exchange.setFault(createFault(e, exchange));
            } catch (Exception e1) {
                logError(new Exception(
                    "Error occured while processing, send a Fault.", e1));
            }
        }
    }

    protected void processRobustInOnly(MessageExchange exchange)
        throws Exception {
        try {
            sendInMessage(exchange.getEndpoint().getServiceName(), exchange
                .getOperation(), exchange.getMessage("in"));

            exchange.setStatus(ExchangeStatus.DONE);
        } catch (Exception e) {
            try {
                exchange.setFault(createFault(e, exchange));
            } catch (Exception e1) {
                logError(new Exception(
                    "Error occured while processing, send a Fault.", e1));
            }
        }
    }

    /**
     * Called to send an In message (InOnly or RobustInOnly).
     * 
     * @param service
     * @param operation
     * @param in
     * @throws Exception
     *             if RobustInOnly, a Fault will be generated with the exception
     */
    protected abstract void sendInMessage(QName service, QName operation,
        NormalizedMessage in) throws Exception;

    /**
     * Called to send an InOut message (InOut or InOptionalOut).
     * 
     * @param service
     * @param operation
     * @param in
     * @param out
     *            the out message on which the response has to be set
     * @param optionalOut
     *            the response is optional or not
     * @return true if the out response has been set (correspond to the
     *         'optionalOut' notion)
     * @throws Exception
     *             a Fault will be generated with the exception, the "out" will
     *             not be treated
     */
    protected abstract boolean sendInOutMessage(QName service, QName operation,
        NormalizedMessage in, NormalizedMessage out, boolean optionalOut)
        throws Exception;

	public DeliveryChannel getChannel() {
		return channel;
	}

	public void setChannel(DeliveryChannel channel) {
		this.channel = channel;
	}

	public Logger getLogger() {
		return l;
	}

	public void setLogger(Logger l) {
		this.l = l;
	}
}
