/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: CsvProcessor.java 1365 2006-12-08 22:17:01Z dutoo $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.engine.csv;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.charset.Charset;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.apache.commons.io.IOUtils;
import org.objectweb.petals.component.common.listener.AbstractInternalMEProcessor;
import org.objectweb.petals.component.common.util.SourceHelper;
import org.objectweb.petals.component.common.util.StringHelper;
import org.objectweb.petals.component.common.util.XMLHelper;
import org.objectweb.petals.engine.csv.exception.CsvException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import com.csvreader.CsvReader;

/**
 * This processor implements to operations : csvToXml and xmlToCsv.
 * 
 * <p/>
 * The csvToXml operation :<br/>
 * This processor uses Java CSV's CsvReader to transform either
 * the <content>-encapsulated content or the single attachment of
 * the message from any CSV format to XML.
 * 
 * It is configured through properties given by the CSV SU handler.
 * Here are their meaning :
 * <ul>
 *     <li>csvInputMode : may be "content", "attachment" or "guess".
 * Default is "guess", meaning that if there is any attachment this
 * will be the chosen mode, and if not, it will be in "content" mode.
 * Not that this works for both ways (xmlToCsv and csvToXml)</li>
 *     <li>customColumnDelimiterChar : default is the comma ","</li>
 *     <li>useCustomRecordDelimiter : whether to use a custom one
 * instead of the default combination of unix & windows line breaks.
 * SO beware, using it lessens the original capabilities of the
 * parser and it is therefore not advised.</li>
 *     <li>customRecordDelimiterChar : custom line delimiter character.
 * See the "useCustomRecordDelimiter" property to activate this feature.</li>
 *     <li>useDoubleEscapeMode : whether to use doubled backslash to
 * trigger escaping instead of a single one.</li>
 *     <li>useComment : whether to parse and make comments available.
 * Default is false.</li>
 *     <li>commentChar : the character used to starts comments.
 * Default is sharp "#".</li>
 * </ul>
 * 
 * <p/>
 * The xmlToCsv operation :<br/>
 * It views a two level-deep XML tree as tabular data and converts it
 * to CSV using the properties csvInputMode, customColumnDelimiterChar, 
 * useCustomRecordDelimiter and customRecordDelimiterChar with the same
 * meaning as above. The CSV is then returned in the response inside a
 * <content> top tag.
 *     
 * @version $Rev: 1081 $Date: {date}
 * @since   Petals 1.0
 * @author Marc DUTOO - Open Wide
 *
 */
public class CsvProcessor extends AbstractInternalMEProcessor {
    
    public static final String CSV_TO_XML_OPERATION = "csvToXml";
    public static final String XML_TO_CSV_OPERATION = "xmlToCsv";

    public static final String CSV_INPUT_MODE_CONTENT = "content";
    public static final String CSV_INPUT_MODE_ATTACHMENT = "attachment";
    public static final String CSV_INPUT_MODE_GUESS = "guess";

    public static final String CSV_INPUT_MODE_PROP = "csvInputMode";
    public static final String CUSTOM_COLUMN_DELIMITER_CHAR_PROP = "customColumnDelimiterChar";
    public static final String USE_CUSTOM_RECORD_DELIMITER_PROP = "useCustomRecordDelimiter";
    public static final String CUSTOM_RECORD_DELIMITER_CHAR_PROP = "customRecordDelimiterChar";
    public static final String USE_DOUBLE_ESCAPE_MODE_PROP = "useDoubleEscapeMode";
    public static final String USE_COMMENT_PROP = "useComment";
    public static final String COMMENT_CHAR_PROP = "commentChar";
    
    // The following constants describe the format of the XML output.
    // They could be configurable in properties.
    // The tag to be used to encapsulate CSV content in content input mode
    private static final String CONTENT_TAG = "content";
    private static final String DATA_TAG = "data";
    private static final String ROW_TAG = "row";
    private static final String CSVXML_NAMESPACE_URI = "http://petals.objectweb.org/components/csv/csvxml";
    private static final String CELL_TAG = "cell";

    
    /** service name to component properties map */
    private Map<String, Properties> mapEndpointCsv;
    
    
    /**
     * Creates a new CSV processor configured to use the given map of
     * service names to component properties.
     * @param c
     * @param l
     * @param mapEndpointCsv
     */
    public CsvProcessor(DeliveryChannel c, Logger l,
        Map<String, Properties> mapEndpointCsv) {
        super(c, l);
        this.mapEndpointCsv = mapEndpointCsv;
    }


    @Override
    protected void sendInMessage(QName service, QName operation,
        NormalizedMessage in) throws CsvException {
        // not supported : since it is a service engine, it outputs something
        throw new CsvException("Only InOut messages are supported");
    }

    @Override
    protected boolean sendInOutMessage(QName service, QName operation,
        NormalizedMessage in, NormalizedMessage out, boolean optionalOut)
    throws CsvException {
        // getting the service properties
        Properties csvProperties = getCsvProperties(service.getLocalPart());

        
        if (CSV_TO_XML_OPERATION.equals(operation.getLocalPart())) {
            try {
                // getting the content to be transformed
                String csvContent = getCsvInputContent(in, csvProperties);
                
                if (StringHelper.isEmpty(csvContent)) {
                    l.log(Level.WARNING, "Empty message content");
                    // this will be automaticly added to a Fault if the message
                    throw new CsvException("operation : message empty");
                }
                
                // actual CSV to XML transformation
                ByteArrayOutputStream xmlOut = new ByteArrayOutputStream();
                transformCsvToXml(new ByteArrayInputStream(csvContent.getBytes()),
                    xmlOut, csvProperties);
    
                // Transform Stream result to String, to be send in NormalizeMessage
                out.setContent(SourceHelper.createSource(new String(xmlOut.toByteArray())));
            } catch (CsvException e) {
                throw e;
            } catch (Exception e) {
                String msg = "Error while transforming csv to XML : " + e.getMessage();
                l.log(Level.SEVERE, msg, e);
                throw new CsvException(msg, e);
            }
        
            
        } else if (XML_TO_CSV_OPERATION.equals(operation.getLocalPart())) {
            Node documentNode = getXmlInputDocument(in, csvProperties);
            
            try {
                Node xmlTableNode = XMLHelper.getFirstChild(documentNode);
                // actual XML to CSV transformation
                StringBuffer csvBuf = new StringBuffer();
                transformXmlToCsv(xmlTableNode, csvBuf, csvProperties);
        
                // Wraps the resulting CSV in a <content> tag and returns it
                out.setContent(SourceHelper.createContentSource(csvBuf.toString()));
            } catch (CsvException e) {
                throw e;
            } catch (Exception e) {
                String msg = "Error while transforming XML to CSV : " + e.getMessage();
                l.log(Level.SEVERE, msg, e);
                throw new CsvException(msg, e);
            }
            
        } else {
            throw new CsvException("Unknown operation " + operation);
        }
        
        return true;
    }


    /**
     * Transforms the given XML tabular data to CSV.
     * There is no requirement on the format (tags...) besides the fact that
     * they have to be a two level tree.
     * @param xmlTableNode the XML tabular data to be written as CSV
     * @param csvBuf where the converted CSV data will be written
     * @param csvProperties the service properties
     * @throws CsvException if conversion error
     */
    private void transformXmlToCsv(Node xmlTableNode, StringBuffer csvBuf,
        Properties csvProperties) throws CsvException {

        // getting CSV formatting properties
        char csvColumnDelimiterChar = csvProperties
              .getProperty(CUSTOM_COLUMN_DELIMITER_CHAR_PROP, ",").charAt(0);
        char csvCustomRecordDelimiterChar = csvProperties
              .getProperty(CUSTOM_RECORD_DELIMITER_CHAR_PROP, "\n").charAt(0);
        boolean useCustomRecordDelimiter = Boolean.valueOf(csvProperties
            .getProperty(USE_CUSTOM_RECORD_DELIMITER_PROP, "false")).booleanValue();
        String recordDelimiter = (useCustomRecordDelimiter)
            ? String.valueOf(csvCustomRecordDelimiterChar) : "\r\n";
        
        NodeList rowNodes = xmlTableNode.getChildNodes();
        int rowNodeNb = rowNodes.getLength();
        for (int i = 0; i < rowNodeNb; i++) {
            Node rowNode = rowNodes.item(i);
            if (!(rowNode instanceof Element)) {
                continue;
            }
            
            // now outputting the current row :
            NodeList columnNodes = rowNode.getChildNodes();
            int columnNodeNb = columnNodes.getLength();
            for (int j = 0; j < columnNodeNb; j++) {
                Node columnNode = columnNodes.item(j);
                if (!(columnNode instanceof Element)) {
                    continue;
                }
                
                // outputting cell value
                String cellValue = XMLHelper.getTextContent(columnNode);
                csvBuf.append(cellValue);
                csvBuf.append(csvColumnDelimiterChar);
            }
            if (columnNodeNb != 0) {
                // removing the last column delimiter
                csvBuf.deleteCharAt(csvBuf.length() - 1);
            }
            
            // appending end of row
            csvBuf.append(recordDelimiter);
        }
        
        if (rowNodeNb != 0) {
            // removing the last row delimiter
            csvBuf.deleteCharAt(csvBuf.length() - recordDelimiter.length());
        }
    }


    /**
     * Returns the actual XML content from the given incoming message
     * according to the processor's settings
     * @param in
     * @param csvProperties
     * @return
     * @throws CsvException
     */
    protected Node getXmlInputDocument(NormalizedMessage in, Properties csvProperties)
            throws CsvException {
        try {
            if (isModeAttachment(in, csvProperties)) {
                // getting first attachment
                String firstAttachmentName = null;
                for (Iterator nameIt = in.getAttachmentNames().iterator(); nameIt.hasNext();) {
                    firstAttachmentName = (String) nameIt.next();
                    break;
                }
                if (firstAttachmentName == null) {
                    String msg = "Error while getting XML source content : "
                        + "inputMode is \"attachment\" but there is none given";
                    l.log(Level.SEVERE, msg);
                    throw new CsvException(msg);
                }
                DataHandler firstAttachment = in.getAttachment(firstAttachmentName);
                String contentString = IOUtils.toString(firstAttachment.getInputStream());
                return XMLHelper.createDocumentFromString(contentString);

            } else {
                Source xmlTableSource = in.getContent();
                return XMLHelper.createDOMNodeFromSource(xmlTableSource);
            }
        } catch (CsvException e) {
            throw e;
        } catch (Exception e) {
            String msg = "Error while getting XML source content : " + e.getMessage();
            l.log(Level.SEVERE, msg, e);
            throw new CsvException(msg, e);
        }
    }


    /**
     * Returns the actual CSV content from the given incoming message
     * according to the processor's settings
     * @param in
     * @param csvProperties
     * @return
     * @throws Exception 
     * @throws  
     * @throws Exception
     */
    protected String getCsvInputContent(NormalizedMessage in,
            Properties csvProperties) throws Exception {
        String contentString;
        if (isModeAttachment(in, csvProperties)) {
            // getting first attachment
            String firstAttachmentName = null;
            for (Iterator nameIt = in.getAttachmentNames().iterator(); nameIt.hasNext();) {
                firstAttachmentName = (String) nameIt.next();
                break;
            }
            if (firstAttachmentName == null) {
                return null; // no content
            }
            DataHandler firstAttachment = in.getAttachment(firstAttachmentName);
            contentString = IOUtils.toString(firstAttachment.getInputStream());
            // NB. XML encoding of the ouput content is done during transformation

        } else {
            // inlined content mode, inside a <content> tag
            Source content = in.getContent();
            contentString = createStringFromContent(content);
            // NB. since it was already XML content, the values are already XML encoded
        }
        return contentString;
    }
    
    /**
     * Unwrapps the given content source : transforms the source
     * to a String and removes the (required) encapsulating <content>
     * tag.
     * @param content
     * @return
     * @throws Exception
     */
    protected static String createStringFromContent(Source content) throws Exception {
        Node contentNode = XMLHelper.getFirstChild(XMLHelper.createDOMNodeFromSource(content));
        // Checking that node element tag is "content"
        if (!CONTENT_TAG.equals(contentNode.getLocalName())) {
            throw new CsvException("Expected <content> encapsulated content "
                + "but found " + XMLHelper.createStringFromDOMNode(contentNode));
        }
        // Usual case : the original content has been encapsulated as XML
        String contentString = contentNode.getTextContent();
        return contentString;
    }

    /**
     * Returns whether the CSV is to be looked for in the
     * message's first attachment instead of in its content
     * @param in
     * @param csvProperties
     * @return
     */
    protected boolean isModeAttachment(NormalizedMessage in,
            Properties csvProperties) {
        String csvInputMode = csvProperties.getProperty(CSV_INPUT_MODE_PROP, "guess");
        if (CSV_INPUT_MODE_ATTACHMENT.equals(csvInputMode)) {
            return true;
        } else if (CSV_INPUT_MODE_CONTENT.equals(csvInputMode)) {
            return false;
        } else { // guessing mode
            if (!in.getAttachmentNames().isEmpty()) {
                return true;
            }
        }
        return false;
    }


    /**
     * Transforms the given CSV content input in XML and
     * outputs it
     * @param csvIn
     * @param xmlOut where the transformed CSV XML is written
     * @param csvProperties used to configure CSV parsing
     * @throws CsvException
     */
    protected void transformCsvToXml(InputStream csvIn, OutputStream xmlOut,
            Properties csvProperties) throws CsvException {
        if (csvIn == null) {
            throw new CsvException("Input stream must not be null");
        }
        // creating a new CsvReader :
        CsvReader csvReader = createCsvReader(csvIn, csvProperties);
        if (csvReader == null) {
            throw new CsvException("not a CSV file");
        }

        try {
            // now creating and outputting the transformed XML :
            // NB. using DOM so content will be XML encoded
            Document csvDoc = XMLHelper.createDocumentFromString(
                "<" + DATA_TAG + " xmlns=\"" + CSVXML_NAMESPACE_URI + "\"/>");
            // NB. dataElement same as csvDoc
            Element dataElement = csvDoc.getDocumentElement();
            
            while (csvReader.readRecord()) {
                Element rowElement = csvDoc.createElementNS(CSVXML_NAMESPACE_URI, ROW_TAG);
                dataElement.appendChild(rowElement);
                int columnNb = csvReader.getColumnCount();
                
                for (int i = 0; i < columnNb; i++) {
                    String value = csvReader.get(i);
                    value = (value == null) ? "" : value; //TODO required ??
                    
                    Element cellElement = csvDoc.createElementNS(CSVXML_NAMESPACE_URI, CELL_TAG);
                    rowElement.appendChild(cellElement);
                    // NB. we could encapsulating CSV values in CDATA so anything should be ok
                    //CDATASection valueData = csvDoc.createCDATASection(value);
                    Text valueData = csvDoc.createTextNode(value);
                    cellElement.appendChild(valueData);
                }
            }
            String csvDocString = XMLHelper.createStringFromDOMDocument(csvDoc);
            xmlOut.write(csvDocString.getBytes());
        } catch (Exception e) {
            throw new CsvException("Error while parsing CSV and building output XML : "
                + e.getMessage());
        }
    }


    /**
     * Creates a new CsvReader configured according to the
     * component's properties, ready to parse CSV content.
     * @param csvIn the CSV content to be parser
     * @param csvProperties used to configure the CsvReader
     * @return a new, configured CsvReader
     * @throws CsvException if creation or configuration error
     */
    protected CsvReader createCsvReader(InputStream csvIn,
            Properties csvProperties) throws CsvException {
        try {
            // TODO use BeanUtils to set the properties ??
            char csvColumnDelimiterChar = csvProperties
                  .getProperty(CUSTOM_COLUMN_DELIMITER_CHAR_PROP, ",").charAt(0);
            char csvCustomRecordDelimiterChar = csvProperties
                  .getProperty(CUSTOM_RECORD_DELIMITER_CHAR_PROP, "\n").charAt(0);
            boolean useCustomRecordDelimiter = Boolean.valueOf(csvProperties
                .getProperty(USE_CUSTOM_RECORD_DELIMITER_PROP, "false")).booleanValue();
            int escapeMode = (Boolean.valueOf(csvProperties.getProperty(
                USE_DOUBLE_ESCAPE_MODE_PROP, "true")).booleanValue()) ?
                CsvReader.ESCAPE_MODE_DOUBLED : CsvReader.ESCAPE_MODE_BACKSLASH;
            boolean useComment = Boolean.valueOf(csvProperties
                .getProperty(USE_COMMENT_PROP, "false")).booleanValue();
            char commentChar = csvProperties.getProperty(COMMENT_CHAR_PROP, "#").charAt(0);
        
            // creating CsvReader for given input
            // csvCharset : Csv SE default is inited in TODO
            CsvReader csvReader = new CsvReader(csvIn, Charset.defaultCharset());
            csvReader.setDelimiter(csvColumnDelimiterChar); // CsvReader default is ','
            if (useCustomRecordDelimiter) {
               csvReader.setRecordDelimiter(csvCustomRecordDelimiterChar);
            } // else CsvReader default is a combination of unix & windows cr & lf
            csvReader.setEscapeMode(escapeMode); // CsvReader default is CsvReader.ESCAPE_MODE_DOUBLED i.e. '\\'
            csvReader.setUseComments(useComment);
            csvReader.setComment(commentChar); // CsvReader default is '#'
            // NB. safetySwitch is activated and may not be disabled for now
            //csvReader.setMergeConsecutive(arg0) //TODO ???
            return csvReader;
        } catch (Exception e) {
            throw new CsvException("Error setting CSV reader configuration."
                + "CSV reader properties are : " + csvProperties);
        }
    }
    
    
    /**
     * Returns the component configuration (for CsvReader etc.)
     * corresponding to the given service name
     * @param serviceName
     * @return
     * @throws Exception
     */
    protected Properties getCsvProperties(String serviceName) throws CsvException{
        Properties csvProperties = this.mapEndpointCsv.get(serviceName);
        if (csvProperties != null) {
            return csvProperties;
        } else {
            throw new CsvException("CSV properties not found for service "
                + serviceName);
        }
    }

}
