/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: Csv.java 1216 2006-11-14 18:27:09Z dutoo $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.engine.forward;

import java.net.URI;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.JBIException;
import javax.jbi.management.DeploymentException;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.InOptionalOut;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.messaging.RobustInOnly;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;

import org.objectweb.petals.component.common.basic.AbstractComponent;
import org.objectweb.petals.component.common.basic.AbstractServiceUnitManager;
import org.objectweb.petals.component.common.util.SourceHelper;
import org.objectweb.petals.component.common.util.XMLHelper;
import org.w3c.dom.DOMException;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;


/**
 * This is the Forward Component.
 * 
 * It provides a single operation (that may be anything according
 * to its corresponding WSDL definition) that chains calls to other
 * services, which must be InOut save for the last one. Chaining calls
 * here means sending to the next service the content and attachments
 * that were returned by the last service to be called previously.
 * 
 * The services and operations to be called are defined in "extension
 * properties" (meaning component-commons 1.0.1 extension properties)
 * provided in the service unit's META-INF/jbi.xml , as the text content
 * of the "service_<i>" and "operation_<i>" extension properties, i
 * being the order of the call within the chain (starting with 0).
 * 
 * Here is a sample extension properties configuration :
 * <petals-extensions:service_0>{http://petals.objectweb.org/}HelloworldService</petals-extensions:service_0>
 * <petals-extensions:operation_0>sayHello</petals-extensions:operation_0>
 * 
 * Note : in order to work, every target service must have available
 * endpoints on the network, according to their exact qualified name
 * as provided in the extension properties. That means that targeting
 * {http://petals.objectweb.org/}HelloworldService won't work if the
 * only available service is {http://petals.objectweb.org}HelloworldService
 * (missing a slash).
 * 
 * @version $Rev: 250 $Date: {date}
 * @since Petals 1.0
 * @author Marc Dutoo - Open Wide
 * 
 */
public class Forward extends AbstractComponent {
    
    public static final String INONLY_METHOD = "InOnly";
    public static final String INOUT_METHOD = "InOut";
    public static final String INOPTIONALOUT_METHOD = "InOptionalOut";
    public static final String ROBUSTINONLY_METHOD = "RobustInOnly";
    

    private Map<QName, Properties> serviceToPropertiesMap = new HashMap<QName, Properties>();

    
    /**
     * Creates a new Forward.
     *
     */
    public Forward() {
        super();
    }

    /**
     * Creates a new Forward.
     *
     */
    public Forward(DeliveryChannel channel, Logger log) {
        super(channel, log);
    }
    

    @Override
    protected boolean handleMessage(QName service, QName operation,
        NormalizedMessage in, NormalizedMessage out, URI messageExhangePattern)
        throws Exception {
        // Loading configuration for service
        Properties props = getServiceProperties(service);
        
        NormalizedMessage currentIn = in;
        NormalizedMessage currentOut = null;
        for (int i = 0; ; i++) {
            // identifying the next service and operation to call
            String targetServiceNameProp = "service_" + i;
            String targetServiceOperationProp = "operation_" + i;
            String targetServiceMethodProp = "method_" + i;
            String targetServiceName = props.getProperty(targetServiceNameProp);
            String targetServiceOperation = props.getProperty(targetServiceOperationProp, "");
            String targetServiceMethod = props.getProperty(targetServiceMethodProp, INOUT_METHOD);
            if (targetServiceName == null || targetServiceName.length() == 0) {
                // no other target service : this is the end of the forward chain
                break;
            }

            // getting available endpoints
            QName targetService = QName.valueOf(targetServiceName);
            ServiceEndpoint[] targetEndpoints = this.getComponentContext()
                .getEndpointsForService(targetService);
            if (targetEndpoints == null || targetEndpoints.length == 0) {
                throw new Exception("No endpoint found for service "
                    + targetServiceName);
            }
            
            if (currentIn == null) {
                throw new Exception("Previous forward had no OUT message "
                    + "so there can't be another forward afterwards to "
                    + targetServiceName + "." + targetServiceOperation);
            }
            
            try {
               currentOut = forwardCopy(currentIn,
                   targetService, targetServiceOperation, targetServiceMethod);
            } catch (Exception e) {
                throw new Exception("Error when forwarding to service and operation "
                    + targetServiceName + "." + targetServiceOperation, e);
            }
            if (currentOut == null) {
                // the out message was already treated
                // NB. we'll check that there will be no more forward
                // and return "<success/>"
                continue;
                
            } else if (currentOut instanceof Fault) {
                String faultString = SourceHelper.createString(currentOut.getContent());
                throw new Exception("Service fault received from call "
                    + "to service and operation " + targetServiceName
                    + "." + targetServiceOperation + " : " + faultString);
            }
            
            // preparing next forward
            currentIn = currentOut;
        }

        if (out != null) {
            if (currentOut == null) {
                // last forward was InOnly, so let's return "<success/>"
                out.setContent(SourceHelper.createSource("<success/>"));
            } else {
                // let's return the result of the last forward
                copy(currentOut, out);
            }
        } // else called by in InOnly mode : nothing to do
        return true;
    }
    
    
    /**
     * Copies a message's content and attachments in another.
     * @param originalNm
     * @param copyNm
     * @throws Exception
     */
    protected final static void copy(NormalizedMessage originalNm,
        NormalizedMessage copyNm) throws Exception {
        // setting content, with hardcopy (because if semantic routing...)
        String contentString = SourceHelper.createString(originalNm.getContent());
        copyNm.setContent(SourceHelper.createSource(contentString));

        // setting attachments if any
        Set inAttachmentNameSet = originalNm.getAttachmentNames();
        if (inAttachmentNameSet != null) {
            for (Iterator nameIt = inAttachmentNameSet.iterator(); nameIt.hasNext();) {
                String attachmentName = (String) nameIt.next();
                DataHandler attDh = originalNm.getAttachment(attachmentName);
                try {
                    copyNm.addAttachment(attachmentName, attDh);
                } catch (MessagingException e) {
                    throw new Exception("Error when attaching file "
                        + attachmentName, e);
                }
            }
        }
    }
    
    
    /**
     * Forwards a copy of the given message to the given service
     * using the given operation and method (i.e. InOut etc.)
     * @param originalIn
     * @param service
     * @param operation
     * @param method
     * @return
     * @throws Exception
     */
    protected NormalizedMessage forwardCopy(NormalizedMessage originalIn,
        QName service, String operation, String method) throws Exception {
        // creating message exchange
        MessageExchangeFactory mef = this.getDeliveryChannel().createExchangeFactory();
        MessageExchange msg;
        if (INONLY_METHOD.equals(method)) {
            msg = mef.createInOnlyExchange();
        } else if (INOUT_METHOD.equals(method)) {
            msg = mef.createInOutExchange();
        } else if (INOPTIONALOUT_METHOD.equals(method)) {
            msg = mef.createInOptionalOutExchange();
        } else if (ROBUSTINONLY_METHOD.equals(method)) {
            msg = mef.createRobustInOnlyExchange();
        } else {
            throw new Exception("Unknown method " + method);
        }
        
        // setting target
        msg.setService(service);
        msg.setOperation(QName.valueOf(operation));
        
        // copying IN message
        NormalizedMessage copyIn = msg.createMessage();
        copy(originalIn, copyIn);
        if (INONLY_METHOD.equals(method)) {
            ((InOnly) msg).setInMessage(copyIn);
        } else if (INOUT_METHOD.equals(method)) {
            ((InOut) msg).setInMessage(copyIn);
        } else if (INOPTIONALOUT_METHOD.equals(method)) {
            ((InOptionalOut) msg).setInMessage(copyIn);
        } else if (ROBUSTINONLY_METHOD.equals(method)) {
            ((RobustInOnly) msg).setInMessage(copyIn);
        } else {
            throw new Exception("Unknown method " + method);
        }

        // doing the actual message send
        boolean ok = this.getDeliveryChannel().sendSync(msg, 0);
        if(ok) {
            if (ExchangeStatus.DONE.equals(msg.getStatus())) {
                return null; // means DONE status received
                
            } else if (ExchangeStatus.ERROR.equals(msg.getStatus())) {
                String err = "ERROR status ! ";
                log.log(Level.SEVERE, err, msg.getError());
                throw new Exception(err, msg.getError()); // timeout
                
            } else {
                try {
                    NormalizedMessage out = null;
                    // NB. maybe necessary to do again full copy with new NormalizedMessageImpl()
                    if (msg.getMessage("OUT") != null) {
                        out = msg.getMessage("OUT");
                    } else if (msg.getFault() != null) {
                        out = msg.getFault();
                    } else {
                        // will be the same as DONE status received
                    }
                    NormalizedMessage copyOut = msg.createMessage();
                    copy(out, copyOut);

                    msg.setStatus(ExchangeStatus.DONE);
                    this.getDeliveryChannel().send(msg);
                    return out;
                    
                } catch (Exception e) {
                    String err = "Error while getting response message";
                    log.log(Level.SEVERE, err, e);
                    throw new Exception(err, e); // timeout
                }
            }
        } else {
            String err = "Timeout !";
            log.log(Level.SEVERE, err);
            throw new Exception(err); // timeout
        }
    }



    /**
     * TODO MAY GO IN COMMONS
     * Returns the given content as a String.
     * Helper method to get everything there is under the top tag
     * as a String. XML nodes will be formatted and included within.
     * NB. For XQuare BC, the point is to allow users to write results
     * formatting in XQuery code without having to XML-encode the tags.
     * @param contentSource the content of an incoming JBI message
     * @return a String. NB. For XQuare BC, an XQuery.
     * @throws ContentException if XML transformation error
     */
    protected String transformToContentString(Source contentSource) throws ContentException {
        try {
            Node contentDocumentNode = XMLHelper.createDOMNodeFromSource(contentSource);
            Node contentNode = XMLHelper.getFirstChild(contentDocumentNode);
            // Now reading the XQuery code in the most "nice" way
            NodeList queryNodes = contentNode.getChildNodes();
            int queryNodeNb = queryNodes.getLength();
            if (queryNodeNb == 1) {
                Node singleQueryNode = queryNodes.item(0);
                if (singleQueryNode instanceof Text) {
                    return ((Text) singleQueryNode).getTextContent();
                }
            }
            StringBuffer contentBuf = new StringBuffer();
            for (int i = 0; i < queryNodeNb ; i++) {
                Node queryNode = queryNodes.item(i);
                // getting the full XML of the node in order to help the user pass
                // along Results XML formatting
                contentBuf.append(XMLHelper.createStringFromDOMNode(queryNode));
                contentBuf.append("\n");
            }
            return contentBuf.toString();
            
        } catch (DOMException e) {
            String msg = "Error while reading content string from XML Source "
                + tryToCreateStringForError(contentSource);
            log.log(Level.SEVERE, msg, e);
            throw new ContentException(msg, e);
        } catch (TransformerException e) {
            String msg = "Error while reading content string from XML Source "
                + tryToCreateStringForError(contentSource);
            log.log(Level.SEVERE, msg, e);
            throw new ContentException(msg, e);
        }
    }

    /**
     * TODO MAY GO IN COMMONS
     * Returns the given content as a String that must be wrapped under
     * a given top tag (checked).
     * Helper method to get everything there is under a given top tag
     * as a String. XML nodes will be formatted and included within.
     * NB. For XQuare BC, the point is to allow users to write results
     * formatting in XQuery code without having to XML-encode the tags.
     * @param contentSource the content of an incoming JBI message
     * @param contentTag the top tag of the contentSource
     * @return a String. NB. For XQuare BC, an XQuery.
     * @throws ContentException if XML transformation error or bad top tag
     */
    protected String transformToContentStringWithRootTag(
        Source contentSource, String contentTag) throws ContentException {
        try {
            Node contentDocumentNode = XMLHelper.createDOMNodeFromSource(contentSource);
            Node contentNode = XMLHelper.getFirstChild(contentDocumentNode);
            // Checking that node element tag is the required one
            if (!contentTag.equals(contentNode.getLocalName())) {
                throw new ContentException("Such messages should be "
                    + "encapsulated in a top <\"" + contentTag + "\"> tag !");
            }
            // Now reading the XQuery code in the most "nice" way
            NodeList queryNodes = contentNode.getChildNodes();
            int queryNodeNb = queryNodes.getLength();
            if (queryNodeNb == 1) {
                Node singleQueryNode = queryNodes.item(0);
                if (singleQueryNode instanceof Text) {
                    return ((Text) singleQueryNode).getTextContent();
                }
            }
            StringBuffer contentBuf = new StringBuffer();
            for (int i = 0; i < queryNodeNb ; i++) {
                Node queryNode = queryNodes.item(i);
                // getting the full XML of the node in order to help the user pass
                // along Results XML formatting
                contentBuf.append(XMLHelper.createStringFromDOMNode(queryNode));
                contentBuf.append("\n");
            }
            return contentBuf.toString();
            
        } catch (DOMException e) {
            String msg = "Error while reading content string from XML Source "
                + tryToCreateStringForError(contentSource);
            log.log(Level.SEVERE, msg, e);
            throw new ContentException(msg, e);
        } catch (TransformerException e) {
            String msg = "Error while reading content string from XML Source "
                + tryToCreateStringForError(contentSource);
            log.log(Level.SEVERE, msg, e);
            throw new ContentException(msg, e);
        }
    }


    /**
     * TODO MAY GO IN COMMONS
     * Returns the given content as a DOM node.
     * @param contentSource the content of an incoming JBI message
     * @return the corresponding DOM node tree
     * @throws ContentException if XML transformation error
     */
    protected Node transformToContentNode(Source contentSource) throws ContentException {
         try {
             Node contentDocumentNode = XMLHelper.createDOMNodeFromSource(contentSource);
             Node contentNode = XMLHelper.getFirstChild(contentDocumentNode);
             return contentNode;
        } catch (TransformerException e) {
            String msg = "Error while reading content node from XML Source "
                + tryToCreateStringForError(contentSource);
            log.log(Level.SEVERE, msg, e);
            throw new ContentException(msg, e);
        }
    }

    /**
     * TODO MAY GO IN COMMONS
     * Returns the given content as a DOM node that must be wrapped under
     * a given top tag (checked).
     * @param contentSource the content of an incoming JBI message
     * @param contentTag the top tag of the contentSource
     * @return the corresponding DOM node tree
     * @throws ContentException if XML transformation error or bad top tag
     */
    protected Node transformToContentNodeWithRootTag(
        Source contentSource, String contentTag) throws ContentException {
        try {
            Node contentDocumentNode = XMLHelper.createDOMNodeFromSource(contentSource);
            Node contentNode = XMLHelper.getFirstChild(contentDocumentNode);
            // Checking that node element tag is the required one
            if (!contentTag.equals(contentNode.getLocalName())) {
                throw new ContentException("Such messages should be "
                    + "encapsulated in a top <\"" + contentTag + "\"> tag !");
            }
            return contentNode;
        } catch (TransformerException e) {
            String msg = "Error while reading content node from XML Source "
                + tryToCreateStringForError(contentSource);
            log.log(Level.SEVERE, msg, e);
            throw new ContentException(msg, e);
        }
    }
    
    
    public static final String tryToCreateStringForError(Source contentSource) {
        String contentString;
        try {
            // NB. SourceHelper.createString() does a reset on Stream sources
            contentString = SourceHelper.createString(contentSource);
        } catch (Exception e) {
            // do nothing
            contentString = "[[error in transforming Source to String]]";
        }
        return contentString;
    }


    /**
     * Returns the targeted service unit's properties.
     * @param service the targeted service
     * @return its configuration
     * @throws XQuareBCException
     */
    protected Properties getServiceProperties(QName service) throws Exception {
        Properties csvProperties = this.serviceToPropertiesMap.get(service);
        if (csvProperties != null) {
            return csvProperties;
        } else {
            throw new Exception("Service properties not found for service "
                + service);
        }
    }


    @Override
    protected AbstractServiceUnitManager createServiceUnitManager() throws DeploymentException {
        // NB. the super impl returns a new SimpleServiceUnitManager with no extension handling
        PropertiesServiceUnitManager suMgr;
        try {
            suMgr = new PropertiesServiceUnitManager(this.getComponentContext(),
                log, serviceToPropertiesMap);
        } catch (JBIException e) {
            throw new DeploymentException("Unable to get component context", e);
        }
        return suMgr;
    }

}
