/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: Csv.java 1216 2006-11-14 18:27:09Z dutoo $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.engine.forward;

import java.io.File;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.management.DeploymentException;
import javax.xml.namespace.QName;

import org.objectweb.petals.component.common.basic.AbstractServiceUnitManager;
import org.objectweb.petals.component.common.util.ManagementMessageUtil;
import org.objectweb.petals.component.common.util.WSDLHelper;
import org.w3c.dom.Document;
import org.w3c.dom.Node;


/**
 * Represents a PropertiesServiceUnitManager.
 * This SU Manager gathers service-level "extension properties"
 * (as defined in component-common 1.0.1) and makes them available
 * as Properties to the component's message processor(s).
 * 
 * NB. services are identified by their full exact qualified name, meaning
 * a missing slash will make it fail.
 * 
 * @version $Rev: 250 $Date: {date}
 * @since Petals 1.0
 * @author Marc Dutoo - Open Wide
 * 
 */
public class PropertiesServiceUnitManager extends AbstractServiceUnitManager {

    private Map<QName, Properties> serviceToPropertiesMap;

    /**
     * Creates a new PropertiesServiceUnitManager that will put all extensions
     * in the given service QName to properties map.
     * NB. services are identified by their full exact qualified name, meaning
     * a missing slash will make it fail.
     * @param context
     * @param logger
     * @param serviceToPropertiesMap
     * @throws DeploymentException
     */
    public PropertiesServiceUnitManager(ComponentContext context, Logger logger,
        Map<QName,Properties> serviceToPropertiesMap) throws DeploymentException {
        super(context, logger);
        this.serviceToPropertiesMap = serviceToPropertiesMap;
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#start(java.lang.String)
     * 
     * Endpoints are activated here
     */
    public void start(String serviceUnitName) throws DeploymentException {
        logger.log(Level.FINE, "start  serviceUnitName  " + serviceUnitName);
        String suRootPath = serviceUnitInstallationRootPath
                .get(serviceUnitName);
        Document serviceDesc = getServiceUnitWSDL(suRootPath);
        if (serviceDesc != null) {
            try {
                activateEndpointsFromJBIDescription(serviceDesc,
                        serviceUnitName, suRootPath);

                // starting extensions
                File jbiXmlFile = new File(suRootPath + File.separator
                        + "META-INF" + File.separator + "jbi.xml");
                Node servicesNode = getNode(jbiXmlFile, "services");
                Map<String, String> extensions = getExtensions(servicesNode);
                start(serviceUnitName, suRootPath, serviceDesc, extensions);
            } catch (Exception ex) {
                logger.log(Level.SEVERE, FAILED_ACTIVATE_ENDPOINT + ex);
                throw new DeploymentException(FAILED_ACTIVATE_ENDPOINT, ex);
            }
        } else {
            throw new DeploymentException(INCOMPLETE_SERVICE_UNIT_PACKAGE);
        }
    }

    /**
     * Here is the core implementation of the Properties SU Manager.
     * This method puts all extension properties that have been found
     * during the deploy phase in a service-wide Properties instance.
     * These properties may be accessed by the component that was the
     * creator of this Properties SU Manager.
     * @param serviceUnitName
     * @param serviceUnitRootPath
     * @param serviceDesc
     * @param extensionMap
     * @throws Exception
     */
    protected void start(String serviceUnitName, String serviceUnitRootPath,
        Document serviceDesc, Map<String, String> extensionMap) throws Exception {
        // Now making extensions available as service properties
        Properties serviceProps = new Properties();
        serviceProps.putAll(extensionMap);
        QName service = WSDLHelper.getServiceNameFromWSDLDocument(serviceDesc)[0];
        serviceToPropertiesMap.put(service, serviceProps);
    }
    

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#stop(java.lang.String)
     * 
     * Endpoints are deactivated here
     */
    public void stop(String serviceUnitName) throws DeploymentException {
        try {
            serviceUnitInstallationRootPath.remove(serviceUnitName);
            deactivateEndpointsFromJBIDescription(serviceUnitName);
        } catch (Exception e) {
            throw new DeploymentException(FAILED_DEACTIVATE_ENDPOINT, e);
        }
        logger.log(Level.FINE, "stop  serviceUnitName  " + serviceUnitName);
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#undeploy(java.lang.String,
     *      java.lang.String)
     */
    public String undeploy(String serviceUnitName, String serviceUnitRootPath)
        throws DeploymentException {
        return ManagementMessageUtil.getComponentTaskResult(context
                .getComponentName(), "undeploy",
                ManagementMessageUtil.TASK_RESULT_SUCCESS);
    }

    /**
     * Deploy a service unit. Actually simply gets extension
     * properties (as defined in component-common 1.0.1) and
     * let subclasses use them by calling deploy(service, extension).
     * NB. This is an override of the orginal one that adds
     * the service being configured to the context of the call.
     * 
     * @param serviceUnitName
     *            name of the service unit to deploy
     * @param serviceUnitRootPath
     *            root path of the service unit to deploy
     * @return a valid management message
     * @throws DeploymentException
     *             If an error occurs during deployment
     */
    public String deploy(String serviceUnitName, String serviceUnitRootPath)
        throws DeploymentException {
        String result = null;
        if (serviceUnitMap.get(serviceUnitName) == null) {
            // deploying extensions
            File jbiXmlFile = new File(serviceUnitRootPath + File.separator
                    + "META-INF" + File.separator + "jbi.xml");
            Node servicesNode = getNode(jbiXmlFile, "services");
            Map<String, String> extensions = getExtensions(servicesNode);
            deploy(serviceUnitName, serviceUnitRootPath, extensions);
            
            logger.log(Level.FINE, "deploy serviceUnitName  "
                    + serviceUnitName + "serviceUnitRootPath"
                    + serviceUnitRootPath);
            result = ManagementMessageUtil.getComponentTaskResult(context
                    .getComponentName(), "deploy",
                    ManagementMessageUtil.TASK_RESULT_SUCCESS);
        } else {
            throw new DeploymentException(ERROR_DUPLICATE_SERVICE_UNIT);
        }
        return result;
    }

    protected void deploy(String serviceUnitName, String serviceUnitRootPath,
        Map<String, String> arg0) {
        // NB. not used here
    }

    @Override
    protected void deploy(Map<String, String> arg0) {
        // TODO remove from the API of AbstractServiceUnitManager
    }

}
