/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: JmsConnexion.java 676 2006-06-27 15:44:03Z alouis $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.jms.su;

import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLClassLoader;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.spi.InitialContextFactory;

import org.objectweb.petals.binding.jms.JMSConnection;
import org.objectweb.petals.component.common.PEtALSComponentSDKException;
import org.objectweb.petals.component.common.su.ServiceUnitListener;
import org.objectweb.petals.component.common.util.PetalsExtensionsUtil;
import org.objectweb.petals.tools.jbicommon.descriptor.Extensions;
import org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor;

import static org.objectweb.petals.component.common.util.ClassLoaderUtil.createClassLoader;

/**
 * This class is used to store information about an external JMS queue acting as
 * an external service
 * 
 * @version $Revision: $ $Date: $
 * @since Petals 1.0
 * @author alouis
 */
public class JMSServiceUnitListener implements ServiceUnitListener {

	private static final String CONNECTION_FACTORY = "connection-factory";

	private static final String ADDRESS = "address";

	private static final String INITIAL_CONTEXT_FACTORY = "initial-context-factory";

	private static final String USER = "user";

	private static final String PASSWORD = "password";

	private static final String TRANSACTED = "transacted";

	/**
	 * Contains jmsProducerConnections, stored by 'address' (ie jms url
	 * provider)
	 */
	private Map<String, JMSConnection> jmsProducerConnections;

	/**
	 * Contains jmsConsumerConnections, stored by 'address' (ie jms url
	 * provider)
	 */
	private Map<String, JMSConnection> jmsConsumerConnections;

	/**
	 * Contains addresses (ie jms url provider), stored by serviceUnitName
	 */
	private Map<String, String> addresses;

	/**
	 * Constructor
	 * 
	 * @param jmsProducerConnections
	 *            jmsConnections map used to store jmsConnections for producers.
	 *            Used by the JBI listener
	 * @param jmsConsumerConnections
	 *            jmsConnections map used to store jmsConnections for consumers.
	 *            Used by the external listeners
	 */
	public JMSServiceUnitListener(
			Map<String, JMSConnection> jmsProducerConnections,
			Map<String, JMSConnection> jmsConsumerConnections) {
		this.jmsProducerConnections = jmsProducerConnections;
		this.jmsConsumerConnections = jmsConsumerConnections;
		addresses = new HashMap<String, String>();
	}

	/**
	 * Prepare JMS connection for a service unit. Set a JMS producer or JMS
	 * consumer depending on JBI 'provides' or JBI 'consumes' descriptor
	 * information. Only one 'consumes' or 'provides' section is allowed
	 */
	public void onSUDeployed(String serviceUnitName, String suRootPath,
			JBIDescriptor descriptor) throws PEtALSComponentSDKException {

		int providesCount = descriptor.getServices().getProvides().size();
		int consumesCount = descriptor.getServices().getConsumes().size();

		if (providesCount + consumesCount != 1) {
			throw new PEtALSComponentSDKException(
					"Only one 'Consumes' or 'Provides' is allowed per JMS-BC ServiceUnit.");

		}
		JMSConnection.Role role = null;
		Extensions extensions = null;

		if (providesCount == 1) {
			role = JMSConnection.Role.JMS_PRODUCER;
			extensions = descriptor.getServices().getProvides().get(0)
					.getExtensions();

		} else {
			role = JMSConnection.Role.JMS_CONSUMER;
			extensions = descriptor.getServices().getConsumes().get(0)
					.getExtensions();
		}

		// Create the JNDI context used to retrieve jms connection and
		// destination
		Context context = setupContext(suRootPath, extensions);

		// create the jmsConnection object that contains all information
		JMSConnection connection = setupConnection(context, extensions, role);

		// retrieve the jms 'address' to be used as key
		String address = PetalsExtensionsUtil
				.extractValueFromKeyValueExtension(extensions, ADDRESS);

		// store the jmsconnection
		addresses.put(serviceUnitName, address);
		if (role == JMSConnection.Role.JMS_PRODUCER) {
			jmsProducerConnections.put(address, connection);
		} else {
			jmsConsumerConnections.put(address, connection);
		}
	}

	/**
	 * Start the jms connection and setup the jms producer/consumer
	 */
	public void onSUStarted(String serviceUnitName)
			throws PEtALSComponentSDKException {

		String address = addresses.get(serviceUnitName);

		if (address != null) {
			JMSConnection jmsConnection = jmsProducerConnections.get(address);
			if (jmsConnection == null) {
				jmsConnection = jmsConsumerConnections.get(address);
			}
			jmsConnection.start();
		}
	}

	/**
	 * Stop the jms connection
	 */
	public void onSUStopped(String serviceUnitName)
			throws PEtALSComponentSDKException {

		String address = addresses.get(serviceUnitName);

		if (address != null) {
			JMSConnection jmsConnection = jmsProducerConnections.get(address);
			if (jmsConnection == null) {
				jmsConnection = jmsConsumerConnections.get(address);
			}
			jmsConnection.stop();
		}
	}

	public void onSUUndeployed(String serviceUnitName, String suRootPath,
			JBIDescriptor descriptor) throws PEtALSComponentSDKException {

		String address = addresses.remove(serviceUnitName);

		if (address != null) {
			JMSConnection jmsConnection = jmsProducerConnections
					.remove(address);
			if (jmsConnection == null) {
				jmsConsumerConnections.remove(address);
			}
		}
	}

	/**
	 * Setup a JNDIContext to access a JMS destination.
	 * 
	 * @param suRootPath
	 *            path where jndi context factory class can be found
	 * @param extensions
	 *            jndi properties
	 * @return a JNDI context initialized with the given jbi.xml properties
	 * @throws PEtALSComponentSDKException
	 */
	private Context setupContext(String suRootPath, Extensions extensions)
			throws PEtALSComponentSDKException {

		// Create a class loader that will contains all JAR
		// It will load the specific JndiFactory class
		// to access the jndi containing Queue and Topic references
		URLClassLoader classLoader = createClassLoader(suRootPath, this
				.getClass().getClassLoader());

		// Create JNDI properties
		Properties jndiProperties = createJNDIProperties(extensions);

		// Create the JNDI context
		return createJNDIContext(classLoader, jndiProperties);
	}

	/**
	 * extract jndi properties from the jbi.xml extensions
	 * 
	 * @param extensions
	 * @return
	 * @throws PEtALSComponentSDKException
	 *             provider url or factory name not found
	 */
	private Properties createJNDIProperties(Extensions extensions)
			throws PEtALSComponentSDKException {
		Properties outProperties = new Properties();

		String url = null;
		try {
			URI address = new URI(PetalsExtensionsUtil
					.extractValueFromKeyValueExtension(extensions, ADDRESS));
			url = address.getScheme() + ":" + address.getSchemeSpecificPart();
		} catch (URISyntaxException e) {
			throw new PEtALSComponentSDKException(
					ADDRESS
							+ " property not found in the Service Unit descriptor file."
							+ e);
		}
		String factory = PetalsExtensionsUtil
				.extractValueFromKeyValueExtension(extensions,
						INITIAL_CONTEXT_FACTORY);

		outProperties.setProperty(InitialContext.PROVIDER_URL, url);
		outProperties.setProperty(InitialContext.INITIAL_CONTEXT_FACTORY,
				factory);
		return outProperties;
	}

	/**
	 * Return a JNDI context loaded from the JAR files contained in the
	 * ServiceUnit and the Properties defined in the JBI.XML file.
	 * 
	 * @param classLoader
	 *            contains the libraries where the JNDI context factory can be
	 *            found
	 * @param properties
	 *            contains the factory class name and the url provider
	 * @return a JNDI context
	 * @throws PEtALSComponentSDKException
	 *             can not retrieved the factory. properties might not be
	 *             correct.
	 */
	private Context createJNDIContext(ClassLoader classLoader,
			Properties properties) throws PEtALSComponentSDKException {
		Context ctx = null;
		try {
			InitialContextFactory icf = (InitialContextFactory) classLoader
					.loadClass(
							properties
									.getProperty(InitialContext.INITIAL_CONTEXT_FACTORY))
					.newInstance();
			ctx = icf.getInitialContext(properties);
		} catch (Exception e) {
			throw new PEtALSComponentSDKException(
					"JNDI context can not be created." + e);
		}
		return ctx;
	}

	/**
	 * Setup a JMS connection to access a JMS destination.
	 * 
	 * @param context
	 *            jndi context where the connection factory and the destination
	 *            can be found
	 * @param extensions
	 *            contains connectionFactory and destination names, user/pwd,
	 *            transacted and acknowledge modes.
	 * @param role
	 *            jms role : consumer or producer
	 * @return a JMSConnection in a stopped state.
	 * @throws PEtALSComponentSDKException
	 *             a mandatory property is not found in the JBI.XML descriptor
	 */
	private JMSConnection setupConnection(Context context,
			Extensions extensions, JMSConnection.Role role)
			throws PEtALSComponentSDKException {
		String connName = PetalsExtensionsUtil
				.extractValueFromKeyValueExtension(extensions,
						CONNECTION_FACTORY);

		String destName = null;
		try {
			URI address = new URI(PetalsExtensionsUtil
					.extractValueFromKeyValueExtension(extensions, ADDRESS));
			destName = address.getFragment();
		} catch (URISyntaxException e) {
			throw new PEtALSComponentSDKException(
					ADDRESS
							+ " property not found in the Service Unit descriptor file."
							+ e);
		}

		String user = PetalsExtensionsUtil.extractValueFromKeyValueExtension(
				extensions, USER);

		String password = PetalsExtensionsUtil
				.extractValueFromKeyValueExtension(extensions, PASSWORD);

		boolean transacted = false;
		String transac = PetalsExtensionsUtil
				.extractValueFromKeyValueExtension(extensions, TRANSACTED);
		transacted = Boolean.valueOf(transac);

		return new JMSConnection(context, connName, destName, transacted, user,
				password, role);

	}

}