/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: SessionDescriptorBuilder.java 154 27 sept. 06 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.mail.listeners;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.List;
import java.util.logging.Logger;

import javax.jbi.JBIException;

import org.objectweb.petals.binding.mail.MailBCImpl;
import org.objectweb.petals.component.common.util.MessageExchangeWrapper;
import org.objectweb.petals.tools.jbicommon.util.StringHelper;

/**
 * This class is used to extract information from an SMTP URI String
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
public class SessionDescriptorBuilder {

    protected Logger log;

    /**
     * @param address
     *            address extracted from a "provides" node of the jbi
     *            descriptor.
     *            <p>
     *            Address conforms to the following pattern (a valid URI
     *            pattern):
     *            <p>
     *            smtp://<i>user</i>:<i>password</i>@<i>host</i>:<i>port</i>/from:<i>fromAddress</i>/to:<i>toAddress</i>.
     *            <p>
     *            <i>fromAddress</i> and <i>toAddress</i> are classical email
     *            addresses like : foo@bar.com.
     * @param log
     *            the component logger
     */
    public SessionDescriptorBuilder(Logger log) {
        super();
        this.log = log;
    }

    /**
     * Retrieve mail session properties from the given address and message
     * exchange.
     * 
     * @param address
     *            the address uri
     * @param exchangeWrapper
     *            the message exchange wrapper used to extract an optionnal
     *            destination address
     * @return a Map of properties extracted from the given address.
     * @throws JBIException
     *             if address isn't a valid URI or if checking fails
     */
    public SessionDescriptor build(String address,
        MessageExchangeWrapper exchangeWrapper) throws JBIException {
        String msg = "";
        SessionDescriptor sessionDescriptor = null;

        // Transform String address into URI address
        URI addressURI;
        try {
            addressURI = new URI(address);
        } catch (URISyntaxException e) {
            msg = "The given address isn't a valid URI: " + address;
            log.severe(msg);
            throw new JBIException(msg);
        }

        /*
         * Retrieve properties values from address URI
         */
        // Retrieve user and password
        String username = null;
        String password = null;
        String userInfo = addressURI.getRawUserInfo();
        if (!StringHelper.isNullOrEmpty(userInfo)) {
            if (userInfo.indexOf(":") > 0) {
                username = addressURI.getUserInfo().split(":")[0];
                password = addressURI.getUserInfo().split(":")[1];
            } else {
                username = addressURI.getUserInfo();
            }
        }

        // Retrieve host and port
        String hostname = addressURI.getHost();
        int port = addressURI.getPort();

        // Retrieve protocol (smtp, imap, pop3)
        String scheme = addressURI.getScheme();

        // Retrieve "from" and "to" addresses
        String fromAddress = "";
        String toAddress = "";
        String path = addressURI.getPath();
        List<String> pathElements = StringHelper.splitPathElements(path);
        for (String string : pathElements) {
            if (string.indexOf(MailBCImpl.FROM_PATHELEMENT) >= 0) {
                fromAddress = string.substring(string
                    .indexOf(MailBCImpl.FROM_PATHELEMENT)
                    + MailBCImpl.FROM_PATHELEMENT.length());
            } else if (string.indexOf(MailBCImpl.TO_PATHELEMENT) >= 0) {
                toAddress = string.substring(string
                    .indexOf(MailBCImpl.TO_PATHELEMENT)
                    + MailBCImpl.TO_PATHELEMENT.length());
            }
        }
        // if the destination address is null, try to find it in the message
        // exchange, in case of an SMTP scheme
        if (MailBCImpl.MAIL_SCHEME_SMTP.equalsIgnoreCase(scheme)
            && StringHelper.isNullOrEmpty(toAddress)) {
            toAddress = (String) exchangeWrapper
                .getProperty(MailBCImpl.DESTINATION_ADDRESS);
        }

        String query = addressURI.getQuery();
        // Retrieve period parameter ("checking for new mail" interval)
        String period = StringHelper.extractValueForAttribute(query,
            MailBCImpl.PERIOD_QUERYELEMENT, MailBCImpl.QUERY_SEPARATOR);

        // Retrieve folder parameter (the folder to search for new mails)
        String folder = StringHelper.extractValueForAttribute(query,
            MailBCImpl.FOLDER_QUERYELEMENT, MailBCImpl.QUERY_SEPARATOR);

        // Check obligatory properties
        checkProperties(hostname, toAddress, scheme);

        // Build the SessionDescriptor, use default value if
        // null or empty
        sessionDescriptor = fillProperties(hostname, port, username, password,
            fromAddress, toAddress, period, folder, scheme);

        return sessionDescriptor;
    }

    /**
     * Check the properties extracted from the address URI. Only hostname is
     * obligatory for all schemes. Only SMTP, POP3 and IMAP schemes are
     * supported. "toAddress" is need for SMTP scheme.
     * 
     * @param hostname
     * @param toAddress
     * @throws JBIException
     *             if a checking step failed
     */
    protected void checkProperties(String hostname, String toAddress,
        String scheme) throws JBIException {
        String msg = "";
        if (StringHelper.isNullOrEmpty(hostname)) {
            msg = "The host name of the requested external "
                + "service must be specified";
            log.severe(msg);
            throw new JBIException(msg);
        }
        if (!(MailBCImpl.MAIL_SCHEME_IMAP.equalsIgnoreCase(scheme)
            || MailBCImpl.MAIL_SCHEME_POP3.equalsIgnoreCase(scheme) || MailBCImpl.MAIL_SCHEME_SMTP
            .equalsIgnoreCase(scheme))) {
            msg = "The specified scheme isn't supported : " + scheme;
            log.severe(msg);
            throw new JBIException(msg);
        }
        if (MailBCImpl.MAIL_SCHEME_SMTP.equals(scheme)) {
            if (StringHelper.isNullOrEmpty(toAddress)) {
                msg = "Destination email address must be specified "
                    + "in the Service Unit descriptor ('provides' node) or in the "
                    + MailBCImpl.DESTINATION_ADDRESS
                    + " Message Exchange property";
                log.severe(msg);
                throw new JBIException(msg);
            }
        }
    }

    /**
     * Build a {@link SessionDescriptor}, use default value if null or empty
     * 
     * @param hostname
     *            the hostname, non null and non empty
     * @param port
     *            the port, default value : "25"
     * @param username
     *            the username, default value : no default
     * @param password
     *            the password, default value : no default
     * @param fromAddress
     *            the return address, default value : no default
     * @param toAddress
     *            the destination address, non null and non empty
     * @param period
     *            time elapse between two new mail checkings, default value :
     *            "60000" ms
     * @param folder
     *            the folder to check for new mails, default value : "INBOX"
     * @param scheme
     *            the access or transport protocol "smtp", "imap" or "pop3", non
     *            null and non empty
     * @return the {@link SessionDescriptor}
     */
    protected SessionDescriptor fillProperties(String hostname, int port,
        String username, String password, String fromAddress, String toAddress,
        String period, String folder, String scheme) {
        SessionDescriptor sessionDescriptor = new SessionDescriptor();

        sessionDescriptor.setHostname(hostname);
        if (port == -1) {
            if (MailBCImpl.MAIL_SCHEME_POP3.equals(scheme)) {
                sessionDescriptor.setPort(MailBCImpl.MAIL_POP3_PORT_DEFAULT);
            } else if (MailBCImpl.MAIL_SCHEME_IMAP.equals(scheme)) {
                sessionDescriptor.setPort(MailBCImpl.MAIL_IMAP_PORT_DEFAULT);
            } else {
                sessionDescriptor.setPort(MailBCImpl.MAIL_SMTP_PORT_DEFAULT);
            }
        } else {
            sessionDescriptor.setPort(new Integer(port).toString());
        }
        sessionDescriptor.setUsername(username);
        sessionDescriptor.setPassword(password);
        sessionDescriptor.setFromAddress(fromAddress);
        sessionDescriptor.setToAddress(toAddress);
        if (!StringHelper.isNullOrEmpty(period)) {
            sessionDescriptor.setPeriod(period);
        } else {
            sessionDescriptor.setPeriod(MailBCImpl.PERIOD_DEFAULT);
        }
        if (!StringHelper.isNullOrEmpty(folder)) {
            sessionDescriptor.setFolder(folder);
        } else {
            sessionDescriptor.setFolder(MailBCImpl.FOLDER_DEFAULT);
        }
        sessionDescriptor.setScheme(scheme);

        return sessionDescriptor;
    }
}
