/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id : $ 
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.xquarebc;

import java.io.File;
import java.util.HashMap;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.Component;
import javax.jbi.component.ComponentContext;
import javax.jbi.component.ComponentLifeCycle;
import javax.jbi.component.ServiceUnitManager;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.ObjectName;
import javax.xml.namespace.QName;

import org.objectweb.petals.binding.xquarebc.listeners.XQuareBCJBIProcessor;
import org.objectweb.petals.binding.xquarebc.listeners.XQuareBCListener;
import org.objectweb.petals.component.common.listener.MessageExchangeListener;
import org.objectweb.petals.component.common.serviceunitmanager.manager.PetalsServiceUnitManager;
import org.objectweb.petals.component.common.util.ComponentLogger;
import org.objectweb.petals.component.common.util.XMLHelper;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Node;

/**
 * Represents an XQuare BC JBI Component.
 * <p>
 * Holds :
 * <ul>
 * <li>the XQuareBCJBIProcessor allowing to interact with the
 * database (insert and query) using JBI messages,</li>
 * <li>and the
 * XQuareBCListener which may emit "new data" event-like JBI messages
 * according to their configuration (see their javadoc for details).</li>
 * </ul>
 * When starting, registers a single XQuareSUHandler to handle service units
 * deployed onto the XQuare BC.
 * 
 * @version $Rev: 250 $Date: {date}
 * @since Petals 1.0
 * @author Marc Dutoo - Open Wide
 * 
 */
public class XQuareBC implements Component, ComponentLifeCycle {

    protected ComponentContext context;
    protected Logger logger;
    protected HashMap<String, Properties> serviceToPropertiesMap;
    protected DeliveryChannel channel;
    
    protected PetalsServiceUnitManager suManager;
    protected MessageExchangeListener jbiListener;

    protected XQuareBCListener xquareListener;
    protected XQuareSUHandler xquareSUHandler;
    

    public ObjectName getExtensionMBeanName() {
        return null;
    }

    public ComponentLifeCycle getLifeCycle() {
        return this;
    }

    public Document getServiceDescription(ServiceEndpoint arg0) {
        return suManager.getServiceDescription(arg0);
    }

    public ServiceUnitManager getServiceUnitManager() {
        return this.suManager;
    }

    public void init(ComponentContext context) throws JBIException {
        try {
            this.context = context;
            Logger log = context.getLogger("", null);
            logger = new ComponentLogger(log, log.getName(), log
                .getResourceBundleName(), context.getComponentName());

            this.channel = context.getDeliveryChannel(); //TODO use Internal if not out only
            this.serviceToPropertiesMap = new HashMap<String, Properties>();

        } catch (Exception e) {
            throw new JBIException(e);
        }
        logger.log(Level.INFO, "init");
    }

    public boolean isExchangeWithConsumerOkay(ServiceEndpoint arg0,
        MessageExchange arg1) {
        return true;
    }

    public boolean isExchangeWithProviderOkay(ServiceEndpoint arg0,
        MessageExchange arg1) {
        return true;
    }

    public ServiceEndpoint resolveEndpointReference(DocumentFragment arg0) {
        ServiceEndpoint result = null;
        if (arg0 != null) {
            String prefix = arg0.getPrefix();
            if (prefix == null) {
                prefix = new String();
            } else {
                prefix += ":";
            }
            Node service = XMLHelper.findChild(arg0, prefix + "service", true);

            if (service != null) {
                String serviceRequested = XMLHelper.getAttributeValue(service,
                "name");
                if (serviceRequested != null) {
                    result = suManager.findEndpointForService(QName
                        .valueOf(serviceRequested));
                }
            }
        }
        return result;
    }

    public void shutDown() throws JBIException {
        logger.log(Level.INFO, "shutDown");
    }

    public void start() throws JBIException {
        logger.log(Level.INFO, "start");

        XQuareBCJBIProcessor processor = new XQuareBCJBIProcessor(context, channel, logger,
            serviceToPropertiesMap);
        jbiListener = new MessageExchangeListener(channel, processor,
            MessageExchangeListener.IgnoredStatus.DONE_AND_ERROR_IGNORED, 0);
        jbiListener.listen(); // start listening for incoming JBI requests
        
        xquareListener = new XQuareBCListener(context, channel, logger,
            serviceToPropertiesMap, processor);
        new Thread(xquareListener).start(); // start "listening" for new data in the db
        
        xquareSUHandler = new XQuareSUHandler(context, logger,
            serviceToPropertiesMap);
        this.suManager = new PetalsServiceUnitManager(context, logger);
        suManager.addNewHandler(xquareSUHandler);
    }

    public void stop() throws JBIException {
        logger.log(Level.INFO, "stop");
        jbiListener.terminate();
        xquareListener.stopProcessing();
    }

    protected String getDirectoryRoot() {
        String baseDir = context.getInstallRoot();
        baseDir = baseDir.substring(0, baseDir.length() - 7);
        baseDir += "work" + File.separator;
        return baseDir;
    }

}
