/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id : $ 
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.xquarebc.listeners;

import java.io.ByteArrayOutputStream;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.NormalizedMessage;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.objectweb.petals.binding.xquarebc.XQuareBCException;
import org.objectweb.petals.binding.xquarebc.XQuareSUHandler;
import org.objectweb.petals.binding.xquarebc.query.StoredQuery;
import org.objectweb.petals.component.common.listener.AbstractInternalMEProcessor;
import org.objectweb.petals.component.common.util.SourceHelper;
import org.objectweb.petals.component.common.util.XMLHelper;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.xml.sax.SAXException;
import org.xquark.bridge.Mapper;
import org.xquark.bridge.Mapping;
import org.xquark.bridge.XQBridge;
import org.xquark.jdbc.datasource.JDBCDataSourceFactory;
import org.xquark.serialize.XMLSerializer;
import org.xquark.xml.xdbc.PreparedXMLStatement;
import org.xquark.xml.xdbc.XMLConnection;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLDocument;
import org.xquark.xml.xdbc.XMLResultSet;
import org.xquark.xml.xdbc.XMLStatement;

/**
 * 
 * This processor builds upon the XQuare Bridge library (see web site,
 * tutorial and reference documentation at http://xquare.objectweb.org/ )
 * in order to allow users to interact with databases through its three
 * operations : insert, query and storedQuery.
 * <p/>
 * 
 * <h4>Operations detail</h4>
 * The "insert" operation lets the user provide as the content of an
 * incoming JBI message an XML document defining insertion or update of
 * database records according to the configured XQuare mapping (mapping
 * XML file and insertion XML schema file). Note that updating existing
 * data or adding new records are both possible. Look at the XQuare
 * Bridge reference documentation for more details. The content of the
 * response is (if no fault because of an internal exception)
 * the single tag "<success/>".
 * <p/>
 * The "query" operation lets the user provide as the content of an
 * incoming JBI message an XQuery wrapped within the "query" top tag.
 * This XQuery must be XQuare compliant and consistent with the
 * configured database and mapping for the targeted service. It should
 * also be XML encoded (even if the parser tries to be nice by formatting
 * and including XML subnodes) - typically the results formatting should
 * be, since they comprise XML tags. The response is an XML document
 * whose top tag is "<Results>" and whose elements are specified by
 * the XQuery's results formatting part. Note that modules (akin to
 * XML / SQL views, defined in *.mod files provided in the service
 * unit's deployment directory) can be imported and used in such
 * queries. Look at the XQuare XQuery reference documentation for
 * more details.
 * There is a way to specify the top tag in the query : by writing
 * it down in xquery (ex. <query><result>{ for $i in...) and using a
 * service configured with the queryResults.wrapInRootTag at false.
 * 
 * <p/>
 * The "storedQuery" operation lets the user provide as the content of
 * an incoming JBI message the id of the stored parameterized query
 * to be executed and the value of its parameters. Concretely it must
 * consist in the "storedQuery" top tag with the "id" attribute and
 * zero or more subtags with "name" and "value" attributes. Thoses
 * "names" must be consistent with the parameters defined within the
 * XQuery code of the stored query (see examples), since the "parameter"
 * mechanism of XQuare / XQuery is used by the processor. Otherwise
 * it is similar to the "query" operation in its behaviour.
 * 
 * <h4>Configuration</h4>
 * The target database is configured through the *.properties file
 * provided by a given service unit and uses the following properties :
 * <ul>
 * <li>database.url : the JDBC URL of the targeted database. Required
 * for accessing the database through JDBC.
 * Ex. jdbc:mysql://localhost/petals_samples_xquare_auction</li>
 * <li>database.user : the user required by the JDBC connection. Required
 * for accessing the database through JDBC. Ex. petals_samples</li>
 * <li>database.password  : the password required by the JDBC connection.
 * Required for accessing the database through JDBC. Ex. petals_samples</li>
 * <li>jndidatasource.name : if provided, is used as the preferred way
 * to access the target database. Note that the given JNDI name must exist
 * in the Petals context and target a J2EE Datasource Ex. jdbc/myDS</li>
 * </ul>
 * <p/>
 * 
 * Additional properties for configuring XQuare Bridge :
 * <ul>
 * <li>database.mapping : the name of the XQuare mapping file to be found
 * in the service unit deployment directory Required. Ex. auction.map</li>
 * <li>queryResults.wrapInRootTag : whether to always wrap the query
 * results in a top "<Results>" tag. If set to false, single results
 * won't be wrapped.</li>
 * <li>xmlconnection.baseuri : The base URI where to find modules (.mod
 * files) to import if such are specified in user or stored queries.
 * Optional, defaults to the service unit's deployment directory</li>
 * <li>(not a property per se) The schema definition of insertion XML
 * documents must be provided in the service unit's deployment directory,
 * with the same prefix than the corresponding mapping file
 * (ex. auction.xsd)</li>
 * </ul>
 * <p/>
 * 
 * Configuring stored queries :<br/>
 * Stored queries must be defined as a property whose name is
 * "storedQuery_" followed by the query id (must be a String), and
 * whose value is its XQuery definition. Note that parameters are
 * implicitly defined using the corresponding XQuare / XQuery concept.
 * See examples.
 * 
 * @version $Rev: 250 $Date: {date}
 * @since Petals 1.0
 * @author Marc Dutoo - Open Wide
 * 
 */
public class XQuareBCJBIProcessor extends AbstractInternalMEProcessor {


    private static final String SQLXML_NAMESPACE_URI = "http://petals.objectweb.org/components/jdbc/sqlxml";
    private static final String SQLXML_RESULTS_TAG = "Results"; //TODO prop ?!
    private static final String INSERT_OPERATION = "insert";
    private static final String INSERT_SUCCESS_CONTENT = "<success/>";
    private static final String QUERY_OPERATION = "query";
    private static final String QUERY_TAG = "query";
    private static final String STORED_QUERY_OPERATION = "storedQuery";
    private static final String STORED_QUERY_TAG = "storedQuery";
    private static final String STORED_QUERY_ID_ATTR = "id";
    private static final String STORED_QUERY_PARAM_NAME_ATTR = "name";
    private static final String STORED_QUERY_PARAM_VALUE_ATTR = "value";
    public static final String STORED_QUERY_PROP_PREFIX = "storedQuery_";


    private Logger logger;
    private HashMap<String, Properties> serviceToPropertiesMap;

    /**
     * Creates a new XQuareBCJBIProcessor
     * @param context
     * @param channel
     * @param logger
     * @param serviceToPropertiesMap
     */
    public XQuareBCJBIProcessor(ComponentContext context, DeliveryChannel channel,
        Logger logger, HashMap<String, Properties> serviceToPropertiesMap) {
        super(channel, logger);
        this.logger = logger;
        this.serviceToPropertiesMap = serviceToPropertiesMap;
    }


    @Override
    protected void sendInMessage(QName service, QName operation, NormalizedMessage in) throws Exception {
        sendInMessage(null, service, operation, in);
    }

    @Override
    protected boolean sendInOutMessage(QName service, QName operation, NormalizedMessage in, NormalizedMessage out, boolean optionalOut) throws Exception {
        return sendInOutMessage(null, service, operation, in, out, optionalOut);
    }


    //@Override
    protected void sendInMessage(String address, QName service, QName operation, NormalizedMessage in) throws Exception {
        //TODO
    }

    /**
     */
    //@Override
    protected boolean sendInOutMessage(String address, QName service, QName operation,
        NormalizedMessage in, NormalizedMessage out, boolean optionalOut) throws Exception {
        // Get the IN message
        Source contentSource = in.getContent();

        // Loading the configuration of the specified service unit
        Properties props = getServiceProperties(service);

        // Loading the xquare bridge of the specified service unit
        XQBridge bridge = getXQBridge(props);

        if (INSERT_OPERATION.equals(operation.getLocalPart())) {
            String content = SourceHelper.createString(contentSource);
            // let's do the SQL INSERT as described in the message :
            doInsert(content, bridge, props);
            
            // Create out "success" message
            // NB. WARNING xquark fails silently ?! TODO check insert result in db ??
            out.setContent(SourceHelper.createSource(INSERT_SUCCESS_CONTENT));


        } else if (STORED_QUERY_OPERATION.equals(operation.getLocalPart())) {
            Node contentNode = getContentNode(contentSource, STORED_QUERY_TAG);
            // load from props known the stored queries and their query template
            StoredQuery storedQuery = getStoredQuery(contentNode, props);

            // NB. could optionally match against them (parameter number and type)
            // let's do the SQL SELECT using a stored query as described in the message :
            Source resXmlSource = doStoredQuery(storedQuery, bridge, props);

            // Create out message with the returned XML results
            // NB. WARNING xquark fails silently ?! TODO check insert result in db ??
            out.setContent(resXmlSource);


        } else if (QUERY_OPERATION.equals(operation.getLocalPart())) {
            String queryString = getContentString(contentSource, QUERY_TAG);
            // let's do the SQL SELECT query as described in the message :
            Source resXmlSource = doQuery(queryString, bridge, props);

            // Create out message with the returned XML results
            out.setContent(resXmlSource);

            
        } else {
            throw new XQuareBCException("Unknown operation " + operation);
        }

        return true;
    }


    /**
     * Returns the cached XQBridge for the given service unit properties
     * or creates a new one and caches it.
     * @param props service unit properties
     * @return a fully configured XQuare Bridge
     * @throws XQuareBCException if an error occurs
     */
    protected XQBridge getXQBridge(Properties props) throws XQuareBCException {
        XQBridge bridge = (XQBridge) props.get(XQuareSUHandler.XQUARE_BRIDGE);
        if (bridge != null) {
            return bridge;
        }

        DataSource xquareDataSource = getDataSource(props);
        try {
            // NB. could use a JNDI / J2EE datasource instead
            bridge = new XQBridge(xquareDataSource);
            // caching it locally in the service properties
            props.put(XQuareSUHandler.XQUARE_BRIDGE, bridge);
            return bridge;
            // NB. Here is what it would be in the explicit way :
            //bridge = new XQBridge(new SingleConnectionDataSource(dbConn));
        } catch (Exception e) {
            throw new XQuareBCException("Error creating the xquare xml / sql bridge", e);
        }
    }


    /**
     * Returns the cached DataSource for the given service unit properties.
     * If a JNDI datasource is configured in the properties, tries to use it
     * (without caching it). Else uses XQuare's JDBCDataSourceFactory
     * to create and configure a datasource according to the JDBC configuration
     * provided in the given service unit properties, before caching and returing it.
     * @param props service unit properties
     * @return a fully configured datasource
     * @throws XQuareBCException if an error occurs
     */
    protected DataSource getDataSource(Properties props) throws XQuareBCException {
        DataSource xquareDataSource = (DataSource) props.get(XQuareSUHandler.STANDALONE_DATASOURCE);
        if (xquareDataSource != null) {
            return xquareDataSource;
        }

        if (props.contains(XQuareSUHandler.JNDI_DATASOURCE)) {
            // trying to use the configured JNDI datasource
            String jndiDataSource = props.getProperty(XQuareSUHandler.JNDI_DATASOURCE);
            try {
                InitialContext jndiContext = new InitialContext(); // NB. no props because Petals-wide
                Object found = jndiContext.lookup(jndiDataSource);
                if (!(found instanceof DataSource)) {
                    throw new XQuareBCException("Unable to find JNDI datasource named " + jndiDataSource);
                }
                xquareDataSource = (DataSource) found;
                // NB. no caching
                return xquareDataSource;
            } catch (NamingException e) {
                throw new XQuareBCException("Error looking for JNDI datasource named "
                    + jndiDataSource, e);
            }
        }

        // Many ways to initiate XQuark's DB connection. We could have done it 
        // explictly in order to validate the connection conf.
        // NB. db drivers beyond xquark's default four (see XQuark JDBCDataSourceFactory) DON'T WORK !!!

        // creating a JDBC datasource with XQuare's helper
        String databaseUrl = props.getProperty(XQuareSUHandler.DATABASE_URL, "[nourl]");
        String databaseUser = props.getProperty(XQuareSUHandler.DATABASE_USER, "[nouser]");
        String databasePassword = props.getProperty(XQuareSUHandler.DATABASE_PASSWORD, "[nopwd]");
        try {
            xquareDataSource = JDBCDataSourceFactory.newInstance()
            .newDataSource(databaseUrl, databaseUser, databasePassword);
            props.put(XQuareSUHandler.STANDALONE_DATASOURCE, xquareDataSource);
            return xquareDataSource;
        } catch (SQLException e) {
            throw new XQuareBCException("Unable to create datasource for XQuare with parameters "
                + databaseUrl + " " + databaseUser + " " + databasePassword);
        }
    }


    /**
     * Implementation of the "insert" operation.
     * @param content must be valid XML describing the insertion or update of database
     * records according to the configured database and XQuare mapping.
     * @param bridge the service unit's bridge with which the inert will be done
     * @param serviceProps the targeted service unit's properties
     * @throws XQuareBCException
     */
    protected void doInsert(String content, XQBridge bridge,
        Properties serviceProps) throws XQuareBCException {
        String databaseMappingFilepath = serviceProps.getProperty(XQuareSUHandler.DATABASE_MAPPING, "[nomap]");

        // Getting the Mapping
        Mapping mapping = null;
        try {
            mapping = bridge.getMapping(databaseMappingFilepath);
        } catch (XMLDBCException e) {
            String msg = "XQuare : error when getting database "
                + "mapping from file " + databaseMappingFilepath;
            logger.log(Level.SEVERE, msg, e);
            if (e.getCause() != null) {
               logger.log(Level.SEVERE, "Caused by : ", e.getCause());
            } else {
                logger.log(Level.SEVERE, "No cause.");
            }
            throw new XQuareBCException(msg, e);
        }
        try {
            // Getting the Mapper and doing the actual insertion
            Mapper mapper = mapping.getMapper();
            // NB. XQuare does content validation against the insertion schema
            mapper.insertDocument(content);
        } catch (XMLDBCException e) {
            String msg = "XQuare XML DB error when trying to insert";
            logger.log(Level.SEVERE, msg, e);
            throw new XQuareBCException(msg, e);
        } catch (SAXException e) {
            String msg = "XQuare : error when inserting data "
                + "using mapping from file " + databaseMappingFilepath;
            logger.log(Level.SEVERE, msg, e);
            throw new XQuareBCException(msg, e);
        } finally {
            try {
                // properly closing the mapping and the resources it relies on
                mapping.close();
            } catch (XMLDBCException e) {
                logger.log(Level.WARNING, "Error closing XML mapping");
            }
        }
    }


    /**
     * Implementation of the "query" operation.
     * @param content must be an XQuery consistent with the configured database
     * and mapping of the specified service unit. It should be XML-encoded and
     * wrapped in a "query" top tag (even if the parser tries to be nice by
     * formatting and including XML subnodes).
     * @param bridge the service unit's bridge whose XML connection will be used
     * to do the query
     * @param serviceProps the targeted service unit's properties
     * @return a Source corresponding of the formatted results of the XQuery,
     * wrapped within the "Results" top tag.
     * @throws XQuareBCException
     */
    protected Source doQuery(String queryString, XQBridge bridge,
        Properties serviceProps) throws XQuareBCException {
        if (queryString == null) {
            throw new XQuareBCException("No query found in a 'query' operation !");
        }

        // Getting the XMl connection
        XMLConnection xc = null;
        try {
            xc = bridge.getXMLConnection();
        } catch (XMLDBCException e) {
            throw new XQuareBCException("XQuare XML DB error when getting XML connection", e);
        }
        try {
            // Creating the corresponding XML Statement
            XMLStatement xs = xc.createStatement();
            xs.setBaseURI(getXmlConnectionBaseUri(serviceProps));

            // executing the query
            XMLResultSet xrs = xs.executeQuery(queryString);
            // TODO line number limit, pagination

            // formating the results
            Source resXmlSource = convertToSource(xrs,
                getWrapQueryResultsInRootTag(serviceProps));

            return resXmlSource;
        } catch (XMLDBCException e) {
            throw new XQuareBCException("XQuare XML DB error when trying to query", e);
        } finally {
            try {
                // properly closing the XML connection and the resources it relies on
                xc.close();
            } catch (XMLDBCException e) {
                logger.log(Level.WARNING, "Error closing XML connection");
            }
        }
    }


    /**
     * TODO MAY GO IN COMMONS
     * Returns the given content as a String that must be wrapped under
     * a given top tag (checked).
     * Helper method to get everything there is under a given top tag
     * as a String. XML nodes will be formatted and included within.
     * NB. For XQuare BC, the point is to allow users to write results
     * formatting in XQuery code without having to XML-encode the tags.
     * @param contentSource the content of an incoming JBI message
     * @param contentTag the top tag of the contentSource
     * @return a String. NB. For XQuare BC, an XQuery.
     * @throws XQuareBCException if bad top content tag
     */
    protected String getContentString(Source contentSource, String contentTag) throws XQuareBCException {
        try {
            Node contentDocumentNode = XMLHelper.createDOMNodeFromSource(contentSource);
            Node contentNode = XMLHelper.getFirstChild(contentDocumentNode);
            // Checking that node element tag is the required one
            if (!contentTag.equals(contentNode.getLocalName())) {
                throw new XQuareBCException("XML XQuare queries should be "
                    + "encapsulated in a top <\"" + contentTag + "\"> tag !");
            }
            // Now reading the XQuery code in the most "nice" way
            NodeList queryNodes = contentNode.getChildNodes();
            int queryNodeNb = queryNodes.getLength();
            if (queryNodeNb == 1) {
                Node singleQueryNode = queryNodes.item(0);
                if (singleQueryNode instanceof Text) {
                    return ((Text) singleQueryNode).getTextContent();
                }
            }
            StringBuffer contentBuf = new StringBuffer();
            for (int i = 0; i < queryNodeNb ; i++) {
                Node queryNode = queryNodes.item(i);
                // getting the full XML of the node in order to help the user pass
                // along Results XML formatting
                contentBuf.append(XMLHelper.createStringFromDOMNode(queryNode));
                contentBuf.append("\n");
            }
            return contentBuf.toString();
        } catch (XQuareBCException e) {
            throw e;
        } catch (Exception e) {
            String contentString;
            try {
                contentString = SourceHelper.createString(contentSource);
            } catch (Exception e1) {
                // do nothing
                contentString = "[[error in transforming Source to String]]";
            }
            throw new XQuareBCException("Error while building content String "
                + "from source : " + contentString, e);
        }
    }


    /**
     * Implementation of the "storedQuery" operation.
     * @param storedQuery the stored query instance to execute. Its id and parameters
     * name and values should be consistent with the specified service unit's
     * configuration.
     * @param bridge the service unit's bridge whose XML connection will be used
     * to do the query
     * @param serviceProps the targeted service unit's properties
     * @return a Source corresponding of the formatted results of the stored XQuery,
     * wrapped within the "Results" top tag.
     * @throws XQuareBCException
     */
    protected Source doStoredQuery(StoredQuery storedQuery, XQBridge bridge, Properties serviceProps)
          throws XQuareBCException {
        XMLConnection xc = null;
        try {
            xc = bridge.getXMLConnection();
        } catch (XMLDBCException e) {
            throw new XQuareBCException("XQuare XML DB error when getting XML connection", e);
        }
        try {
            xc.setBaseURI(getXmlConnectionBaseUri(serviceProps));
            // loading the corresponding xquare prepared statement
            PreparedXMLStatement pxs = xc.prepareStatement(storedQuery.getStatement());

            // filling the query template with parameters
            for (String key : storedQuery.getParameterMap().keySet()) {
                String value = storedQuery.getParameterMap().get(key);
                pxs.setString(key, value);
            }

            // executing the query
            XMLResultSet xrs = pxs.executeQuery();
            // NB. here could be done line number limit, pagination...

            // formating the results
            Source resXmlSource = convertToSource(xrs,
                getWrapQueryResultsInRootTag(serviceProps));

            return resXmlSource;
        } catch (Exception e) {
            throw new XQuareBCException("Error while doing stored XML query", e);
        } finally {
            try {
                xc.close();
            } catch (XMLDBCException e) {
                logger.log(Level.WARNING, "Error closing XML connection");
            }
        }
    }

    /**
     * TODO MAY GO IN COMMONS
     * Returns the given content as a DOM node that must be wrapped under
     * a given top tag (checked).
     * @param contentSource the content of an incoming JBI message
     * @param contentTag the top tag of the contentSource
     * @return the corresponding DOM node tree
     * @throws XQuareBCException if bad top tag
     */
    protected Node getContentNode(Source contentSource, String contentTag) throws XQuareBCException {
        try {
            Node contentDocumentNode = XMLHelper.createDOMNodeFromSource(contentSource);
            Node contentNode = XMLHelper.getFirstChild(contentDocumentNode);
            // Checking that node element tag is the required one
            if (!contentTag.equals(contentNode.getLocalName())) {
                throw new XQuareBCException("XML XQuare queries should be "
                    + "encapsulated in a top <\"" + contentTag + "\"> tag !");
            }
            return contentNode;
        } catch (Exception e) {
            throw new XQuareBCException("XML XQuare queries should be "
                + "encapsulated in a top \"" + contentTag + "\"> tag !"
                + e.getMessage(), e);
        }
    }


    /**
     * Converts the given XQuery results to XML and wrap them in a top "Results" tag.
     * NB. It generates the XML with the namespace defined in the SQLXML_NAMESPACE_URI
     * constant.
     * @param xrs results from an XQuery (or a stored one)
     * @return the corresponding, formatted XML Source
     * @throws XQuareBCException 
     * @throws XMLDBCException if any error
     */
    protected Source convertToSource(XMLResultSet xrs, boolean wrapInRootTag)
            throws XQuareBCException {
        String resXmlString;
        try {
            if (wrapInRootTag) { // handling "always wrap" policy
                XMLDocument resXmlDoc = xrs.getFragmentsAsDocument(SQLXML_NAMESPACE_URI,
                    SQLXML_RESULTS_TAG, "sqlxml:" + SQLXML_RESULTS_TAG);
                // NB. getAsSAX is the root output method, but we can't do a thing with it ?!
                resXmlString = resXmlDoc.getAsString();
                // NB. no "<?xml ..." processing instruction with the xquark serializer
                // removing forbidden xml processing instruction :
                resXmlString = resXmlString.trim();
                if (resXmlString.startsWith("<?xml ")) {
                    resXmlString = resXmlString.substring(resXmlString.indexOf('>') + 1);
                }
            
            } else {     // handling "try not to wrap" policy        
                ByteArrayOutputStream resBos = new ByteArrayOutputStream();
                int rowNb = 0;
                XMLSerializer serializer = new XMLSerializer(resBos);
                xrs.setContentHandler(serializer);
                for (; xrs.hasNext(); rowNb++) {
                    xrs.nextAsSAX();
                }
                resXmlString = resBos.toString();
                // NB. no "<?xml ..." processing instruction with the xquark serializer
                if (rowNb != 1) {
                    // needs a root tag to be a valid XML document
                    resXmlString = "<sqlxml:" + SQLXML_RESULTS_TAG + " xmlns=\""
                        + SQLXML_NAMESPACE_URI + "\">\n" + resXmlString + "\n</sqlxml:"
                        + SQLXML_RESULTS_TAG + ">";
                }
            }
        } catch (Exception e) {
            throw new XQuareBCException("Error while serializing "
                + "XML result set", e);
        }
        
        // now creating corresponding source
        // NB. not adding encapsulating <content> tag, because there is SQLXML_RESULTS_TAG
        Source resXmlSource = SourceHelper.createSource(resXmlString);
        logger.log(Level.FINE, "Query result : " + resXmlString);
        return resXmlSource;
    }


    /**
     * Builds the stored query model.
     * @param contentNode the node to build it from
     * @param props the properties to check the provided query id against
     * @return a filled StoredQuery model
     * @throws XQuareBCException if any error
     */
    protected StoredQuery getStoredQuery(Node contentNode,
        Properties props) throws XQuareBCException {
        StoredQuery storedQuery = new StoredQuery();

        // get query id
        String storedQueryId = XMLHelper.getAttributeValue(contentNode,
            STORED_QUERY_ID_ATTR);
        if (storedQueryId == null || storedQueryId.length() == 0) {
            throw new XQuareBCException("Found stored query without its id !");
        }
        storedQuery.setId(storedQueryId);

        // get query statements
        String storedQueryPropName = STORED_QUERY_PROP_PREFIX + storedQueryId;
        String storedQueryStatement = props.getProperty(storedQueryPropName);
        if (storedQueryStatement == null || storedQueryStatement.length() == 0) {
            throw new XQuareBCException("Stored query with id " + storedQueryId
                + " has no query statement !");
        }
        storedQuery.setStatement(storedQueryStatement);

        // get params
        HashMap<String, String> storedQueryParamMap = new HashMap<String, String>();
        Node currentParamNode = XMLHelper.getFirstChild(contentNode);
        while (currentParamNode != null) {
            String paramName = XMLHelper.getAttributeValue(currentParamNode,
                STORED_QUERY_PARAM_NAME_ATTR);
            if (paramName == null || paramName.length() == 0) {
                throw new XQuareBCException("Found stored query parameter without a name "
                    + "for stored query with id " + storedQueryId);
            }
            String paramValue = XMLHelper.getAttributeValue(currentParamNode,
                STORED_QUERY_PARAM_VALUE_ATTR);
            storedQueryParamMap.put(paramName, paramValue);
            currentParamNode = XMLHelper.getNextSibling(currentParamNode);
        }
        storedQuery.setParameterMap(storedQueryParamMap);

        return storedQuery;
    }



    /**
     * Returns the value of the XQuareSUHandler.XML_CONNECTION_BASE_URI property
     * within the specified service unit
     * @param props the targeted service unit's configuration
     * @return
     */
    protected String getXmlConnectionBaseUri(Properties props) {
        return props.getProperty(XQuareSUHandler.XML_CONNECTION_BASE_URI);
    }
    
    /**
     * Returns the value of the XQuareSUHandler.WRAP_QUERY_RESULTS_IN_ROOT_TAG
     * property within the specified service unit
     * @param props the targeted service unit's configuration
     * @return
     */
    protected boolean getWrapQueryResultsInRootTag(Properties props) {
        return Boolean.valueOf(props.getProperty(XQuareSUHandler
            .WRAP_QUERY_RESULTS_IN_ROOT_TAG, "false")).booleanValue();
    }

    
    /**
     * Returns the targeted service unit's properties.
     * @param service the targeted service
     * @return its configuration
     * @throws XQuareBCException
     */
    protected Properties getServiceProperties(QName service) throws XQuareBCException {
        //TODO rather use endpoint address (by getting the WSDL and from there the service name) ??
        String serviceName = service.getLocalPart();
        Properties csvProperties = this.serviceToPropertiesMap.get(serviceName);
        if (csvProperties != null) {
            return csvProperties;
        } else {
            throw new XQuareBCException("XQuare properties not found for service "
                + serviceName);
        }
    }
    
}
