package org.objectweb.rentacar.agency.services.availability;

import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import javax.jws.WebMethod;
import javax.jws.WebParam;
import javax.jws.WebService;
import javax.jws.soap.SOAPBinding;

import org.apache.log4j.Logger;
import org.objectweb.rentacar.AgencyException;
import org.objectweb.rentacar.persistance.bo.Car;
import org.objectweb.rentacar.persistance.bo.CarCriteria;
import org.objectweb.rentacar.persistance.bo.CarDeLuxeVO;
import org.objectweb.rentacar.persistance.bo.CarVO;
import org.objectweb.rentacar.persistance.dao.CarDAO;
import org.objectweb.rentacar.persistance.dao.DAOException;

/**
 * This class expose web services that allow to check car availability in an agency.
 * @author wjoseph
 *
 */
@WebService
@SOAPBinding(style=SOAPBinding.Style.DOCUMENT, use=SOAPBinding.Use.LITERAL)
public class AgencyCheckAvailabilityServiceImpl{
	
	
	private static Logger logger = Logger.getLogger( AgencyCheckAvailabilityServiceImpl.class);
	
	/**
	 * Retrieve the list of cars that match the given criterium 
	 * and that are available during the given period
	 * @param criterium the car criteria for the research
	 * @param startDate the starting date
	 * @param endDate the ending date
	 * @return
	 * @throws AgencyException 
	 */
	@WebMethod
	public CarVO[] consultAvailability(@WebParam(name = "criterium")CarCriteria criterium, 
								@WebParam(name = "startDate")Date startDate, 
								@WebParam(name = "endDate")Date endDate) throws AgencyException {
		logger.info("Agency . consultAvailability : "+criterium.toString()+
				", "+startDate.toString()+", "+endDate.toString());
		Set<CarVO> availableCarInAgency;
		availableCarInAgency = availableCarsByCriterium(criterium, startDate, endDate);
		CarVO[] availableCar= {};
		return availableCarInAgency.toArray(availableCar);
	}
	
	/**
	 * It is a test method to validate polymorphic strategy with jaxws
	 */
	@WebMethod
	public CarVO[] getAListOfPolymorphicCars() throws AgencyException {
		CarDeLuxeVO carDeLuxeVO = new CarDeLuxeVO("G", "mercedes-benz", "TwinportV12", "f45", "sw45", "smb45", "tv45");
		CarVO carVO = new CarVO("G", "peugeot", "206");
		CarVO[] cars = new CarVO[2];
		cars[0] = carDeLuxeVO;
		cars[1] = carVO;
		return cars;
	}
	
	/**
	 * It is a test method to validate polymorphic strategy with jaxws
	 */
	@WebMethod
	public void dummy(CarDeLuxeVO carDeLuxeVO) {
		
	}
	
	/**
	 * Retrieve the list of cars that match the given criterium 
	 * and that are available during the given period
	 * @param criterium the car criteria for the research
	 * @param start the starting date
	 * @param end the ending date
	 * @return
	 * @throws AgencyException 
	 */
	private Set<CarVO> availableCarsByCriterium(CarCriteria criterium, Date start, Date end) throws AgencyException {
			Set<CarVO> availableCars= new HashSet<CarVO>();
			boolean type=false, model= false, brand=false, carId=false;
			
			if( criterium.getType()== null)
				type=true;
			
			if(criterium.getModel()== null )
				model=true;
			
			if(criterium.getBrand()==null)
				brand=true;
			
			if(criterium.getCarId()==null)
				carId=true;

			try {
					Set<Car> cars= getAvailableCars(start, end);
					Set<CarVO> carsVO= new HashSet<CarVO>();
						
					for (Car car : cars) {
						carsVO.add(new CarVO(car));
					}
					
					for (CarVO carVO : carsVO) {
						//logger.debug(carVO);
						if(carId || carVO.getCarId().equals(criterium.getCarId()))
							if(type || carVO.getType().equals(criterium.getType()))
									if(model || carVO.getModel().equals(criterium.getModel()))
											if(brand || carVO.getBrand().equals(criterium.getBrand()))
														availableCars.add(carVO);
				}
				
			} catch (DAOException e) {
				logger.error("Error when retrieving available cars", e);
				throw new AgencyException("Error when retrieving available cars", e);
			}
			
			return availableCars;
		}
	
	/**
	 * Returns all available cars in this agency for the given period
	 * @param start Starting date of the search
	 * @param end Ending date of the search
	 * @return A Set of avalaible cars in the agency for this period
	 * @throws DAOException
	 */
	private Set<Car> getAvailableCars(Date start, Date end) throws DAOException{
		Set<Car> result = getAllAgencyCars();
		result.removeAll(getReservedCars(start,end));
		return result;
	}
	
	/**
	 * Returns all cars of this agency
	 * @return A Set of cars belonging to this agency
	 * @throws DAOException
	 */
	private Set<Car> getAllAgencyCars() throws DAOException{
		return CarDAO.getInstance().getAllAgencyCars();
	}
	
	/**
	 * Returns all reserved cars for the givien period
	 * @param start Starting date of the search
	 * @param end Ending date of the search
	 * @return A Set of reserved cars for this period
	 * @throws DAOException
	 */
	private Set<Car> getReservedCars(Date start, Date end) throws DAOException{
		return (Set<Car>) CarDAO.getInstance().getReservedCarsByDates(start,end);
	}
	
}
