/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: ProfilerImpl.java 154 6 juin 2006 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.demo.mortgage.profiler;

import javax.jws.WebMethod;
import javax.jws.WebParam;
import javax.jws.WebService;

/**
 * A mortgage webservice used to evaluate the profile of a customer
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
@WebService(targetNamespace = "http://petals.objectweb.org/")
public class ProfilerImpl {

    /**
     * Evaluate the profile of the mortgage requester from his cash flow
     * information. This is a detailed evaluation : the response depends on the
     * debt level
     * 
     * @param salary
     *            The salary of the customer
     * @param propertyTaxes
     *            Property Taxes of the customer
     * @param insurance
     *            Amount of insurance of the customer
     * @param autoPayment
     *            All the auto payment of the customer
     * @param creditCards
     *            Amount of the credit cards of the customer
     * @param otherPayments
     *            Other payments (credits, subscriptions...) of the customer
     * @return good for a low debt level (debt rate<33%), medium for an average
     *         debt level (33%<debt rate<50%) and risky for a high debt
     *         level(debt rate>50%). Must be non null and non empty.
     */
    @WebMethod
    public String evaluateDetailedProfile(@WebParam(name = "salary")
    float salary, @WebParam(name = "propertyTaxes")
    float propertyTaxes, @WebParam(name = "insurance")
    float insurance, @WebParam(name = "autoPayment")
    float autoPayment, @WebParam(name = "creditCards")
    float creditCards, @WebParam(name = "otherPayments")
    float otherPayments) {
        System.out.println("ProfilerImpl : evaluateDetailedProfile - " + salary
                + " " + propertyTaxes + " " + insurance + " " + autoPayment
                + " " + creditCards + " " + otherPayments);
        String result = "good";
        float debtRate = calculateDebtRate(salary, propertyTaxes, insurance,
                autoPayment, creditCards, otherPayments);
        /*
         * If debt rate is upper than 50 percent, mortgage profile is risky else
         * if debt rate is upper than 33 percent, mortgage profile is just
         * medium
         */
        if (debtRate > 50) {
            result = "risky";
        } else if (debtRate > 33) {
            result = "medium";
        }
        System.out.println("ProfilerImpl : evaluateDetailedProfile - debRate "
                + debtRate + " Evaluation " + result);
        return result;
    }

    /**
     * Evaluate the profile of the mortgage requester from his cash flow
     * information. This is a simple evaluation : ok or not to according
     * mortgage
     * 
     * @param salary
     *            The salary of the customer
     * @param propertyTaxes
     *            Property Taxes of the customer
     * @param insurance
     *            Amount of insurance of the customer
     * @param autoPayment
     *            All the auto payment of the customer
     * @param creditCards
     *            Amount of the credit cards of the customer
     * @param otherPayments
     *            Other payments (credits, subscriptions...) of the customer
     * @return yes if the mortgage is allowed (debt rate<50%) or no instead
     */
    @WebMethod
    public String evaluateSimpleProfile(@WebParam(name = "salary")
    float salary, @WebParam(name = "propertyTaxes")
    float propertyTaxes, @WebParam(name = "insurance")
    float insurance, @WebParam(name = "autoPayment")
    float autoPayment, @WebParam(name = "creditCards")
    float creditCards, @WebParam(name = "otherPayments")
    float otherPayments) {
        System.out.println("ProfilerImpl : evaluateSimpleProfile - " + salary
                + " " + propertyTaxes + " " + insurance + " " + autoPayment
                + " " + creditCards + " " + otherPayments);
        String result = "yes";

        float debtRate = calculateDebtRate(salary, propertyTaxes, insurance,
                autoPayment, creditCards, otherPayments);
        /*
         * If debt rate is upper than 50 percent, mortgage isn't allowed
         */
        if (debtRate > 50) {
            result = "no";
        }
        System.out.println("ProfilerImpl : evaluateSimpleProfile - debRate "
                + debtRate + " Mortgage possible " + result);
        return result;
    }

    /**
     * Calculate the debt rate using salary and all outgoings. Outgoings :
     * propertyTaxes + insurance + autoPayment + creditCards + otherPayments.
     * debtRate : (100 * outgoings) / salary.
     * 
     * @param salary
     *            The salary of the customer
     * @param propertyTaxes
     *            Property Taxes of the customer
     * @param insurance
     *            Amount of insurance of the customer
     * @param autoPayment
     *            All the auto payment of the customer
     * @param creditCards
     *            Amount of the credit cards of the customer
     * @param otherPayments
     *            Other payments (credits, subscriptions...) of the customer
     * @return the debt rate in percent
     */
    private float calculateDebtRate(float salary, float propertyTaxes,
            float insurance, float autoPayment, float creditCards,
            float otherPayments) {
        /*
         * Calculate sum of the outgoings
         */
        float outgoings = propertyTaxes + insurance + autoPayment + creditCards
                + otherPayments;
        /*
         * Calculate debt rate in percent
         */
        float debtRate = (100 * outgoings) / salary;

        return debtRate;
    }
}
