/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: DetailedMortgageProfiling.java 154 7 juin 2006 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.demo.mortgage.workflow.orchestration;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.objectweb.petals.component.common.HandlingException;
import org.objectweb.petals.component.common.PEtALSComponentSDKException;
import org.objectweb.petals.component.common.util.SourceHelper;

/**
 * Workflow for detailed mortgage
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
public class DetailedMortgageProfiling implements MortgageProfiling {

	public final static String MORTGAGE_SERVICE = "ProfilerImplService";

	public final static String MORTGAGE_SERVICE_OP = "evaluateDetailedProfile";

	protected DeliveryChannel channel;

	protected Logger l;

	protected MortgageProfilingUtil profilingUtil;

	public DetailedMortgageProfiling() {
		super();
	}

	/**
	 * Build a new DetailedMortgageProfiling setted with the given
	 * ComponentContext and Logger. Set a new MortgageProfilingUtil.
	 * 
	 * @param ctx
	 *            the ComponentContext of the MortgageWorkflow service
	 * @param l
	 *            a logger
	 */
	public DetailedMortgageProfiling(DeliveryChannel channel, Logger l) {
		super();
		this.channel = channel;
		this.l = l;
		profilingUtil = new MortgageProfilingUtil(channel, l);

	}

	/**
	 * Launch the detailed mortgage workflow : evaluate the profile, evaluate
	 * the mortgage rate and create the MortgageResponse bo
	 * 
	 * @param request
	 *            the MortgageRequest bo. Must be non null.
	 * @return a MortgageResponse bo. Can't be null
	 */
	public MortgageResponse orchestrate(MortgageRequest request)
			throws HandlingException {
		MortgageResponse response = null;

		if (request == null) {
			throw new HandlingException("MortgageRequest must not be null !");
		}

		l.info("###################################################");
		l.info("#### PROCESSING A DETAILED MORTGAGE RATE EVALUATION");
		l.info("###################################################");

		/*
		 * Evaluate Mortgage Profile
		 */
		l.info("##### First step : Evalute Mortgage Profile #####");
		String profile = doEvalProfile(request.getSalary(), request
				.getPropertyTaxes(), request.getInsurance(), request
				.getAutoPayment(), request.getCreditCards(), request
				.getOtherPayments());

		/*
		 * Evaluate Mortgage rate
		 */
		l.info("##### Second Step : Evaluate Mortgage Rate #####");
		if (profilingUtil == null) {
			profilingUtil = new MortgageProfilingUtil(channel, l);
		}
		float rate = profilingUtil.doEvalMortgageRate(profile);

		/*
		 * Create Mortgage Response
		 */
		response = profilingUtil.createMortgageResponse(rate, request);
		l
				.info("##### Last Step : Send the response to the Mortgage requester#####");

		/*
		 * Send a summary of the profiling
		 */
		profilingUtil.doSendSummary(response);

		return response;
	}

	/**
	 * Create a new Source content that handles a Soap representation of the
	 * Profiler service's evaluation method call from given customer personal
	 * information
	 * 
	 * @param salary
	 *            The salary of the customer
	 * @param propertyTaxes
	 *            Property Taxes of the customer
	 * @param insurance
	 *            Amount of insurance of the customer
	 * @param autoPayment
	 *            All the auto payment of the customer
	 * @param creditCards
	 *            Amount of the credit cards of the customer
	 * @param otherPayments
	 *            Other payments (credits, subscriptions...) of the customer
	 * @return a new Source content that handles a Soap representation of the
	 *         Profiler service's evaluation method call
	 * @throws PEtALSComponentSDKException
	 */
	protected Source createInMessageContent(float salary, float propertyTaxes,
			float insurance, float autoPayment, float creditCards,
			float otherPayments) throws PEtALSComponentSDKException {

		StringBuffer soap = new StringBuffer();
		soap.append("<ns1:" + MORTGAGE_SERVICE_OP
				+ " xmlns:ns1=\"http://petals.objectweb.org/\">");
		soap.append("<salary>" + salary + "</salary>");
		soap.append("<propertyTaxes>" + propertyTaxes + "</propertyTaxes>");
		soap.append("<insurance>" + insurance + "</insurance>");
		soap.append("<autoPayment>" + autoPayment + "</autoPayment>");
		soap.append("<creditCards>" + creditCards + "</creditCards>");
		soap.append("<otherPayments>" + otherPayments + "</otherPayments>");
		soap.append("</ns1:" + MORTGAGE_SERVICE_OP + ">");

		return SourceHelper.createSource(soap.toString());
	}

	/**
	 * Launch the mortgage profile evaluation : create a Soap content with given
	 * data, create and send a new MessageExchange handling this content to the
	 * Profiler service and finaly process the response.
	 * 
	 * @param salary
	 * @param propertyTaxes
	 * @param insurance
	 * @param autoPayment
	 * @param creditCards
	 * @param otherPayments
	 * @return
	 * @throws Exception
	 */
	protected String doEvalProfile(float salary, float propertyTaxes,
			float insurance, float autoPayment, float creditCards,
			float otherPayments) throws HandlingException {

		MessageExchange msg = null;
		try {
			/*
			 * Create message content from given data
			 */
			Source source = createInMessageContent(salary, propertyTaxes,
					insurance, autoPayment, creditCards, otherPayments);

			/*
			 * Create a new message exchange for the Profiler Service
			 */
			msg = channel.createExchangeFactory().createInOutExchange();
			msg.setService(new QName(MortgageProfilingUtil.DEFAULT_NS,
					MORTGAGE_SERVICE));
			msg.setOperation(new QName(MORTGAGE_SERVICE_OP));

			/*
			 * Create the inMessage to send
			 */
			NormalizedMessage inNM = msg.createMessage();
			inNM.setContent(source);
			((InOut) msg).setInMessage(inNM);

			/*
			 * Send the message exchange
			 */
			channel.sendSync(msg);
		} catch (Exception e) {
			throw new HandlingException("Error evaluating profile", e);
		}
		/*
		 * Get response from EvalRate service
		 */
		NormalizedMessage outNM = ((InOut) msg).getOutMessage();
		String response = "ProfileNotDefined";
		if (outNM != null) {
			try {
				response = SourceHelper.createString(outNM.getContent());
			} catch (PEtALSComponentSDKException e) {
				throw new HandlingException(e);
			}
			if (response.indexOf("return") > -1) {
				response = response.substring(response.indexOf("<return>") + 8,
						response.indexOf("</return>"));
				l.log(Level.INFO, response);
			}
		} else {
			throw new HandlingException(
					"Mortgage profiler web service response is null");
		}
		l.info("##### Evaluated profile : " + response + " #####");
		return response;
	}

}
