package org.objectweb.petals.demo.mortgage.workflow.sdk;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.List;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.objectweb.petals.component.common.HandlingException;
import org.objectweb.petals.demo.mortgage.workflow.orchestration.DetailedMortgageProfiling;
import org.objectweb.petals.demo.mortgage.workflow.orchestration.MortgageProfiling;
import org.objectweb.petals.demo.mortgage.workflow.orchestration.MortgageRequest;
import org.objectweb.petals.demo.mortgage.workflow.orchestration.MortgageResponse;
import org.objectweb.petals.demo.mortgage.workflow.orchestration.SimpleMortgageProfiling;
import org.objectweb.petals.tools.jbicommon.util.XMLUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

public class MortgageWorkflowImpl implements MortgageWorkflow {

	private Logger logger;

	private DeliveryChannel channel;

	public MortgageWorkflowImpl(Logger logger, DeliveryChannel channel) {
		super();
		this.logger = logger;
		this.channel = channel;
	}

	public String orchestrate(String endpointName, String content)
			throws HandlingException {

		/*
		 * retrieve request parameters
		 */
		MortgageRequest request = retrieveRequestParameters(content);

		/*
		 * execute workflow to process
		 */
		MortgageResponse response = executeWorkflow(endpointName, request);

		return createContentFromMortgageResponse(response);
	}

	/**
	 * Build a MortgageRequest bo from information contained into the String
	 * content (a Soap representation of the MortgageRequest).
	 * 
	 * @param content
	 *            the String content (a Soap representation of the
	 *            MortgageRequest)
	 * @return a MortgageRequest bo. Can't be null.
	 * @throws HandlingException
	 * @throws Exception
	 */
	protected MortgageRequest retrieveRequestParameters(String stringContent)
			throws HandlingException {

		String address = "";
		String autoPayment = "";
		String creditCards = "";
		String firstName = "";
		String insurance = "";
		String lastName = "";
		String otherPayments = "";
		String propertyTaxes = "";
		String salary = "";
		
		System.out.println(stringContent);

		Document doc = parseXml(stringContent);
		Node requestValues = doc.getDocumentElement().getFirstChild();

		List<Node> values = XMLUtil.getNodeChildren(requestValues);
		for (Node node : values) {
			if ("address".equalsIgnoreCase(node.getLocalName())) {
				address = node.getTextContent();
			} else if ("autoPayment".equalsIgnoreCase(node.getLocalName())) {
				autoPayment = node.getTextContent();
			} else if ("creditCards".equalsIgnoreCase(node.getLocalName())) {
				creditCards = node.getTextContent();
			} else if ("firstName".equalsIgnoreCase(node.getLocalName())) {
				firstName = node.getTextContent();
			} else if ("insurance".equalsIgnoreCase(node.getLocalName())) {
				insurance = node.getTextContent();
			} else if ("lastName".equalsIgnoreCase(node.getLocalName())) {
				lastName = node.getTextContent();
			} else if ("otherPayments".equalsIgnoreCase(node.getLocalName())) {
				otherPayments = node.getTextContent();
			} else if ("propertyTaxes".equalsIgnoreCase(node.getLocalName())) {
				propertyTaxes = node.getTextContent();
			} else if ("salary".equalsIgnoreCase(node.getLocalName())) {
				salary = node.getTextContent();
			}
		}

		MortgageRequest request = new MortgageRequest(firstName, lastName,
				address, new Float(salary).floatValue(), new Float(
						propertyTaxes).floatValue(), new Float(insurance)
						.floatValue(), new Float(autoPayment).floatValue(),
				new Float(creditCards).floatValue(), new Float(otherPayments)
						.floatValue());

		return request;
	}

	private Document parseXml(final String xmlString) throws HandlingException {

		DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory
				.newInstance();
		docBuilderFactory.setNamespaceAware(true);

		DocumentBuilder docBuilder = null;

		try {
			docBuilder = docBuilderFactory.newDocumentBuilder();
		} catch (ParserConfigurationException pce) {
			throw new HandlingException("Bad XML parser configuration", pce);
		}

		Document document = null;

		try {
			document = docBuilder.parse(new ByteArrayInputStream(xmlString
					.getBytes()));
		} catch (SAXException saxe) {
			throw new HandlingException(
					"JBI descriptor is not well formed XML", saxe);
		} catch (IOException ioe) {
			throw new HandlingException("Can't read JBI descriptor", ioe);
		}
		return document;
	}

	/**
	 * Execute the workflow depending on the endpoint information : detailed
	 * workflow for "detailedMortgageProfiler" endpoint and simple workflow for
	 * "simpleMortgageProfiler" endpoint.
	 * 
	 * @param endpointName
	 *            name of the targeted endpoint
	 * @param request
	 *            MortgageRequest bo
	 * @return MortgageResponse bo. Can't be null.
	 * @throws Exception
	 *             if given endpoint doesn't exist for MortgageWorkflow service
	 */
	private MortgageResponse executeWorkflow(String endpointName,
			MortgageRequest request) throws HandlingException {
		MortgageResponse response = null;

		MortgageProfiling profiling = null;
		if (MortgageWorkflowEngine.ENDPOINT1.equals(endpointName)) {
			profiling = new SimpleMortgageProfiling(channel, logger);
		} else if (MortgageWorkflowEngine.ENDPOINT2.equals(endpointName)) {
			profiling = new DetailedMortgageProfiling(channel, logger);
		} else {
			throw new HandlingException(
					"Given endpoint doesn't exist for MortgageWorkflow");
		}
		response = profiling.orchestrate(request);

		return response;
	}

	/**
	 * Create a String content from the MortgageResponse. This String handles a
	 * Soap representation of the MortgageResponse bo.
	 * 
	 * @param response
	 *            the MortgageResponse bo.
	 * @return the String handling the Soap representation of the
	 *         MortgageResponse. Can't be null.
	 */
	protected String createContentFromMortgageResponse(MortgageResponse response) {

		StringBuffer soap = new StringBuffer();
		soap
				.append("<orchestrateResponse xmlns:soapenc=\"http://schemas.xmlsoap.org/soap/encoding/\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\">");
		soap
				.append("<orchestrateReturn xmlns:tns0=\"http://orchestration.workflow.mortgage.demo.petals.objectweb.org\" xsi:type=\"tns0:MortgageResponse\">");
		soap.append("<firstName xsi:type=\"xsd:string\">"
				+ response.getFirstName() + "</firstName>");
		soap.append("<lastName xsi:type=\"xsd:string\">"
				+ response.getLastName() + "</lastName>");
		soap.append("<mortgageRate xsi:type=\"xsd:float\">"
				+ response.getMortgageRate() + "</mortgageRate>");
		soap.append("</orchestrateReturn>");
		soap.append("</orchestrateResponse>");

		return soap.toString();
	}

}
