/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: JNDIUtil.java 16:03:11 ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.util;

import java.util.Map;
import java.util.Properties;
import java.util.Map.Entry;

import javax.naming.Binding;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.LinkRef;
import javax.naming.Name;
import javax.naming.NameClassPair;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;

/**
 * This utility class is used to help us to manage JNDI directories
 * 
 * @author ddesjardins - eBMWebsourcing
 */
public final class JNDIUtil {

    private JNDIUtil() {

    }

    /**
     * Browse the distributed JNDI directory
     * 
     * Connect to the JNDI directory of this container if no args are specified
     * 
     * Else can connect to any JNDI directory if -host and -port are specified
     * 
     * @param args
     *            argument of the program
     * @return a String representing the JNDI tree
     * 
     * @throws NamingException
     */
    public static String browseJNDI(String[] args) throws NamingException {
        String host = SystemUtil.getHost();
        String port = SystemUtil.getJndiPort();
        String factory = SystemUtil.getJndiFactory();

        if (args != null) {
            int i = 0;
            for (; i < args.length; i++) {
                if ("-host".equals(args[i])) {
                    i++;
                    host = args[i];
                } else if ("-port".equals(args[i])) {
                    i++;
                    port = args[i];
                } else if ("-factory".equals(args[i])) {
                    i++;
                    factory = args[i];
                }
            }
        }
        Properties properties = new Properties();
        properties.setProperty("java.naming.factory.host", host);
        properties.setProperty("java.naming.factory.port", port);
        properties.setProperty("java.naming.factory.initial", factory);
        InitialContext rootContext;

        StringBuffer result = new StringBuffer();

        rootContext = retrieveContext(properties);

        result.append("JNDI directory on " + host + ":" + port
                + "\n<== / ==>\n");

        recurseBindings((Context) rootContext, rootContext.listBindings("/"),
                "\t", result);

        return result.toString();
    }

    /**
     * see isBound(Context parentContext,String childContextName, String name)
     * 
     * @param parentContext
     * @param childContextName
     * @param name
     * @return
     */
    public static boolean isBound(Context parentContext, Name childContextName,
            Name name) {
        return isBound(parentContext, NamingHelper
                .nameToString(childContextName), NamingHelper
                .nameToString(name));
    }

    /**
     * see isBound(Context parentContext,String childContextName, String name)
     * 
     * @param parentContext
     * @param childContextName
     * @param name
     * @return
     */
    public static boolean isBound(Context parentContext, Name childContextName,
            String name) {
        return isBound(parentContext, NamingHelper
                .nameToString(childContextName), name);
    }

    /**
     * see isBound(Context parentContext,String childContextName, String name)
     * 
     * @param parentContext
     * @param childContextName
     * @param name
     * @return
     */
    public static boolean isBound(Context parentContext,
            String childContextName, Name name) {
        return isBound(parentContext, childContextName, NamingHelper
                .nameToString(name));
    }

    /**
     * Check if a name is bound is a specific context
     * 
     * @param parentContext
     *            parent context where to list the content of the child context
     * @param childContextName
     *            name of the child context, the context in we want to look for
     *            the binding name
     * @param name
     *            binding name
     * @return
     */
    public static boolean isBound(Context parentContext,
            String childContextName, String name) {
        boolean test = false;
        try {
            NamingEnumeration<NameClassPair> namingEnumeration = parentContext
                    .list(childContextName);
            while (namingEnumeration.hasMoreElements()) {
                NameClassPair classPair = (NameClassPair) namingEnumeration
                        .next();
                if (classPair.getName().equals(name)) {
                    return true;
                }
            }
        } catch (NamingException e) {
            return false;
        }
        return test;
    }

    /**
     * Recurse the contexts
     * 
     * @param bindings
     * @throws NamingException
     */
    protected static void recurseBindings(Context context,
            NamingEnumeration<Binding> bindings, String tabs,
            StringBuffer result) throws NamingException {

        while (bindings.hasMoreElements()) {

            Binding binding = (Binding) bindings.next();

            Object object = binding.getObject();

            if (object instanceof Context) {

                result.append("\n" + tabs + "<=" + binding.getName() + "=>\n");

                try {

                    recurseBindings((Context) object, context
                            .listBindings(binding.getName()), tabs + "\t",
                            result);
                } catch (Exception e) {

                    recurseNames((Context) object, context.list(binding
                            .getName()), tabs + "\t", result);
                }
            } else if (object instanceof Map) {

                result.append(tabs + "-> " + binding.getName() + "\n");

                Map<?, ?> map = (Map<?, ?>) object;

                for (Entry<?, ?> entry : map.entrySet()) {

                    result.append(tabs + "\t" + "-> (" + entry.getKey() + ") "
                            + entry.getValue() + "\n");
                }
            } else if (object instanceof LinkRef) {

                LinkRef linkRef = (LinkRef) binding.getObject();

                result.append(tabs + "-> " + binding.getName() + " [Link to "
                        + linkRef.getLinkName() + "]\n");

            } else {

                result.append(tabs + "-> " + binding.getObject() + "\n");
            }
        }
    }

    /**
     * Recurse the contexts
     * 
     * @param bindings
     * @throws NamingException
     */
    protected static void recurseNames(Context context,
            NamingEnumeration<NameClassPair> names, String tabs,
            StringBuffer result) throws NamingException {

        while (names.hasMoreElements()) {

            NameClassPair name = (NameClassPair) names.next();

            result.append(tabs + "-> " + name.getName() + "\n");
        }
    }

    /**
     * Retrieve an initial context
     * 
     * @param properties
     *            properties of the JNDI directory
     * @return Initialcontext
     * @throws NamingException
     *             if an error occurs
     */
    protected static InitialContext retrieveContext(Properties properties)
        throws NamingException {
        return new InitialContext(properties);
    }

}
