/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.util;

import java.io.DataInputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.util.Enumeration;

/**
 * Network utilities
 * 
 * @author Christophe Hamerling - EBM WebSourcing
 * 
 */
public class NetworkUtil {

    /** The IPIv4 loopback address */
    public static String IPV4_LOOPBACK_ADDRESS = "127.0.0.1";
    
    /**
     * Creates a new instance of NetworkUtil
     */
    private NetworkUtil() {
    }

    /**
     * Get the IPv4 address of current host. The loopback address is returned if
     * no other interface has been found.
     * 
     * @return
     * @throws SocketException
     */
    public static String getIPv4Address() throws SocketException {
        Enumeration<NetworkInterface> e = NetworkInterface
            .getNetworkInterfaces();

        while (e.hasMoreElements()) {
            NetworkInterface netItf = e.nextElement();
            Enumeration<InetAddress> addresses = netItf.getInetAddresses();

            while (addresses.hasMoreElements()) {
                InetAddress ip = addresses.nextElement();
                if (!ip.isLoopbackAddress() && isIPv4(ip.getHostAddress())) {
                    return ip.getHostAddress();
                }
            }
        }
        return NetworkUtil.IPV4_LOOPBACK_ADDRESS;
    }

    /**
     * Resolve the "localhost" IP address. Call the system command "ipconfig" or
     * "ifconfig" and parse the result to extract the IP.
     * 
     * @return the real IP address of the machine, or 127.0.0.1 if the system
     *         call fails or if not parsed.
     */
    protected static String getAddressFromSystemCall() {
        String host = IPV4_LOOPBACK_ADDRESS;
        String hostIP = "";
        String os = System.getProperty("os.name").toLowerCase();

        String command = "/sbin/ifconfig";

        if (os.indexOf("windows") != -1) {
            // We are on Windows
            command = "ipconfig";

        } else if (os.indexOf("linux") != -1) {
            // We are on Linux
            command = "/sbin/ifconfig";

        } else if (os.indexOf("mac") != -1) {
            // We are on a Mac :)
            command = "ifconfig";
        }

        // Execute the command
        try {
            Process process = Runtime.getRuntime().exec(command);
            try {
                process.waitFor();
            } catch (InterruptedException e) {
                // Do nothing
            }
            DataInputStream dis = new DataInputStream(process.getInputStream());
            byte[] content = new byte[dis.available()];
            dis.readFully(content);
            String output = new String(content);

            // search address from result data
            if (os.indexOf("windows") != -1 && output.length() > 45) {
                hostIP = getAddressFromWindowsSystemCallResult(output);
            } else if (os.indexOf("linux") != -1) {
                hostIP = getAddressFromLinuxSystemCallResult(output);
            } else if (os.indexOf("mac") != -1) {
                hostIP = getAddressFromMacSystemCallResult(output);
            } else {
                // other platforms ?
            }

        } catch (IOException e) {
            // Do nothing
        }

        // If connection cable is disconnected, the retrieved IP is a
        // bad IP ("Media disconnected" for exemple), so we have to test
        // if the retrieved IP starts with a number.
        if (hostIP.charAt(0) >= '0' && hostIP.charAt(0) <= '9') {
            host = hostIP;
        }

        return host.trim();
    }

    /**
     * 
     * @param output
     * @return
     */
    private static String getAddressFromWindowsSystemCallResult(String output) {
        String hostIP = "";
        // We are on Windows, and the ipconfig command return a
        // String with information
        // (size >45 char, otherwise, pb with IPCONFIG command)
        if (output.indexOf("IP") > -1) {
            hostIP = output.substring(output.indexOf(": ", output.indexOf("IP",
                output.indexOf("IP") + 10)) + 2, output.indexOf(": ", output
                .indexOf("IP", output.indexOf("IP") + 10)) + 15);

        }
        return hostIP;
    }

    /**
     * TODO
     * 
     * @param output
     * @return
     */
    private static String getAddressFromMacSystemCallResult(String output) {
        return "";
    }

    /**
     * 
     * @param output
     * @return
     */
    private static String getAddressFromLinuxSystemCallResult(String output) {
        String hostIP = "";
        // We are on Linux
        if (output.indexOf("lo") > -1) {
            // Remove the loopback part
            output = output.substring(0, output.indexOf("lo"))
                + output.substring(output.indexOf("\n", output.indexOf(")",
                    output.indexOf("lo"))), output.length());
        }
        if (output.indexOf("inet addr:") > -1) {
            // Extract the first ip address
            hostIP = output.substring(output.indexOf("inet addr:") + 10, output
                .indexOf(" ", output.indexOf("inet addr:") + 10));
        } else if (output.indexOf("inet adr:") > -1) {
            // Extract the first ip address
            hostIP = output.substring(output.indexOf("inet adr:") + 9, output
                .indexOf(" ", output.indexOf("inet adr:") + 10));
        }
        return hostIP;
    }

    /**
     * 
     * @param hostAddress
     * @return
     */
    public static boolean isIPv4(String hostAddress) {
        return hostAddress.split("[.]").length == 4;
    }

}
