/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: PropertyUtil.java 14:47:22 ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.util;

import java.io.File;
import java.io.FileOutputStream;
import java.net.URL;
import java.util.Properties;

import org.objectweb.petals.PetalsException;

/**
 * The property util is used to load and retrieve properties from file
 * 
 * @author ddesjardins - eBMWebsourcing
 */
public final class PropertyUtil {

    /**
     * Name of the server properties file
     */
    public static final String SERVER_PROPS = "/server.properties";

    private PropertyUtil() {
        // Do nothing
    }

    /**
     * Retrieve the JNDI properties of a properties file
     * 
     * @return bunch of JNDI properties
     */
    public static Properties retrieveJNDIProperties() {
        Properties outProperties = new Properties();
        outProperties.setProperty("java.naming.factory.host", SystemUtil
                .getHost());
        outProperties.setProperty("java.naming.factory.port", SystemUtil
                .getJndiPort());
        outProperties.setProperty("java.naming.factory.initial", SystemUtil
                .getJndiFactory());
        return outProperties;
    }

    /**
     * Update the container properties
     * 
     * @param joramId
     *            new joram id
     * @param joramDomainPort
     *            new joram port
     * @param joramTCPPort
     *            new joram tcp port
     * @param jmxPort
     *            new jmx port
     * @param htmlPort
     *            new html port
     * @param jndiPort
     *            new jndi port
     * @throws PetalsException 
     */
    public static void updateContainerProperties(String joramId,
            String joramDomainPort, String joramTCPPort, String jmxPort,
            String htmlPort, String jndiPort, long uid) throws PetalsException {
        try {
            URL url = PropertyUtil.class.getResource(SERVER_PROPS);
            Properties properties = new Properties();
            properties.load(url.openStream());
            properties.setProperty("joram.domainport", joramDomainPort);
            properties.setProperty("joram.id", joramId);
            properties.setProperty("joram.tcpport", joramTCPPort);
            properties.setProperty("jmx.port", jmxPort);
            properties.setProperty("java.naming.factory.port", jndiPort);
            properties.setProperty("html.port", htmlPort);
            properties.setProperty("container.uid", Long.toString(uid));

            properties.store(new FileOutputStream(new File(url.toURI())),
                    "New container properties set");
        } catch (Exception e) {
            throw new PetalsException("Error updating container properties", e);
        }
    }

    /**
     * Retreive a property from a properties file
     * 
     * @param propertiesFile
     *            name of the properties file (must be in the classpath)
     * @param key
     *            key of the property we want to retrieve
     * @return property associated with the key
     * @throws PetalsException
     *             If a key is not defined in the property file or if any
     *             problems occurs while opening the file
     */
    protected static String getProperty(String propertiesFile, String key)
        throws PetalsException {
        String value = null;
        try {
            URL url = PropertyUtil.class.getResource(propertiesFile);
            Properties properties = new Properties();
            properties.load(url.openStream());
            if (properties.containsKey(key)) {
                value = properties.getProperty(key);
            } else {
                throw new PetalsException("Key " + key
                        + " is not defined in the property file "
                        + propertiesFile);
            }
        } catch (Exception e) {
            throw new PetalsException("Problem while trying to get property "
                    + key + " of " + propertiesFile, e);
        }
        return value;
    }

    /**
     * Retreive a property from a properties file
     * 
     * @param propertiesFile
     *            name of the properties file (must be in the classpath)
     * @param key
     *            key of the property we want to retrieve
     * @param defaultValue
     *            default value if no value is found with the given key
     * @return property associated with the key
     * @throws PetalsException
     *             If any problems occurs while opening the file
     */
    protected static String getProperty(String propertiesFile, String key,
            String defaultValue) throws PetalsException {
        String value = null;
        try {
            URL url = PropertyUtil.class.getResource(propertiesFile);
            Properties properties = new Properties();
            properties.load(url.openStream());
            value = properties.getProperty(key, defaultValue);
        } catch (Exception e) {
            throw new PetalsException("Problem while trying to get property "
                    + key + " of " + propertiesFile, e);
        }
        return value;
    }
}
