/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: Installer.java 154 2006-03-27 15:30:10Z ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.component.lifecycle;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.component.Bootstrap;
import javax.jbi.component.Component;
import javax.jbi.component.InstallationContext;
import javax.jbi.management.LifeCycleMBean;
import javax.management.AttributeChangeNotification;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanNotificationInfo;
import javax.management.Notification;
import javax.management.NotificationBroadcaster;
import javax.management.NotificationBroadcasterSupport;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectName;

import org.objectweb.petals.PetalsException;
import org.objectweb.petals.classloader.LoaderManager;
import org.objectweb.petals.classloader.PetalsClassLoader;
import org.objectweb.petals.jbi.component.context.ComponentContextImpl;
import org.objectweb.petals.jbi.component.context.InstallationContextImpl;
import org.objectweb.petals.jbi.component.event.StateChangeFailedEvent;
import org.objectweb.petals.jbi.component.event.StateChangedEvent;
import org.objectweb.petals.jbi.component.thread.BootstrapThread;
import org.objectweb.petals.jbi.component.thread.ComponentLifeCycleThread;
import org.objectweb.petals.jbi.management.service.LifeCycleManagerService;
import org.objectweb.petals.jbi.management.service.MBeanNamesImpl;
import org.objectweb.petals.jbi.management.service.ManagementException;
import org.objectweb.petals.jbi.management.systemstate.SystemState;
import org.objectweb.petals.tools.jbicommon.descriptor.ComponentDescription;
import org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor;
import org.objectweb.petals.tools.jbicommon.descriptor.SharedLibraryList;
import org.objectweb.petals.util.ConvertSpecialCharactersUtil;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * 
 * @author Olivier Fabre - EBM Websourcing
 * @author wjoseph - eBMWebsourcing
 * 
 */
public class Installer implements InstallerMBean, NotificationBroadcaster {

    /**
     * Bootstrap classloader id prefix
     */
    public static final String BS_PREFIX = "BS";

    /**
     * Component classloader id prefix
     */
    public static final String COMP_PREFIX = "COMP";

    public static final String INSTALLED = "Installed";

    public static final String INSTALLING = "Installing";

    public static final String UNINSTALLED = "Uninstalled";

    public static final String UNINSTALLING = "Uninstalling";

    public static final String UNKNOWN = "Unknown";

    private static final String BINDING_COMPONENT = "binding-component";

    private static final String SERVICE_ENGINE = "service-engine";

    // /////////////////
    // Notifications support
    // /////////////////
    protected Object activitySynchronizer;

    protected String state;

    NotificationBroadcasterSupport notifSupport = new NotificationBroadcasterSupport();

    protected ComponentContextImpl componentContext;

    protected ComponentDescription componentDescription;

    protected ObjectName componentJmxName;

    protected URI installationRoot;

    protected Bootstrap jbiBootstrap;

    protected LifeCycleManagerService jmxAdmin;

    protected LoaderManager loadSrv;

    protected BootstrapThread bootstrapThread;

    protected LoggingUtil log;

    /**
     * The ObjectName of this installer
     */
    protected ObjectName mBeanName;

    protected SystemState recoverySrv;

    /**
     * A sequence number used by state change notification
     */
    protected long sequenceNumber;

    /**
     * Create the Installer that will manage installation / uninstallation of
     * the Jbi component. The related bootstrap is created and initialize here.
     * 
     * @param installationRoot
     * @param loadSrv
     * @param componentContext
     * @param componentDescription
     * @param jmxAdmin
     * @param logger
     * @throws JBIException
     */
    public Installer(URI installationRoot, LoaderManager loadSrv,
        ComponentContextImpl componentContext,
        ComponentDescription componentDescription,
        LifeCycleManagerService jmxAdmin, SystemState recoverySrv,
        ObjectName mbeanName, Logger logger) {
        super();
        log = new LoggingUtil(logger);

        log.start();

        this.jmxAdmin = jmxAdmin;

        this.recoverySrv = recoverySrv;

        this.componentDescription = componentDescription;

        this.installationRoot = installationRoot;

        this.componentContext = componentContext;

        this.loadSrv = loadSrv;

        this.mBeanName = mbeanName;

        this.state = UNINSTALLED;

        activitySynchronizer = new Object();

        log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.NotificationBroadcaster#addNotificationListener(javax.management.NotificationListener,
     *      javax.management.NotificationFilter, java.lang.Object)
     */
    public void addNotificationListener(NotificationListener arg0,
        NotificationFilter arg1, Object arg2) throws IllegalArgumentException {
        notifSupport.addNotificationListener(arg0, arg1, arg2);
    }

    /**
     * Get the unique component name of this component, as assigned by the
     * identification section of this component's installation descriptor.
     * 
     * @return the component name; must be non-null and non-empty.
     */
    public String getComponentName() {
        return componentContext.getComponentName();
    }

    // ////////////////////////////////////////////
    // ElementListener implementation
    // ////////////////////////////////////////////
    /**
     * @see javax.jbi.management.LifeCycleMBean#getCurrentState()
     */
    public String getCurrentState() {
        if (!(LifeCycleMBean.SHUTDOWN.equals(state)
            || LifeCycleMBean.STARTED.equals(state) || LifeCycleMBean.STOPPED
            .equals(state))) {
            return LifeCycleMBean.UNKNOWN;
        }
        return state;
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.management.InstallerMBean#getInstallerConfigurationMBean()
     */
    public ObjectName getInstallerConfigurationMBean() throws JBIException {
        LifeCycleAbstract lifeCycle = jmxAdmin
            .getComponentByName(componentContext.getComponentName());
        if (!INSTALLED.equals(state)
            || (lifeCycle != null && !lifeCycle.getCurrentState().equals(
                LifeCycleMBean.SHUTDOWN))) {
            throw new JBIException("Component is not in the SHUTDOWN state");
        }
        
        if (jbiBootstrap == null) {
            throw new JBIException("Bootstrap should not be null");
        }
        
        return bootstrapThread.getExtensionMBeanName();
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.management.InstallerMBean#getInstallRoot()
     */
    public String getInstallRoot() {
        return componentContext.getInstallRoot();
    }

    /**
     * 
     * @return the installer ObjectName
     */
    public ObjectName getMBeanName() {
        return mBeanName;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.NotificationBroadcaster#getNotificationInfo()
     */
    public MBeanNotificationInfo[] getNotificationInfo() {
        String[] types = new String[] {AttributeChangeNotification.ATTRIBUTE_CHANGE};

        String name = AttributeChangeNotification.class.getName();

        String description = "Notification about a change of the lifeCycle state or a state change failure.";

        MBeanNotificationInfo info = new MBeanNotificationInfo(types, name,
            description);

        return new MBeanNotificationInfo[] {info};
    }

    /**
     * Install the jbi component. Call the Bootstrap onInstall() and cleanup()
     * methods, register the jbiComponent object, and return its jmxName.
     * 
     * @throws JBIException
     *             the component is already installed
     */
    public ObjectName install() throws JBIException {
        log.start();
        if (UNINSTALLED.equals(state)) {
            synchronized (activitySynchronizer) {
                try {
                    setState(INSTALLING);

                    doInstall();

                    setState(INSTALLED);
                } catch (Throwable re) {
                    setState(UNKNOWN);

                    String msg = "An exception occured during the installation.";

                    JBIException e = new JBIException(msg, re);

                    log.error(msg, e);

                    stateChangeFailed(new StateChangeFailedEvent(this, e));

                    throw e;
                }
            }
        } else {
            String msg = "The Object can not be installed in this state.";

            JBIException e = new JBIException(msg);

            log.error(msg, e);

            stateChangeFailed(new StateChangeFailedEvent(this, e));

            throw e;
        }
        log.end();

        return componentJmxName;
    }

    /**
     * Test the installation flag
     * 
     * @return true if the component is already installed
     */
    public boolean isInstalled() {
        boolean result = false;
        if (INSTALLED.equals(state)) {
            result = true;
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.NotificationBroadcaster#removeNotificationListener(javax.management.NotificationListener)
     */
    public void removeNotificationListener(NotificationListener arg0)
        throws ListenerNotFoundException {
        notifSupport.removeNotificationListener(arg0);
    }

    /**
     * Set the <code>State</code> reference of the current state. An
     * <code>LifeCycleStateEventAbstract</code> is thrown.
     * 
     * @param state
     * @throws JBIException
     */
    public synchronized void setState(String state) throws JBIException {
        stateChanged(new StateChangedEvent(this, this.state, state));

        this.state = state;

        try {
            recoverySrv.updateComponentInstallationState(getComponentName(),
                state);
        } catch (Exception e) {
            String msg = "Component state can't be persisted for recovery";
            log.error(msg, e);
            throw new JBIException(msg, e);
        }
    }

    /**
     * <i>stateChanged</i> is called when the state of a lifecycle element
     * changes. A JMX-notification is thrown.
     * 
     * @see petals.management.element.ElementStateListener#stateChanged(java.util.EventObject)
     */
    public synchronized void stateChanged(StateChangedEvent event) {
        log.call();
        Notification n = new AttributeChangeNotification(mBeanName,
            sequenceNumber++, System.currentTimeMillis(), "State of '"
                + mBeanName.toString() + "' changed : " + event.getOldState()
                + " -> " + event.getNewState(), "state", "String", event
                .getOldState(), event.getNewState());

        notifSupport.sendNotification(n);
    }

    /**
     * <i>stateChangeFailed</i>> is called when the change of state of a
     * lifecycle element fails. A JMX-notification throws the relative
     * JBIException, .
     * 
     * @see petals.management.element.ElementStateListener#stateChanged(java.util.EventObject)
     */
    public synchronized void stateChangeFailed(StateChangeFailedEvent event) {
        log.call();
        Notification n = new AttributeChangeNotification(mBeanName,
            sequenceNumber++, System.currentTimeMillis(), "State change of '"
                + mBeanName.toString() + "' failed : "
                + event.getException().getMessage(), "exception", "String",
            null, event.getException().toString());

        notifSupport.sendNotification(n);
    }

    /**
     * Uninstall the jbi component. Call the Bootstrap onUninstall() and
     * cleanup() methods, deregister the jbiComponent object,
     * 
     * @throws JBIException
     *             the component is already uninstalled
     */
    public void uninstall() throws JBIException {
        log.start();
        if (INSTALLED.equals(state)) {
            synchronized (activitySynchronizer) {
                try {

                    ComponentLifeCycle componentLifeCycle = getComponentLifeCycle();
                    String compState = componentLifeCycle.getCurrentState();

                    if (ComponentLifeCycle.STARTED.equals(compState)) {
                        componentLifeCycle.stop();
                        componentLifeCycle.shutDown();
                    } else if (ComponentLifeCycle.STOPPED.equals(compState)) {
                        componentLifeCycle.shutDown();
                    }
                    
                    compState = componentLifeCycle.getCurrentState();

                    if (ComponentLifeCycle.SHUTDOWN.equals(compState)) {
                        setState(UNINSTALLING);
                        doUninstall();
                        setState(UNINSTALLED);
                    }
                } catch (Throwable re) {
                    setState(UNKNOWN);

                    String msg = "An exception occured during the uninstallation.";

                    JBIException e = new JBIException(msg, re);

                    log.error(msg, e);

                    stateChangeFailed(new StateChangeFailedEvent(this, e));

                    throw e;
                }
            }
        } else {
            String msg = "The Object can not be uninstalled in this state.";

            JBIException e = new JBIException(msg);

            log.error(msg, e);

            stateChangeFailed(new StateChangeFailedEvent(this, e));

            throw e;
        }
        log.end();
    }

    /**
     * Create the classLoader for Bootstrap
     * 
     * @return classLoader for Bootstrap
     */
    protected PetalsClassLoader createBootstrapClassloader() {
        log.call();

        List<String> classpath = componentDescription.getBootstrapClassPath();

        boolean parentFirst = JBIDescriptor.isParentFirst(componentDescription
            .getBootstrapClassLoaderDelegation());

        return createClassLoader(BS_PREFIX, classpath, parentFirst, null);
    }

    /**
     * Create a classloader with the specified parameters.
     * 
     * @param prefix
     *            bootstrap or component prefix
     * @param classpath
     *            classpath of the bootstrap/component
     * @param parentFirst
     *            delegation mode
     * @param slNameList
     *            list of the name of the required sharedLibraries
     * @return a classloader
     */
    protected PetalsClassLoader createClassLoader(String prefix,
        List<String> classpath, boolean parentFirst, List<String> slNameList) {
        log.call();

        // prepare bootstrap and component classloader
        List<URL> urls = new ArrayList<URL>();

        try {
            // TODO Find a better way of managing URI
            String installRootPath;
            if (installationRoot.toURL().getPath().startsWith("/")) {
                installRootPath = "file://"
                    + installationRoot.toURL().getPath();
            } else {
                installRootPath = "file:///"
                    + installationRoot.toURL().getPath();
            }

            installRootPath = ConvertSpecialCharactersUtil
                .convertPath(installRootPath);

            urls.add(new URL(installRootPath));
        } catch (MalformedURLException e) {
            // simple do nothing - JBI Package has responsibility in
            // checking the archive contents against informations
            // contained in JBI Descriptor first.
            String msg = "Unexpected error. " + e.getMessage();
            log.info(msg, e);
        }
        URL[] baseUrls = urls.toArray(new URL[] {});

        PetalsClassLoader classLoader = null;

        try {
            classLoader = (PetalsClassLoader) loadSrv
                .createComponentClassLoader(prefix
                    + componentDescription.getIdentification().getName(),
                    baseUrls, classpath, parentFirst, slNameList);
        } catch (PetalsException pe) {
            // just log when some error occur
            log.error("Error creating bootstrap class loader. "
                + pe.getMessage(), pe);

            log.end();
            return null;
        }
        return classLoader;
    }

    /**
     * Create the classloader for the component Be careful, we use the classpath
     * set in the InstallationContext. (during the installation phase, the
     * Bootstrap can modify the component's classpath by setting it in the
     * InstallationContext object.)
     * 
     * @return classloader for the component
     */
    @SuppressWarnings("unchecked")
    protected PetalsClassLoader createComponentClassloader() {
        log.call();

        InstallationContext installContext = new InstallationContextImpl(
            componentDescription, componentContext, true);
        
        List<String> classpath = (List<String>) installContext
            .getClassPathElements();

        // delegation model
        boolean parentFirst = JBIDescriptor.isParentFirst(componentDescription
            .getComponentClassLoaderDelegation());

        // sharedlibraries used by this component
        List<SharedLibraryList> slList = componentDescription
            .getSharedLibraryList();

        List slNameList = new ArrayList();

        if (slList != null) {
            for (Iterator iter = slList.iterator(); iter.hasNext();) {
                SharedLibraryList element = (SharedLibraryList) iter.next();
                slNameList.add(element.getName());
            }
        }

        return createClassLoader(COMP_PREFIX, classpath, parentFirst,
            slNameList);
    }

    /**
     * Do the bootstrap installation.
     * 
     * @throws JBIException
     */
    protected void doInstall() throws JBIException {

        // load the bootstrap
        if (jbiBootstrap != null) {
            throw new JBIException(
                "Bootstrap should be null, the installer has already been initialized");
        } 
        
        // 1. create the bootstrap classloader
        PetalsClassLoader classloader = createComponentClassloader();
        
        // 2. create the bootstrap object
        jbiBootstrap = loadBootstrap(classloader);
        
        // 3. create the bootstrap thread
        bootstrapThread = createBootstrapThread(classloader);
        
        // 4. create installation context
        InstallationContextImpl installContext = new InstallationContextImpl(
            componentDescription, componentContext, true);

        // 5. init and install
        bootstrapThread.init(installContext);
        bootstrapThread.onInstall();

        // 6. create the component life cycle
        componentJmxName = createComponentLifeCycleMBean();

        bootstrapThread.cleanUp();
    }
    
    /**
     * Create the thread that will be used to process component installation steps.
     * The thread-classloader is set with the given paremeter.
     * 
     * @param classloader
     * @return the BootstrapThread
     */
    protected BootstrapThread createBootstrapThread(PetalsClassLoader classloader) {
        BootstrapThread it = new BootstrapThread(this, jbiBootstrap);
        it.setContextClassLoader(classloader);
        it.start();
        return it;
    }

    /**
     * Create the thread that will be used to process component life cycle steps.
     * The thread-classloader is set with the given paremeter.
     * 
     * @param classloader
     * @return the BootstrapThread
     */
    protected ComponentLifeCycleThread createComponentLifeCycleThread(
        Component component, PetalsClassLoader classloader) {
        ComponentLifeCycleThread clt = new ComponentLifeCycleThread(component);
        clt.setContextClassLoader(classloader);
        clt.start();
        return clt;
    }

    /**
     * Register the installer engine/binding in the life cycle manager service
     *  
     * @return
     * @throws JBIException
     */
    protected ObjectName createComponentLifeCycleMBean() throws JBIException {
        ObjectName objectName = null;

        // 1. create component classloader
        PetalsClassLoader cl = createComponentClassloader();
        
        // 2. create the component object
        Component jbiComponent = loadComponent(cl);
        
        // 3. create the componentLifeCycle thread
        ComponentLifeCycleThread lifeCycleThread = createComponentLifeCycleThread(
            jbiComponent, cl);
        
        // 4. create the Mbean (as binding or engine)
        MBeanNamesImpl mbeanNames = jmxAdmin.getMBeanNames();
        if (BINDING_COMPONENT.equals(componentDescription.getType())) {
            objectName = mbeanNames
                .createBindingComponentMBeanName(componentDescription
                    .getIdentification().getName());
        } else if (SERVICE_ENGINE.equals(componentDescription.getType())){
            objectName = mbeanNames
            .createEngineComponentMBeanName(componentDescription
                .getIdentification().getName());
        }
        
        // 5. create life cycle
        ComponentLifeCycle lifecycle = new ComponentLifeCycle(
            componentDescription, jbiComponent, componentContext, recoverySrv,
            objectName, lifeCycleThread, log);
        
        // 6. register engine or binding
        if (BINDING_COMPONENT.equals(componentDescription.getType())) {
            objectName = jmxAdmin.registerBindingComponent(lifecycle);
        } else if (SERVICE_ENGINE.equals(componentDescription.getType())) {
            objectName = jmxAdmin.registerEngineComponent(lifecycle);
        }
        
        return objectName;
    }

    /**
     * Unregister the installer engine/binding from the life cycle manager service

     * @throws ManagementException
     */
    protected void unregisterComponentLifeCycleMBean() throws ManagementException {
        if (BINDING_COMPONENT.equals(componentDescription.getType())) {
            jmxAdmin.unregisterBindingComponent(componentJmxName);
        } else if (SERVICE_ENGINE.equals(componentDescription.getType())) {
            jmxAdmin.unregisterEngineComponent(componentJmxName);
        }
    }

    /**
     * Do the bootstrap uninstallation
     * 
     * @throws JBIException
     */
    protected void doUninstall() throws JBIException {

        if (jbiBootstrap == null) {
            throw new JBIException("Bootstrap should not be null");
        }

        // call uninstall on bootstrap
        // create an uninstallation context
        InstallationContextImpl uninstallContext = new InstallationContextImpl(
            componentDescription, componentContext, false);

        bootstrapThread.init(uninstallContext);

        bootstrapThread.onUninstall();

        // unregister the installer mbean
        unregisterComponentLifeCycleMBean();

        // unregister the optionnal extensionmbean
        if (componentContext.getComponent().getLifeCycle() != null) {
            ObjectName extName = componentContext.getComponent().getLifeCycle()
                .getExtensionMBeanName();
            try {
                jmxAdmin.getMBeanServer().unregisterMBean(extName);
            } catch (Exception e) {
                // nothing
            }
        }

        bootstrapThread.cleanUp();

        // stop current bootstrap thread
        // a new thread will be created on the next do install call
        bootstrapThread.shutdownThread();

        /*
         * unload component and bootstrap class loaders
         */
        String componentName = componentDescription.getIdentification()
            .getName();

        loadSrv.deleteClassLoader(Installer.COMP_PREFIX + componentName);
        loadSrv.deleteClassLoader(Installer.BS_PREFIX + componentName);
        
        // set parameters to null
        jbiBootstrap = null;

    }

    /**
     * Get the component life cycle
     * 
     * @return the component life cycle
     */
    protected ComponentLifeCycle getComponentLifeCycle() {
        String compName = componentDescription.getIdentification().getName();
        return jmxAdmin.getComponentByName(compName);
    }

    /**
     * load the bootstrap, with its own classloader
     * 
     * @return
     */
    protected Bootstrap loadBootstrap(PetalsClassLoader classloader)
        throws JBIException {
        log.call();
        Class bootstrapClass;
        Bootstrap jbiBootstrapTmp = null;
        try {
            // 1. get the classloader
            // PetalsClassLoader classloader = createBootstrapClassloader();

            bootstrapClass = classloader.loadClass(componentDescription
                .getBootstrapClassName());

            if (bootstrapClass == null) {
                throw new JBIException("Could not find bootstrap class: "
                    + componentDescription.getBootstrapClassName());
            }
            jbiBootstrapTmp = (Bootstrap) bootstrapClass.newInstance();

            // 2. set the logger factory
            // LoggerFactory loggerFactory = new LoggerFactory();
            // loggerFactory.getLogger("Petals");
            // LogManager logManager = LogManager.getLogManager();
            // this.componentContext.setRootLogger(loggerFactory.getLogger(descr));

        } catch (ClassNotFoundException e) {
            throw new JBIException("Class not found: "
                + componentDescription.getBootstrapClassName(), e);
        } catch (InstantiationException e) {
            throw new JBIException("Could not instantiate : "
                + componentDescription.getBootstrapClassName(), e);
        } catch (IllegalAccessException e) {
            throw new JBIException("Illegal access on: "
                + componentDescription.getBootstrapClassName(), e);
        }
        return jbiBootstrapTmp;
    }

    /**
     * load the component, with its own classloader
     * 
     * @return the component
     */
    protected javax.jbi.component.Component loadComponent(PetalsClassLoader classloader) throws JBIException {
        log.call();

        Class componentClass;
        javax.jbi.component.Component jbiComponent = null;

        try {
            componentClass = classloader.loadClass(componentDescription
                .getComponentClassName());

            if (componentClass == null) {
                throw new JBIException("Could not find component class: "
                    + componentDescription.getComponentClassName());
            }

            jbiComponent = (javax.jbi.component.Component) componentClass
                .newInstance();

            // 2. set component logger
            Class clazz = classloader.loadClass("java.util.logging.Logger");
            Method method = clazz.getMethod("getLogger", new Class[] {
                Class.forName("java.lang.String"),
                Class.forName("java.lang.String")});
            java.util.logging.Logger logger = (java.util.logging.Logger) method
                .invoke(null, new Object[] {
                    "root0.Petals" + ".components.container-"
                        + SystemUtil.getContainerName() + "."
                        + componentDescription.getIdentification().getName(),
                    null});
            componentContext.setRootLogger(logger);

        } catch (ClassNotFoundException e) {
            throw new JBIException("Class not found: "
                + componentDescription.getComponentClassName(), e);
        } catch (InstantiationException e) {
            throw new JBIException("Could not instantiate : "
                + componentDescription.getComponentClassName(), e);
        } catch (IllegalAccessException e) {
            throw new JBIException("Illegal access on: "
                + componentDescription.getComponentClassName(), e);
        } catch (SecurityException e) {
            throw new JBIException(
                "Security exception for java.util.logging.Logger", e);
        } catch (NoSuchMethodException e) {
            throw new JBIException(e);
        } catch (IllegalArgumentException e) {
            throw new JBIException(e);
        } catch (InvocationTargetException e) {
            throw new JBIException(e);
        }
        return jbiComponent;
    }
}
