/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: InstallationService.java 154 2006-03-27 15:30:10Z ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.component.lifecycle;

import java.net.URI;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.jbi.JBIException;
import javax.management.AttributeChangeNotification;
import javax.management.Notification;
import javax.management.ObjectName;

import org.objectweb.petals.PetalsException;
import org.objectweb.petals.jbi.component.event.StateChangeFailedEvent;
import org.objectweb.petals.jbi.component.event.StateChangedEvent;
import org.objectweb.petals.jbi.management.systemstate.SystemState;
import org.objectweb.petals.repository.RepositoryService;
import org.objectweb.petals.tools.jbicommon.descriptor.ServiceAssembly;
import org.objectweb.petals.util.LoggingUtil;

/**
 * This class manage the life cycle of deployed service assemblies
 * 
 * @author Olivier Fabre - EBM Websourcing
 * @author ddesjardins - eBMWebsourcing
 */
public class ServiceAssemblyLifeCycle extends LifeCycleAbstract {

    private URI installationRoot;

    private SystemState recoverySrv;

    private RepositoryService repositoryService;

    private ServiceAssembly serviceAssembly;

    private Map<String, ServiceUnitLifeCycle> serviceUnitsLifeCycles;

    /**
     * 
     * @param mbeanName
     *            name of the MBean for this LifeCycle. In fact, the
     *            DeploymentServiceMBean name is used as there is no
     *            SALifeCycleMBean
     * @param log
     * @param serviceAssembly
     * @param repositoryService
     * @param recoverySrv
     * @param installationRoot
     */
    public ServiceAssemblyLifeCycle(ObjectName mbeanName, LoggingUtil log,
            ServiceAssembly serviceAssembly,
            RepositoryService repositoryService, SystemState recoverySrv,
            URI installationRoot) {
        super(mbeanName, log);
        this.installationRoot = installationRoot;
        this.serviceAssembly = serviceAssembly;
        this.repositoryService = repositoryService;
        this.recoverySrv = recoverySrv;
        this.serviceUnitsLifeCycles = Collections
                .synchronizedMap(new HashMap<String, ServiceUnitLifeCycle>());
    }

    /**
     * Do nothing
     */
    @Override
    public void doInit() throws JBIException {

    }

    @Override
    public void doShutdown() throws JBIException {
        Map<String, ServiceUnitLifeCycle> sus = this
                .getServiceUnitsLifeCycles();

        synchronized (sus) {
            for (Iterator iter = sus.values().iterator(); iter.hasNext();) {
                ServiceUnitLifeCycle lifeCycle = (ServiceUnitLifeCycle) iter
                        .next();
                lifeCycle.shutDown();
            }
        }
    }

    @Override
    public void doStart() throws JBIException {
        Map<String, ServiceUnitLifeCycle> sus = this
                .getServiceUnitsLifeCycles();

        synchronized (sus) {
            for (Iterator iter = sus.values().iterator(); iter.hasNext();) {
                ServiceUnitLifeCycle lifeCycle = (ServiceUnitLifeCycle) iter
                        .next();
                lifeCycle.start();
            }
        }
    }

    @Override
    public void doStop() throws JBIException {
        Map<String, ServiceUnitLifeCycle> sus = this
                .getServiceUnitsLifeCycles();

        synchronized (sus) {
            for (Iterator iter = sus.values().iterator(); iter.hasNext();) {
                ServiceUnitLifeCycle lifeCycle = (ServiceUnitLifeCycle) iter
                        .next();
                lifeCycle.stop();
            }
        }
    }

    public URI getInstallationRoot() {
        return installationRoot;
    }

    public ServiceAssembly getServiceAssembly() {
        return serviceAssembly;
    }

    /**
     * Return the synchronized map of deployed service units.
     * 
     * @return
     */
    public Map<String, ServiceUnitLifeCycle> getServiceUnitsLifeCycles() {
        return serviceUnitsLifeCycles;
    }

    // ////////////////////////////////////////////
    // ElementListener implementation
    // ////////////////////////////////////////////

    /**
     * Register a ServiceUnitLifeCycle into a synchronized map of the
     * ServiceAssemblyLifeCycle
     * 
     * @param suName
     *            name of a service unit
     * @param serviceUnitLifeCycle
     *            life cycle of the service unit
     * @throws JBIException
     */
    public void registerSUIntoSA(String suName,
            ServiceUnitLifeCycle serviceUnitLifeCycle) throws JBIException {
        Map<String, ServiceUnitLifeCycle> sus = getServiceUnitsLifeCycles();
        synchronized (sus) {
            if (!sus.containsKey(suName)) {
                sus.put(suName, serviceUnitLifeCycle);
            } else {
                throw new JBIException("Duplicate service unit : same name");
            }

        }
    }

    /**
     * Unregister a ServiceUnitLifeCycle from a synchronized map of the
     * ServiceAssemblyLifeCycle
     * 
     * @param suName
     *            name of a service unit
     */
    public void unregisterSUFromSA(String suName) {
        Map<String, ServiceUnitLifeCycle> sus = getServiceUnitsLifeCycles();
        synchronized (sus) {
            sus.remove(suName);
        }
    }

    /**
     * Remove the package of the service assembly
     * 
     * @throws PetalsException
     *             if an error occured during package deletion
     * 
     */
    public void removePackage() throws PetalsException {
        repositoryService.removeServiceAssemblyPackage(serviceAssembly
                .getIdentification().getName());
    }

    @Override
    public synchronized void setState(String state) throws JBIException {
        super.setState(state);
        String saName = this.serviceAssembly.getIdentification().getName();
        try {
            recoverySrv.updateServiceAssemblyState(saName, state);
        } catch (Exception e) {
            String msg = "Service assembly state can't be persisted for recovery";
            log.error(msg, e);
            throw new JBIException(msg, e);
        }
    }

    /**
     * No init state for ServiceAssemblyLifeCycle
     */
    @Override
    public void start() throws JBIException {
        log.start();

        if (SHUTDOWN.equals(state) || STOPPED.equals(state)) {
            synchronized (activitySynchronizer) {
                try {
                    setState(STARTING);

                    doStart();

                    setState(STARTED);
                } catch (Throwable re) {
                    setState(UNKNOWN);

                    String msg = "A runtime exception occured during the start.";

                    JBIException e = new JBIException(msg, re);

                    log.error(msg, e);

                    stateChangeFailed(new StateChangeFailedEvent(this, e));

                    throw e;
                }
            }
        } else {
            String msg = "The Object can not be started in this state.";

            JBIException e = new JBIException(msg);

            log.error(msg, e);

            stateChangeFailed(new StateChangeFailedEvent(this, e));

            throw e;
        }
        log.end();
    }

    @Override
    public synchronized void stateChanged(StateChangedEvent event) {
        log.call();
        Notification n = new AttributeChangeNotification(mBeanName,
                sequenceNumber++, System.currentTimeMillis(),
                "State changed : " + event.getOldState() + " -> "
                        + event.getNewState(), "state", "String", event
                        .getOldState(), event.getNewState());

        notifSupport.sendNotification(n);
    }

    @Override
    public synchronized void stateChangeFailed(StateChangeFailedEvent event) {
        log.call();
        Notification n = new AttributeChangeNotification(mBeanName,
                sequenceNumber++, System.currentTimeMillis(),
                "State change failed : " + event.getException().getMessage(),
                "exception", "String", null, event.getException().toString());

        notifSupport.sendNotification(n);
    }

    /**
     * No init state for ServiceAssemblyLifeCycle
     */
    @Override
    public void stop() throws JBIException {
        log.start();

        if (STARTED.equals(state)) {
            synchronized (activitySynchronizer) {
                try {

                    setState(STOPPING);

                    doStop();

                    setState(STOPPED);

                } catch (Throwable re) {
                    setState(UNKNOWN);

                    String msg = "A runtime exception occured during the stop.";

                    JBIException e = new JBIException(msg, re);

                    log.error(msg, e);

                    stateChangeFailed(new StateChangeFailedEvent(this, e));

                    throw e;
                }
            }
        } else {
            String msg = "The Object can not be stoped in this state.";

            JBIException e = new JBIException(msg);

            log.error(msg, e);

            stateChangeFailed(new StateChangeFailedEvent(this, e));

            throw e;
        }
        log.end();
    }
}
