/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.component.thread;

import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;

import javax.jbi.JBIException;

/**
 * @author Christophe Hamerling - EBM WebSourcing
 * 
 */
public abstract class AbstractThread extends Thread {

    /**
     * 
     */
    protected JBIException jbiException;

    /**
     * The blocking working queue is equal to a FIFO. The worker thread is
     * waiting for an element to be added in this queue by the caller thread.
     * 
     * This queue is the synchronization object between caller thread and
     * execution thread. When an element is added by the caller thread, the
     * execution thread wakes up and do the task.
     */
    protected BlockingQueue<Integer> requestQueue;

    /**
     * The response queue used to wait for the task to complete. The caller
     * thread is blocked on this queue while the task is not complted by the
     * worker thread.
     * 
     * This queue is the synchronization object between execution thread and
     * caller thread. When the execution thread put an object in this queue, the
     * caller thread wakes up.
     */
    protected BlockingQueue<Integer> responseQueue;

    /**
     * Shutdown thread action
     */
    public static final int SHUTDOWNTHREAD = -1;

    /**
     * Creates a new instance of AbstractThread
     */
    public AbstractThread() {
        super();
        this.requestQueue = new ArrayBlockingQueue<Integer>(1);
        this.responseQueue = new ArrayBlockingQueue<Integer>(1);
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Runnable#run()
     */
    @Override
    public void run() {
        boolean run = true;
        int ret = 0;
        while (run) {
            try {
                // waits until there is one element available in the request
                // queue
                Integer action = requestQueue.take();
                ret = doTask(action.intValue());

                // if return < 0, stop the while loop and so stop the current
                // thread
                if (ret < 0) {
                    run = false;
                }

                // put element in response element wakes up the caller thread
                responseQueue.put(action);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        return;
    }

    /**
     * Excute the task. A JBI exception can occur during the task execution. In
     * this case, the local jbiException must be filled with this exception and
     * in shared between client and thread.
     * 
     * @return -1 if the current thread needs to be shutdowned
     */
    protected abstract int doTask(int action);

    /**
     * Execute the action. If the thread is not alive, throw an exception. On
     * the other cases, put the action in the FIFO and wait for the task to
     * complete. This method is the entry point of the task submission, multiple
     * threads will wait for other threads to complete.
     * 
     * @param state
     */
    protected synchronized void execute(int action) {
        // thread is alive test
        if (!this.isAlive()) {
            jbiException = new JBIException("Thread " + this.getName()
                + " is dead");
            return;
        }

        try {
            // put action in request FIFO, this will wait for an available place
            // in the queue. Since the queue is one element size, block if there
            // is a current task execution
            requestQueue.put(new Integer(action));

            // wait response == wait end of execution of the current query,
            // waiting for the doTask method to complete to be sure that tasks
            // are not done simultaneously
            responseQueue.take();
        } catch (InterruptedException e1) {
            e1.printStackTrace();
        }
    }

    /**
     * Get the JBIException that has occured during execution. The exception
     * could be null. Reset the current Exception.
     * 
     * @return the jbiException or null if no exception occured
     */
    public JBIException getJbiException() {
        JBIException jbie = null;
        if (jbiException != null) {
            jbie = new JBIException(jbiException);
            jbiException = null;
        }
        return jbie;
    }

    /**
     * Shutdown the current thread
     * 
     */
    public void shutdownThread() throws JBIException {
        execute(SHUTDOWNTHREAD);

        JBIException jbie = getJbiException();
        if (jbie != null) {
            throw jbie;
        }
    }
}
