/** 
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------  
 */
package org.objectweb.petals.jbi.component.thread;

import javax.jbi.JBIException;
import javax.jbi.component.Bootstrap;
import javax.jbi.component.Component;
import javax.jbi.component.InstallationContext;
import javax.management.ObjectName;

import org.objectweb.petals.jbi.component.lifecycle.Installer;

/**
 * The installer thread. This thread is used by the Installer component to have
 * each component in a separate bootstrap environment.
 * 
 * @author Christophe Hamerling - EBM Websourcing
 * 
 */
public class BootstrapThread extends AbstractThread implements Bootstrap {

    /**
     * The installer that is linked to this thread
     */
    private Installer installer;

    /**
     * The JBI component
     */
    private Component jbiComponent;

    /**
     * Management object name
     */
    private ObjectName objectName;

    /**
     * The Bootstrap
     */
    private Bootstrap jbiBootStrap;

    /**
     * The installation context attribute is required by the init method. This
     * attribute is set by the init method and used by the current thread.
     */
    private InstallationContext installContext;

    /*
     * STATICS : The different states that are used to process operations
     */

    public static final int INSTALL = 0;

    public static final int UNINSTALL = 1;

    public static final int INIT = 2;

    public static final int CLEAN = 3;

    public static final int EXTMBEAN = 4;

    /**
     * Creates a new instance of BootstrapThread
     * 
     * @param installer
     * @param jbiBootstrap
     */
    public BootstrapThread(Installer installer, Bootstrap jbiBootstrap) {
        super();
        this.installer = installer;
        this.jbiBootStrap = jbiBootstrap;
        this.setName(this.installer.getComponentName() + "-Installer Thread");
    }

    /**
     * Process the requested task.
     * 
     * @param action
     *            the task to execute
     * @return
     */
    protected int doTask(int action) {
        
        int result = 0;
        try {
            switch (action) {
                case CLEAN:
                    jbiBootStrap.cleanUp();
                    break;

                case INSTALL:
                    jbiBootStrap.onInstall();
                    break;

                case UNINSTALL:
                    jbiBootStrap.onUninstall();
                    break;

                case INIT:
                    jbiBootStrap.init(installContext);
                    break;

                case EXTMBEAN:
                    objectName = jbiBootStrap.getExtensionMBeanName();
                    break;

                case SHUTDOWNTHREAD:
                    result = -1;
                    break;

                default:
                    // NOTHING
                    break;
            }

        } catch (JBIException jbie) {
            jbiException = jbie;
        }
        
        return result;
    }

    /**
     * 
     * @return the jbiComponent
     */
    protected Component getJbiComponent() {
        return jbiComponent;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Bootstrap#cleanUp()
     */
    public void cleanUp() throws JBIException {
        execute(CLEAN);

        JBIException jbie = getJbiException();
        if (jbie != null) {
            throw jbie;
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Bootstrap#getExtensionMBeanName()
     */
    public ObjectName getExtensionMBeanName() {
        execute(EXTMBEAN);
        return objectName;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Bootstrap#init(javax.jbi.component.InstallationContext)
     */
    public void init(InstallationContext installContext) throws JBIException {
        this.installContext = installContext;
        execute(INIT);

        JBIException jbie = getJbiException();
        if (jbie != null) {
            throw jbie;
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Bootstrap#onInstall()
     */
    public void onInstall() throws JBIException {
        execute(INSTALL);

        JBIException jbie = getJbiException();
        if (jbie != null) {
            throw jbie;
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Bootstrap#onUninstall()
     */
    public void onUninstall() throws JBIException {
        execute(UNINSTALL);

        JBIException jbie = getJbiException();
        if (jbie != null) {
            throw jbie;
        }
    }
}