/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: AllConnectionRegistrationTask.java 154 6 oct. 06 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.management.deployment.deploy;

import java.util.HashMap;
import java.util.List;

import javax.jbi.JBIException;
import javax.xml.namespace.QName;

import org.objectweb.petals.jbi.management.deployment.DeploymentContextConstants;
import org.objectweb.petals.jbi.management.deployment.DeploymentUtils;
import org.objectweb.petals.jbi.management.service.EndpointService;
import org.objectweb.petals.jbi.management.service.ManagementException;
import org.objectweb.petals.processor.Task;
import org.objectweb.petals.tools.jbicommon.descriptor.Connection;
import org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor;
import org.objectweb.petals.tools.jbicommon.descriptor.ServiceAssembly;
import org.objectweb.petals.util.LoggingUtil;

/**
 * This task registers connections if exist
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
public class AllConnectionRegistrationTask implements Task {

    protected EndpointService endpointService;

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    public AllConnectionRegistrationTask(EndpointService endpointService,
            LoggingUtil log) {
        super();
        this.endpointService = endpointService;
        this.log = log;
    }

    public void execute(HashMap context) throws Exception {
        JBIDescriptor descriptor = (JBIDescriptor) context
                .get(DeploymentContextConstants.SA_DESCRIPTOR);

        List<Connection> connections = descriptor.getServiceAssembly()
                .getConnections();
        if (connections != null) {
            for (Connection connection : connections) {
                registerConnection(connection, descriptor);
            }
        }

    }

    /**
     * Register a connection into the endpoint registry of the JBI container
     * 
     * @param connection
     *            The {@link Connection} to register
     * @param descriptor
     *            the service assembly descriptor
     * @throws ManagementException
     *             when container fails to register a connection into the
     *             endpoint registry
     */
    protected void registerConnection(Connection connection,
            JBIDescriptor descriptor) throws ManagementException {
        String saName = DeploymentUtils.getServiceAssemblyName(descriptor);

        QName consInterface = connection.getConsumerInterfaceName();
        QName consService = connection.getConsumerServiceName();
        String consEndpoint = connection.getConsumerEndpointName();
        QName provService = connection.getProviderServiceName();
        String provEndpoint = connection.getProviderEndpointName();

        try {
            if (consInterface != null) {

                endpointService.createConnection(consInterface, provService,
                        provEndpoint);

            } else {
                endpointService.createConnection(consService, consEndpoint,
                        provService, provEndpoint);
            }
        } catch (JBIException e) {
            String msg = "Error while registering a service unit connection for the service assembly: "
                    + saName + ".";
            log.error(msg, e);
            throw new ManagementException(msg, e);
        }

    }

    public void undo(HashMap context) {
        JBIDescriptor descriptor = (JBIDescriptor) context
                .get(DeploymentContextConstants.SA_DESCRIPTOR);

        ServiceAssembly serviceAssembly = descriptor.getServiceAssembly();

        try {
            removeConnections(serviceAssembly);
        } catch (Exception e) {
            String msg = "Failed to revert a AllConnectionRegistrationTask";
            log.error(msg, e);
        }
    }

    /**
     * Remove all connections registered during the SA deployment process
     * 
     * @param saDescriptor
     *            the service assembly production element
     * @throws Exception
     *             if problem occurs during the deletion of a specific
     *             connection
     */
    protected void removeConnections(ServiceAssembly saDescriptor)
        throws Exception {
        List<Connection> connections = saDescriptor.getConnections();
        if (connections != null) {
            for (Connection connection : connections) {
                QName consInterface = connection.getConsumerInterfaceName();
                QName consService = connection.getConsumerServiceName();
                String consEndpoint = connection.getConsumerEndpointName();
                QName provService = connection.getProviderServiceName();
                String provEndpoint = connection.getProviderEndpointName();

                if (provEndpoint != null && provService != null) {
                    if (consInterface != null) {
                        endpointService.deleteConnection(consInterface,
                                provService, provEndpoint);
                    } else if (consService != null && consEndpoint != null) {
                        endpointService.deleteConnection(consService,
                                consEndpoint, provService, provEndpoint);
                    }
                }
            }
        }
    }

}
