/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: AllSUUndeploymentTask.java 154 10 oct. 06 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.management.deployment.undeploy;

import java.io.InputStream;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;

import javax.jbi.component.ServiceUnitManager;
import javax.jbi.management.DeploymentException;

import org.objectweb.petals.jbi.component.lifecycle.ComponentLifeCycle;
import org.objectweb.petals.jbi.component.lifecycle.LifeCycleAbstract;
import org.objectweb.petals.jbi.component.lifecycle.ServiceAssemblyLifeCycle;
import org.objectweb.petals.jbi.component.lifecycle.ServiceUnitLifeCycle;
import org.objectweb.petals.jbi.management.deployment.DeploymentContextConstants;
import org.objectweb.petals.jbi.management.deployment.DeploymentUtils;
import org.objectweb.petals.jbi.management.service.LifeCycleManagerService;
import org.objectweb.petals.jbi.management.service.util.XMLResult;
import org.objectweb.petals.processor.Task;
import org.objectweb.petals.tools.jbicommon.descriptor.ServiceUnit;
import org.objectweb.petals.util.LoggingUtil;

/**
 * This task executes the undeployment process of a service unit, for each
 * Service Unit included into the service assembly.
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
public class AllSUUndeploymentTask implements Task {

    /**
     * JBI Container :: JMX Agent Admin
     */
    protected LifeCycleManagerService managerService;

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    public AllSUUndeploymentTask(LifeCycleManagerService managerService,
            LoggingUtil log) {
        super();
        this.managerService = managerService;
        this.log = log;
    }

    public void execute(HashMap context) throws Exception {
        ServiceAssemblyLifeCycle saLifeCycle = (ServiceAssemblyLifeCycle) context
                .get(DeploymentContextConstants.SA_LIFECYCLE);

        XMLResult xmlResult = (XMLResult) context
                .get(DeploymentContextConstants.XML_RESULT);

        /*
         * Undeploy sus
         */
        undeploySUS(saLifeCycle, xmlResult);

    }

    /**
     * Undeploy all SUs by calling the "undeploySU" method of the corresponding
     * SUDeploymentManager.
     * 
     * @param saLifeCycle
     *            the parent {@link ServiceAssemblyLifeCycle}
     * @param xmlResult
     *            xml task result
     * @throws Exception
     *             if an error occurs during the undeployment of a service unit
     */
    protected void undeploySUS(ServiceAssemblyLifeCycle saLifeCycle,
            XMLResult xmlResult) throws Exception {
        /*
         * Put the service assembly in the shutdown state
         */
        String saState = saLifeCycle.getCurrentState();
        if (LifeCycleAbstract.STARTED.equals(saState)) {
            saLifeCycle.stop();
            saLifeCycle.shutDown();
        } else if (LifeCycleAbstract.STOPPED.equals(saState)) {
            saLifeCycle.shutDown();
        }
        saState = saLifeCycle.getCurrentState();
        if (saState.equals(LifeCycleAbstract.SHUTDOWN)) {
            /*
             * The service assembly is in the SHUTDOWN state We can try to
             * undeploy it
             */

            /*
             * Jbi spec specifies that service units must be undeployed in the
             * reverse order of the declared list in service assembly descriptor
             */
            List<ServiceUnit> suList = saLifeCycle.getServiceAssembly()
                    .getServiceUnits();
            Collections.reverse(suList);
            StringBuffer componentOutput = new StringBuffer();
            for (ServiceUnit suDesc : suList) {
                String xmlReturn = undeploySU(saLifeCycle, suDesc);
                if (!isValidComponentTask(xmlReturn, this.getClass()
                        .getClassLoader().getResourceAsStream(
                                "schema/management-message-component.xsd"))) {
                    xmlReturn = xmlResult.wrapComponentTaskResult(suDesc
                            .getTargetComponentName(), "undeploy", xmlReturn);
                }
                componentOutput.append(xmlReturn);
            }
            xmlResult.addComponentTaskResult(componentOutput);
        }
    }

    /**
     * Process the undeployment of the related service unit.
     * 
     * @param saLifeCycle
     *            the service assembly life cycle
     * @param serviceUnit
     *            the service unit production element
     * @return a {@link String} xml result of the service unit undeployment (see
     *         section 6.9 of the JSR 208)
     * @throws Exception
     *             if an error occurs during the undeployment process : can't
     *             retrieve the {@link ComponentLifeCycle}, can't retrieve the
     *             component {@link ServiceUnitManager} (SUM) or an exception is
     *             thrown during the call of the undeploy method of the SUM
     */
    protected String undeploySU(ServiceAssemblyLifeCycle saLifeCycle,
            ServiceUnit serviceUnit) throws Exception {
        String msg = "";
        String xmlReturn = null;

        /*
         * Undeploy it
         */
        String suName = DeploymentUtils.getServiceUnitName(serviceUnit);
        ServiceUnitLifeCycle serviceUnitLifeCycle = saLifeCycle
                .getServiceUnitsLifeCycles().get(suName);
        String targetedComponent = serviceUnit.getTargetComponentName();
        // Retrieve the component life cycle
        ComponentLifeCycle clc = managerService
                .getComponentByName(targetedComponent);
        if (clc == null) {
            msg = "The component ("
                    + targetedComponent
                    + ") is not deployed. The service unit cannot be undeployed.";
            log.error(msg);
            throw new Exception(msg);
        }
        // Retrieve the service unit manager of the component
        ServiceUnitManager sum = clc.getComponent().getServiceUnitManager();
        if (sum == null) {
            msg = "The service unit manager is null for the component ("
                    + targetedComponent
                    + "). You may need to initialise the component. The service units cannot be undeployed.";
            log.error(msg);
            throw new Exception(msg);
        }

        try {
            xmlReturn = sum.undeploy(suName, serviceUnitLifeCycle
                    .getServiceUnitRootPath());
        } catch (DeploymentException e) {
            msg = "Service unit manager of the component (" + targetedComponent
                    + ") failed to undeploy su (" + suName + ")";
            log.error(msg, e);
            throw new Exception(msg, e);
        }

        /*
         * Remove the service unit from the service assembly
         */
        saLifeCycle.getServiceUnitsLifeCycles().remove(suName);

        return xmlReturn;
    }

    /**
     * Check if a component task result is valid against an xml schema
     * 
     * @param componentTask
     *            xml of the component task
     * @param xsd
     *            an {@link InputStream} of the xsd file
     * @return true if valid, false if not valid
     */
    protected boolean isValidComponentTask(String componentTask, InputStream xsd) {
        return XMLResult.isValidComponentTask(componentTask, xsd);
    }

    public void undo(HashMap context) throws Exception {
        // Nothing to do

    }

}
