/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: InstallationServiceImpl.java 154 2006-03-27 15:30:10Z alouis $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.management.service;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.zip.ZipFile;

import javax.jbi.management.LifeCycleMBean;
import javax.jbi.management.MBeanNames;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.commons.io.FileUtils;
import org.objectweb.fractal.fraclet.annotation.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.Interface;
import org.objectweb.fractal.fraclet.annotation.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.LifeCycleType;
import org.objectweb.fractal.fraclet.annotation.Monolog;
import org.objectweb.fractal.fraclet.annotation.Provides;
import org.objectweb.fractal.fraclet.annotation.Requires;
import org.objectweb.petals.PetalsException;
import org.objectweb.petals.classloader.LoaderManager;
import org.objectweb.petals.jbi.component.context.ComponentContextImpl;
import org.objectweb.petals.jbi.component.context.ComponentInitialContext;
import org.objectweb.petals.jbi.component.lifecycle.ComponentLifeCycle;
import org.objectweb.petals.jbi.component.lifecycle.Installer;
import org.objectweb.petals.jbi.management.autoload.AutoLoaderImpl;
import org.objectweb.petals.jbi.management.systemstate.ComponentState;
import org.objectweb.petals.jbi.management.systemstate.SharedLibraryState;
import org.objectweb.petals.jbi.management.systemstate.SystemState;
import org.objectweb.petals.jbi.registry.ConsumerEndpoint;
import org.objectweb.petals.jbi.registry.Registry;
import org.objectweb.petals.jbi.routing.Router;
import org.objectweb.petals.repository.RepositoryService;
import org.objectweb.petals.tools.jbicommon.descriptor.ComponentDescription;
import org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor;
import org.objectweb.petals.tools.jbicommon.descriptor.SharedLibrary;
import org.objectweb.petals.tools.jbicommon.descriptor.SharedLibraryList;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.ParameterCheckHelper;
import org.objectweb.petals.util.StringHelper;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.petals.util.ZipUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * JBI installation service implementation.
 * 
 * @version $Rev: 154 $ $Date: 2006-03-27 17:30:10 +0200 (lun., 27 mars 2006) $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 * @author ofabre - EBM Websourcing
 */
@FractalComponent
@Provides(interfaces=@Interface(name="service",signature=org.objectweb.petals.jbi.management.service.InstallationService.class))
public class InstallationServiceImpl implements InstallationService {

    public static final String ERROR_INSTALL_SHARED_LIBRARY = "Could not install the shared library ";

    /**
     * The endpointService is used to manage all the endpoint interfaces in a
     * distributed context
     */
    @Requires(name="endpoint-service",signature=org.objectweb.petals.jbi.management.service.EndpointService.class)
    protected EndpointService endpointService;

    /**
     * JBI Container :: JMX Agent Admin
     */
    @Requires(name="lifecyclemanager",signature=org.objectweb.petals.jbi.management.service.LifeCycleManagerService.class)
    protected LifeCycleManagerService jmxAdmin;

    /**
     * Platform Component :: Loader Service
     */
    @Requires(name="classloader",signature=org.objectweb.petals.classloader.LoaderManager.class)
    protected LoaderManager loaderSrv;

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Monolog Logger instance.
     */
    @Monolog(name="logger")
    protected Logger logger;

    /**
     * Package Handler
     */
    protected PackageHandler packageHandler;

    /**
     * A {@link List} of loaded components.
     */
    protected List<String> componentLoadedCache;

    /**
     * A {@link List} of loaded shared libaries.
     */
    protected List<String> slLoadedCache;

    /**
     * JBI Container :: Recovery Service
     */
    @Requires(name="systemstate",signature=org.objectweb.petals.jbi.management.systemstate.SystemState.class)
    protected SystemState recoverySrv;

    /**
     * JBI Container :: Recovery Service
     */
    @Requires(name="registry",signature=org.objectweb.petals.jbi.registry.Registry.class)
    protected Registry registryService;

    /**
     * Platform Component :: Repository Service
     */
    @Requires(name="repository",signature=org.objectweb.petals.repository.RepositoryService.class)
    protected RepositoryService repositorySrv;

    /**
     * The container router instance.
     */
    @Requires(name="router",signature=org.objectweb.petals.jbi.routing.Router.class)
    protected Router router;

    protected InitialContext componentInitialContext;

    /**
     * Default constructor.
     */
    public InstallationServiceImpl() {

    }

    /**
     * @see javax.jbi.management.InstallationServiceMBean#installSharedLibrary(java.lang.String)
     *      Return ERROR_INSTALL_SHARED_LIBRARY if shared library can't be
     *      loaded correctly. It's an implementation choice !
     */
    public synchronized String installSharedLibrary(String slZipURL) {
        log.start();
        ParameterCheckHelper.isNullParameterWithLog(slZipURL, "slZipURL", log);
        ParameterCheckHelper.isEmptyParameterWithLog(slZipURL, "slZipURL", log);
        try {
            new URL(slZipURL);
        } catch (MalformedURLException e) {
            throw new IllegalArgumentException("slZipURL is not a valid URL");
        }

        String msg;
        String result = null;

        JBIDescriptor descriptor = null;
        String slName = null;
        URI archiveURI = null;

        try {
            /*
             * Transforming URL into URI
             */

            archiveURI = packageHandler.processAndGetPackageURI(slZipURL, true);

            /*
             * Loading descriptor and getting name
             */

            descriptor = packageHandler.loadDescriptor(archiveURI);

            /*
             * Checking shared lib package
             */

            checkSharedLibrary(descriptor, archiveURI);
        } catch (Throwable e) {
            msg = "Shared library package (" + slZipURL + ") isn't validated.";
            log.error(msg, e);
            throw new RuntimeException(msg);
        }
        try {
            slName = getSharedLibraryName(descriptor);
            /*
             * Exploding shared library into the installation repository
             */

            URI installationRoot = explodeSL(slName, archiveURI);

            /*
             * Performing shared library installation
             */

            result = installSharedLibrary(installationRoot, descriptor);

            /*
             * Creating state holder for shared lib
             */
            File file = createFileFromURI(archiveURI);
            File installedArchive = getInstalledArchive(file);
            recoverySrv.createSharedLibraryStateHolder(slName, installationRoot
                .toURL().toString(), installedArchive.toURL().toString());

            /*
             * Copy archive file to installed dir
             */
            copyFile(file, installedArchive);

            /*
             * Create success file in workdir
             */
            createSuccessFile(installedArchive);

            log.info("Shared library " + slName + " correctly installed !");
        } catch (Throwable e) {
            msg = ERROR_INSTALL_SHARED_LIBRARY + slZipURL;
            log.error(msg, e);

            // Reverse sl installation process in order to reput the system in
            // its
            // state before sl installation
            reverseSLInstallation(slName);

            throw new RuntimeException(msg);
        }

        log.end();

        return result;
    }

    /**
     * Expand the given package into the system repository
     * 
     * @param slName
     *            the unique identifier of the shared lib
     * @param archiveURI
     *            the shared lib zip archive URI
     * @return return installation URI of the expanded package, null if it fails
     * @throws ManagementException
     *             if the zip archive fails to deploy or if returned install
     *             root is null
     */
    protected URI explodeSL(String slName, URI archiveURI)
        throws ManagementException {
        String msg;
        URI installationRoot = null;

        try {
            installationRoot = repositorySrv.addSharedLibPackage(slName,
                archiveURI);

        } catch (PetalsException pe) {
            msg = "Unexpected error adding JBI installation "
                + "package into repository. ";
            log.error(msg, pe);
            throw new ManagementException(msg, pe);
        }

        if (installationRoot == null) {
            msg = "Unexpected Platform repository service error. The SL"
                + "installation root must NOT be null:  " + slName;
            log.error(msg);
            throw new ManagementException(msg);
        }
        return installationRoot;
    }

    /**
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.jbi.management.service.InstallationService#installSharedLibrary(java.net.URI,
     *      org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor)
     */
    public synchronized String installSharedLibrary(URI installationRoot,
        JBIDescriptor descriptor) throws PetalsException {
        log.start();
        ParameterCheckHelper.isNullParameterWithLog(installationRoot,
            "installationRoot", log);
        ParameterCheckHelper.isNullParameterWithLog(descriptor, "descriptor",
            log);
        String msg;

        /*
         * Get shared library production element
         */
        SharedLibrary sl = descriptor.getSharedLibrary();

        /*
         * Get classloader delegation mode
         */
        boolean parentFirst = isParentFirst(sl);

        /*
         * Create baseUrls Array for shared lib classloader
         */

        List<URL> urls = new ArrayList<URL>();

        try {
            urls.add(installationRoot.toURL());
        } catch (MalformedURLException e) {
            // simple do nothing - JBI Package has responsibility in
            // checking the archive contents against informations
            // contained in JBI Descriptor first.
            msg = "Unexpected error creating URL for shared lib classloader.";
            log.info(msg, e);
            throw new PetalsException(msg, e);
        }

        URL[] baseUrls = urls.toArray(new URL[] {});

        /*
         * Create shared lib classloader
         */
        String slName = sl.getIdentification().getName();
        try {
            loaderSrv.createSharedLibraryClassLoader(slName, baseUrls, sl
                .getSharedLibraryClassPath(), parentFirst);
        } catch (PetalsException pe) {
            msg = "Unexpected error creating class loader. ";
            log.error(msg, pe);
            throw new PetalsException(msg, pe);
        }

        /*
         * add shared library to the loaded cache list
         */
        slLoadedCache.add(slName);

        log.end();
        return slName;
    }

    /**
     * Check if the given sl is already loaded
     * 
     * @param slID
     * @return true if already loaded, false otherwise
     */
    public boolean isSLLoaded(String slID) {
        return slLoadedCache.contains(slID);
    }

    /**
     * Check if the given component is already loaded
     * 
     * @param compID
     * @return true if already loaded, false otherwise
     */
    public boolean isComponentLoaded(String compID) {
        return componentLoadedCache.contains(compID);
    }

    /**
     * @see javax.jbi.management.InstallationServiceMBean#loadInstaller(java.lang.String)
     */
    public synchronized ObjectName loadInstaller(String componentName) {
        ParameterCheckHelper.isNullParameterWithLog(componentName,
            "componentName", log);
        ParameterCheckHelper.isEmptyParameterWithLog(componentName,
            "componentName", log);
        ObjectName result = null;
        result = jmxAdmin.getInstallerByName(componentName);
        return result;
    }

    /**
     * @see javax.jbi.management.InstallationServiceMBean#loadNewInstaller(java.lang.String)
     *      Return a specific ObjectName if component can't be installed
     *      correctly. It's an implementation choice !
     * @throws RuntimeException
     *             if an error occurs during installation process. We have
     *             chosen this type of exception because, in the specification,
     *             this method doesn't return an exception
     */
    public synchronized ObjectName loadNewInstaller(String installZipURL) {
        log.start();
        ParameterCheckHelper.isNullParameterWithLog(installZipURL,
            "installZipURL", log);
        ParameterCheckHelper.isEmptyParameterWithLog(installZipURL,
            "installZipURL", log);
        try {
            new URL(installZipURL);
        } catch (Throwable e) {
            throw new IllegalArgumentException(
                "installZipUrl must be a legal url", e);
        }
        String msg = null;
        ObjectName result = null;
        JBIDescriptor descriptor = null;
        String componentName = null;
        URI archiveURI = null;

        try {
            /*
             * Transforming URL into URI
             */

            archiveURI = packageHandler.processAndGetPackageURI(installZipURL,
                true);

            /*
             * Loading descriptor and getting name
             */

            descriptor = packageHandler.loadDescriptor(archiveURI);

            /*
             * Checking component package
             */

            checkComponent(descriptor, archiveURI);

        } catch (Throwable e) {
            msg = "Component package (" + installZipURL + ") isn't validated.";
            log.error(msg, e);

            throw new RuntimeException(msg);
        }

        try {
            componentName = getComponentNameFromJBIDescriptor(descriptor);

            /*
             * Exploding component into the installation repository
             */

            URI installationRoot = explodeComponent(componentName, archiveURI);

            /*
             * Performing shared library installation
             */

            result = loadNewInstaller(installationRoot, descriptor);

            /*
             * Creating state holder for component
             */
            File file = createFileFromURI(archiveURI);
            File installedArchive = getInstalledArchive(file);
            recoverySrv.createComponentStateHolder(componentName,
                installationRoot.toURL().toString(), installedArchive.toURI()
                    .toString(), Installer.UNINSTALLED, LifeCycleMBean.UNKNOWN);

            /*
             * Copy archive file to installed dir
             */
            copyFile(file, installedArchive);

            /*
             * Create success file in workdir
             */
            createSuccessFile(installedArchive);

            log.info("Component " + componentName + " correctly installed ! ");
        } catch (Throwable e) {
            msg = "Component can't be installed: " + installZipURL;
            log.error(msg, e);

            // Reverse installation process in order to reput the system in its
            // state before component installation
            reverseComponentInstallation(componentName, result);

            throw new RuntimeException(msg);
        }
        log.end();

        return result;
    }

    /**
     * Expand the given package into the system repository
     * 
     * @param componentName
     *            the unique identifier of the component
     * @param archiveURI
     *            the component zip archive URI
     * @return return installation URI of the expanded package, null if it fails
     * @throws ManagementException
     *             if the zip archive fails to deploy or if returned install
     *             root is null
     */
    protected URI explodeComponent(String componentName, URI archiveURI)
        throws ManagementException {
        String msg;
        URI installationRoot = null;

        try {
            installationRoot = repositorySrv.addComponentPackage(componentName,
                archiveURI);

        } catch (PetalsException pe) {
            msg = "Unexpected error adding JBI installation "
                + "package into repository. ";
            log.error(msg, pe);
            throw new ManagementException(msg, pe);
        }

        if (installationRoot == null) {
            msg = "Unexpected Platform repository service error. The Component"
                + "installation root must NOT be null:  " + componentName;
            log.error(msg);
            throw new ManagementException(msg);
        }
        return installationRoot;
    }

    /**
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.jbi.management.service.InstallationService#loadNewInstaller(java.net.URI,
     *      org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor)
     */
    public synchronized ObjectName loadNewInstaller(URI installationRoot,
        JBIDescriptor descriptor) throws PetalsException {
        log.start();
        ParameterCheckHelper.isNullParameterWithLog(installationRoot,
            "installationRoot", log);
        ParameterCheckHelper.isNullParameterWithLog(descriptor, "descriptor",
            log);

        /*
         * Check installed shared lib for component : a shared lib must be
         * installed before a component can be installed that makes use of the
         * shared lib
         */
        checkInstalledSLForComp(descriptor);

        /*
         * Creating Component and Installation contexts
         */

        ComponentContextImpl compContextImpl = createComponentContext(
            descriptor, installationRoot);

        /*
         * Creating and registering installer mbean
         */

        ObjectName result = createAndRegisterInstallerMBean(compContextImpl,
            descriptor, installationRoot);

        /*
         * Adding component to the loaded cache list
         */

        componentLoadedCache.add(getComponentNameFromJBIDescriptor(descriptor));

        log.end();
        return result;
    }

    /**
     * @see javax.jbi.management.InstallationServiceMBean#uninstallSharedLibrary(java.lang.String)
     */
    public synchronized boolean uninstallSharedLibrary(String slName) {
        log.start();
        ParameterCheckHelper.isNullParameterWithLog(slName, "slName", log);
        ParameterCheckHelper.isEmptyParameterWithLog(slName, "slName", log);
        boolean result = false;

        /*
         * The slName must be non null and all components that makes use of the
         * given sl must be in the shutdown state
         */
        if (checkComponentShutdownStateForSL(slName)) {
            /*
             * unload shared library class loaders and clean cache
             */
            loaderSrv.deleteClassLoader(slName);

            slLoadedCache.remove(slName);

            /*
             * Remove shared lib from state holder file
             */
            SharedLibraryState slState = null;
            try {
                slState = recoverySrv.deleteSharedLibraryStateHolder(slName);
            } catch (Exception e) {
                String msg = "Shared library isn't completely uninstalled ! "
                    + "State holder isn't deleted";
                log.error(msg, e);
                return false;
            }

            /*
             * clean package installation from platform repository
             */
            try {
                result = repositorySrv.removeSharedLibPackage(slName);
            } catch (PetalsException e1) {
                String msg = "Shared library isn't completely uninstalled ! "
                    + "Failed to remove shared lib package.";
                log.error(msg, e1);
                return false;
            }

            /*
             * Move archive file to uninstalled dir
             */
            try {
                File installedArchive = null;
                if (slState != null) {
                    URI archiveURI = packageHandler.processAndGetPackageURI(
                        slState.getArchiveURL(), false);
                    installedArchive = createFileFromURI(archiveURI);
                }
                if (installedArchive != null && installedArchive.exists()) {
                    File uninstalledArchive = getUninstalledArchive(installedArchive);
                    installedArchive.renameTo(uninstalledArchive);
                    installedArchive = null;

                }
            } catch (IllegalArgumentException e) {
                log.error("Error getting URI from package url. ", e);
                return false;
            }
        }
        log.end();

        return result;
    }

    /**
     * TODO isToBeDeleted parameter isn't used for the moment : this method
     * completly deletes the component from the container TODO Refactoring
     * failure cases
     * 
     * @see javax.jbi.management.InstallationServiceMBean#unloadInstaller(java.lang.String,
     *      boolean)
     */
    public synchronized boolean unloadInstaller(String componentName,
        boolean isToBeDeleted) {
        log.start();
        ParameterCheckHelper.isNullParameterWithLog(componentName,
            "componentName", log);
        ParameterCheckHelper.isEmptyParameterWithLog(componentName,
            "componentName", log);
        boolean result = false;

        /*
         * Call uninstall on Installer
         */
        ObjectName instName = jmxAdmin.getInstallerByName(componentName);
        MBeanServer mbeanSrv = jmxAdmin.getMBeanServer();
        Object[] nullObjects = new Object[0];
        String[] nullStrings = new String[0];
        try {
            boolean isInstalled = ((Boolean) mbeanSrv.getAttribute(instName,
                "Installed")).booleanValue();
            if (isInstalled) {
                mbeanSrv
                    .invoke(instName, "uninstall", nullObjects, nullStrings);
            }
        } catch (Throwable e) {
            log.error("Error uninstalling component : " + componentName + "."
                + e.getMessage(), e);
            return false;
        }

        /*
         * unload installer from manager service
         */
        try {
            jmxAdmin.unregisterInstaller(instName);
        } catch (ManagementException e) {
            log.error("Error uninstalling component : " + componentName + "."
                + e.getMessage(), e);
            return false;
        }

        /*
         * Clean package loaded
         */
        componentLoadedCache.remove(componentName);

        /*
         * Remove component from state holder file
         */
        ComponentState componentState = null;
        try {
            componentState = recoverySrv
                .deleteComponentStateHolder(componentName);
        } catch (Exception e) {
            String msg = "Component isn't completely uninstalled ! "
                + "State holder isn't deleted";
            log.error(msg, e);
            return false;
        }

        /*
         * clean package installation from platform repository
         */

        try {
            result = repositorySrv.removeComponentPackage(componentName);
        } catch (PetalsException e1) {
            String msg = "Component isn't completely uninstalled ! "
                + "Failed to delete component package";
            log.error(msg, e1);
            return false;
        }

        /*
         * Move archive file to uninstalled dir
         */
        try {
            File installedArchive = null;
            if (componentState != null) {
                URI archiveURI = packageHandler.processAndGetPackageURI(
                    componentState.getArchiveURL(), false);
                installedArchive = createFileFromURI(archiveURI);

            }
            if (installedArchive != null && installedArchive.exists()) {
                File uninstalledArchive = getUninstalledArchive(installedArchive);
                installedArchive.renameTo(uninstalledArchive);
                installedArchive = null;
                /*
                 * FileUtils.copyFile(installedArchive, uninstalledArchive);
                 * FileUtils.forceDelete(installedArchive);
                 */

            }
            /*
             * } catch (IOException e) { log.error( "Error moving archive from
             * installed to uninstalled dir. " + e.getMessage(), e); return
             * false;
             */
        } catch (IllegalArgumentException e) {
            log.error("Error uninstalling component : " + componentName + "."
                + e.getMessage(), e);
            return false;
        }

        log.end();
        return result;
    }

    /**
     * Check the classpath element of the jbi descriptor
     * 
     * @param archive
     *            a ZipFile object mapping a real zip file entity
     * @param classPath
     *            a List of classpath elements to check
     * @throws PetalsException
     */
    protected void checkClassPathElements(ZipFile archive,
        List<String> classPath) throws PetalsException {
        String msg;

        for (String pathElement : classPath) {

            /*
             * empty class-path elements is not allowed
             */
            if (pathElement == null || "".equals(pathElement.trim())) {
                msg = "An empty class-path element found "
                    + "in the JBI Descriptor file (jbi.xml).";
                throw new PetalsException(msg);
            }

            /*
             * check if the specified path element exists within archive
             */
            if (!hasEntry(archive, pathElement)) {
                msg = "A class-path element " + pathElement + " declared in "
                    + "JBI Descriptor not found within installation "
                    + "package Zip archive file.";
                throw new PetalsException(msg);
            }
        }
    }

    /**
     * Check the validity of the component archive
     * 
     * @param descriptor
     *            the component <code>JBIDescriptor</code>
     * @param archiveURI
     *            URI representing the component archive
     * @return true if checking is completly ok and false otherwise
     * @throws PetalsException
     *             if component element or name is null or empty, if component
     *             is already installed or if some classpath elements of the
     *             component are missing
     */
    protected void checkComponent(JBIDescriptor descriptor, URI archiveURI)
        throws PetalsException {
        String msg = null;

        /*
         * check if jbi descriptor has a component element
         */

        ComponentDescription comp = descriptor.getComponent();
        if (comp == null) {
            msg = "The component production element was not found in "
                + "JBI Descriptor within installation package: "
                + archiveURI.getPath();
            log.error(msg);
            throw new PetalsException(msg);
        } else {

            /*
             * check if it is already loaded by package uuid
             */

            String compName = comp.getIdentification().getName();

            if (StringHelper.isEmpty(compName)) {
                msg = "The component name must be non null and non empty for the new component :"
                    + archiveURI.getPath();
                log.error(msg);
                throw new PetalsException(msg);
            }

            if (isComponentLoaded(compName)) {
                msg = "Other installation package already exists loaded in "
                    + "the component framework: " + compName;

                log.error(msg);
                throw new PetalsException(msg);
            } else {

                /*
                 * Check the classpath elements of the component
                 */

                ZipFile zipArchive = openZipFile(archiveURI);

                try {
                    checkClassPathElements(zipArchive, comp
                        .getBootstrapClassPath());
                    checkClassPathElements(zipArchive, comp
                        .getComponentClassPath());
                } finally {
                    try {
                        zipArchive.close();
                    } catch (IOException e) {
                        msg = "Component archive can't be closed";
                        log.error(msg);
                        throw new PetalsException(msg, e);
                    }
                }
            }
        }
    }

    /**
     * Check that all components that makes use of the given sl are in the
     * shutdown state.
     * 
     * @param slName
     *            the sl identifier
     * @return true if all components that makes use of the given sl are in the
     *         shutdown state, false otherwise
     */
    protected boolean checkComponentShutdownStateForSL(String slName) {
        boolean result = true;

        /*
         * Check for bindings
         */
        Collection<ComponentLifeCycle> bindings = jmxAdmin
            .getBindingCompoLifeCycles().values();
        for (Iterator iter = bindings.iterator(); iter.hasNext() && result;) {
            ComponentLifeCycle cycle = (ComponentLifeCycle) iter.next();

            /*
             * Retrieve the list of used sl
             */
            ComponentDescription description = cycle.getComponentDescription();
            List<SharedLibraryList> sls = description.getSharedLibraryList();

            if (sls != null) {
                for (Iterator iter2 = sls.iterator(); iter2.hasNext() && result;) {
                    SharedLibraryList sl = (SharedLibraryList) iter2.next();

                    /*
                     * If a used component isn't in the shutdown state, shared
                     * lib can't be uninstalled
                     */
                    if (slName.equals(sl.getName())
                        && !(LifeCycleMBean.SHUTDOWN.equals(cycle
                            .getCurrentState()))) {
                        log
                            .error("The shared lib can't be uninstalled because the following "
                                + "component isn't in the shutdown state : "
                                + description.getIdentification().getName());
                        result = false;
                    }
                }
            }
        }

        /*
         * Check for engines
         */
        Collection<ComponentLifeCycle> engines = jmxAdmin
            .getEngineCompoLifeCycles().values();
        for (Iterator iter = engines.iterator(); iter.hasNext() && result;) {
            ComponentLifeCycle cycle = (ComponentLifeCycle) iter.next();

            /*
             * Retrieve the list of used sl
             */
            ComponentDescription description = cycle.getComponentDescription();
            List<SharedLibraryList> sls = description.getSharedLibraryList();

            if (sls != null) {
                for (Iterator iter2 = sls.iterator(); iter2.hasNext() && result;) {
                    SharedLibraryList sl = (SharedLibraryList) iter2.next();

                    /*
                     * If a used component isn't in the shutdown state, shared
                     * lib can't be uninstalled
                     */
                    if (slName.equals(sl.getName())
                        && !(LifeCycleMBean.SHUTDOWN.equals(cycle
                            .getCurrentState()))) {
                        log
                            .error("The shared lib "
                                + slName
                                + " can't be uninstalled because"
                                + " the following component isn't in the shutdown state : "
                                + description.getIdentification().getName());
                        result = false;
                    }
                }
            }
        }

        return result;
    }

    /**
     * Check installed shared lib for component
     * 
     * @param descriptor
     *            the component <code>JBIDescriptor</code>
     * @throws PetalsException
     *             if a shared library is missing for the component being
     *             installed
     */
    protected void checkInstalledSLForComp(JBIDescriptor descriptor)
        throws PetalsException {
        List<SharedLibraryList> sls = descriptor.getComponent()
            .getSharedLibraryList();

        if (sls != null) {
            for (SharedLibraryList list : sls) {
                if (!slLoadedCache.contains(list.getName())) {
                    String msg = "Component depends on a not installed shared library : "
                        + list.getName();

                    log.error(msg);
                    throw new PetalsException(msg);
                }
            }
        }
    }

    /**
     * Check the validity of the shared lib archive
     * 
     * @param descriptor
     *            the shared lib <code>JBIDescriptor</code>
     * @param archiveURI
     *            URI representing the shared lib archive
     * @return true if checking is completly ok and false otherwise
     * @throws PetalsException
     * @throws ComponentCheckingException
     */
    protected void checkSharedLibrary(JBIDescriptor descriptor, URI archiveURI)
        throws PetalsException {
        String msg = null;

        /*
         * check if jbi descriptor has a shared-library element
         */

        SharedLibrary sl = descriptor.getSharedLibrary();
        if (sl == null) {
            msg = "The shared-library production element was not found in "
                + "JBI Descriptor within installation package: "
                + archiveURI.getPath();

            log.error(msg);
            throw new PetalsException(msg);
        } else {

            /*
             * check if it is already loaded by package uuid
             */

            String slName = sl.getIdentification().getName();

            if (isSLLoaded(slName)) {
                msg = "Other installation package already exists loaded in "
                    + "the component framework: " + slName;

                log.error(msg);

                throw new PetalsException(msg);
            } else {

                /*
                 * Check the classpath element of the shared library
                 */

                ZipFile zipArchive = openZipFile(archiveURI);

                try {
                    checkClassPathElements(zipArchive, sl
                        .getSharedLibraryClassPath());
                } finally {
                    try {
                        zipArchive.close();
                    } catch (IOException e) {
                        msg = "Shared library archive can't be closed";
                        log.error(msg, e);
                        throw new PetalsException(msg, e);
                    }
                }
            }

        }
    }

    /**
     * Copies a file from source to destination
     * 
     * @param src
     *            source file
     * @param dest
     *            destination file
     * @throws IOException
     *             if an error occurs during copy process
     */
    protected void copyFile(File src, File dest) throws IOException {
        FileUtils.copyFile(src, dest);
    }

    /**
     * Create and register a new InstallerMBean for the loaded component
     * 
     * @param componentContext
     *            the new Component execution context
     * @param installationContext
     *            the new Component installaion context
     * @param descriptor
     *            the component JBIDescriptor
     * @param installationRoot
     *            the URI representing the component installation folder
     * @return the new InstallerMBean name
     * @throws PetalsException
     */
    protected ObjectName createAndRegisterInstallerMBean(
        ComponentContextImpl componentContext, JBIDescriptor descriptor,
        URI installationRoot) throws PetalsException {
        log.start();
        ObjectName result = null;
        ComponentDescription componentDescription = descriptor.getComponent();
        try {
            MBeanNamesImpl mbeanNames = jmxAdmin.getMBeanNames();
            result = mbeanNames.createInstallerMBeanName(componentDescription
                .getIdentification().getName());
            
            Installer installerMBean = new Installer(installationRoot,
                loaderSrv, componentContext, componentDescription, jmxAdmin,
                recoverySrv, result, logger);

            jmxAdmin.registerInstaller(installerMBean);
        } catch (ManagementException e) {
            String msg = "Error creating installerMBean.";
            log.error(msg, e);
            throw new PetalsException(msg, e);
        }
        log.end();
        return result;
    }

    /**
     * Create a new ComponentContext for the loaded component with a new
     * DeliveryChannel
     * 
     * @param descriptor
     *            the component <code>JBIDescriptor</code>
     * @param installationRoot
     *            the component installation folder
     * @return a new <code>ComponentContext</code> implementation
     * @throws PetalsException
     *             if the creation of an initial context for components fails
     */
    protected ComponentContextImpl createComponentContext(
        JBIDescriptor descriptor, URI installationRoot) throws PetalsException {
        log.start();

        String componentName = descriptor.getComponent().getIdentification()
            .getName();

        String installationRootPath = new File(installationRoot)
            .getAbsolutePath();

        String workingRootPath = repositorySrv.getComponentWorkRoot(
            componentName).getAbsolutePath();

        MBeanServer beanServer = jmxAdmin.getMBeanServer();
        MBeanNames beanNames = jmxAdmin.getMBeanNames();

        ConsumerEndpoint address = new ConsumerEndpoint(componentName,
            SystemUtil.getContainerName());

        if (componentInitialContext == null) {
            try {
                componentInitialContext = new ComponentInitialContext(
                    registryService.getUsersContext());
            } catch (NamingException e) {
                String msg = "Error creating InitialContext for components";
                log.error(msg, e);
                throw new PetalsException(msg, e);
            }
        }

        ComponentContextImpl componentContextImpl = new ComponentContextImpl(
            descriptor, installationRootPath, workingRootPath, beanNames,
            beanServer, jmxAdmin, address, endpointService, logger, router,
            componentInitialContext);

        log.end();

        return componentContextImpl;
    }
    
    /**
     * Create a new File instance with the fileURI
     * 
     * @param fileURI
     *            URI of the file to create
     * @return a new File instance
     */
    protected File createFileFromURI(URI fileURI) {
        return new File(fileURI);
    }

    /**
     * Create a ".success" file in the work directory. The name of the new file
     * is the name of the "file" parameter with the suffix ".succes"
     * 
     * @param file
     *            file from which the name is computed
     * @throws IOException
     *             If an error occurs at the file creation
     */
    protected void createSuccessFile(File file) throws IOException {
        new File(AutoLoaderImpl.getWorkDirectory(), file.getName() + ".success")
            .createNewFile();
    }

    /**
     * Return the name of from the component description found in the descriptor
     * 
     * @param descriptor
     *            the JBI descriptor
     * @return the name of the component
     */
    protected String getComponentNameFromJBIDescriptor(JBIDescriptor descriptor) {
        return descriptor.getComponent().getIdentification().getName();
    }

    /**
     * Returns the shared library installedArchive
     * 
     * @param file
     *            the shared library file name
     * @return a File pointing at the shared library installed archive
     */
    protected File getInstalledArchive(File file) {
        return new File(AutoLoaderImpl.getInstalledDirectory(), file.getName());
    }

    /**
     * Get the sharedLibrary name from the JBI descriptor
     * 
     * @param descriptor
     *            the JBI descriptor
     * @return shared library name
     */
    protected String getSharedLibraryName(JBIDescriptor descriptor) {
        return descriptor.getSharedLibrary().getIdentification().getName();
    }

    /**
     * Returns the shared library uninstalledArchive
     * 
     * @param file
     *            the shared library file name
     * @return a File pointing at the shared library uninstalled archive
     */
    protected File getUninstalledArchive(File file) {
        return new File(AutoLoaderImpl.getUninstalledDirectory(), file
            .getName());
    }

    /**
     * Checks if the zip archive has one element with the pathElement name
     * 
     * @param archive
     *            the zip archive
     * @param pathElement
     *            the element to find
     * @return true if the element was found in the zip archive
     */
    protected boolean hasEntry(ZipFile archive, String pathElement) {
        return ZipUtil.hasEntry(archive, pathElement);
    }

    /**
     * Initialize the InstallationServiceImpl
     */
    @LifeCycle(on=LifeCycleType.START)
    protected void start() {
        log = new LoggingUtil(logger);
        log.start();

        /*
         * initialize entity loaded cache
         */
        slLoadedCache = new ArrayList<String>();
        componentLoadedCache = new ArrayList<String>();

        /*
         * Initialize package handler
         */
        packageHandler = new PackageHandler(logger, repositorySrv);

        log.end();
    }

    /**
     * @param sl
     * @return
     */
    protected boolean isParentFirst(SharedLibrary sl) {
        return JBIDescriptor.isParentFirst(sl.getClassLoaderDelegation());
    }

    /**
     * Get the zip archive at the archiveURI
     * 
     * @param archiveURI
     *            the URI of the zip archive
     * @return the zip archive
     * @throws PetalsException
     *             if the archive can not be found or opened
     */
    protected ZipFile openZipFile(URI archiveURI) throws PetalsException {
        return ZipUtil.openZipFile(archiveURI);
    }

    /**
     * Reverse component installation process in order to reput the system in
     * its state before component installation
     * 
     * @param componentName
     *            the component unique identifier
     * @param installerMBeanName
     *            the created installer MBean name
     */
    protected void reverseComponentInstallation(String componentName,
        ObjectName installerMBeanName) {
        if (!StringHelper.isEmpty(componentName)) {
            // Reverse installer MBean registration
            if (installerMBeanName != null) {
                try {
                    jmxAdmin.unregisterInstaller(installerMBeanName);
                } catch (ManagementException e) {
                    // Nothing to do just log
                    log
                        .warning("Component installer can't be unregistered or doesn't exist");
                }
            }

            // Delete component from already loaded component
            if (isComponentLoaded(componentName)) {
                componentLoadedCache.remove(componentName);
            }

            // Reverse state holder creation
            try {
                recoverySrv.deleteComponentStateHolder(componentName);
            } catch (Exception e) {
                // Nothing to do just log
                log
                    .warning("Component state can't be deleted or doesn't exist");
            }

            // Reverse component file system creation
            try {
                File compRoot = repositorySrv.getComponentRoot(componentName);
                if (compRoot.exists()) {

                    FileUtils.forceDelete(compRoot);
                }
            } catch (Exception e) {
                // Nothing to do just log
                log.warning("Component root can't be deleted or doesn't exist");
            }
        }
    }

    /**
     * Reverse sl installation process in order to reput the system in its state
     * before sl installation
     * 
     * @param slName
     *            the sl unique identifier
     */
    protected void reverseSLInstallation(String slName) {
        if (!StringHelper.isEmpty(slName)) {

            // Reverse classloader creation
            loaderSrv.deleteClassLoader(slName);

            // Delete sl from already loaded sl
            if (isComponentLoaded(slName)) {
                slLoadedCache.remove(slName);
            }

            // Reverse state holder creation
            try {
                recoverySrv.deleteSharedLibraryStateHolder(slName);
            } catch (Exception e) {
                // Nothing to do just log
                log
                    .warning("Shared lib state can't be deleted or doesn't exist");
            }

            // Reverse sl file system creation
            try {
                File compRoot = repositorySrv.getSharedLibRoot(slName);
                if (compRoot.exists()) {

                    FileUtils.forceDelete(compRoot);
                }
            } catch (Exception e) {
                // Nothing to do just log
                log
                    .warning("Shared lib root can't be deleted or doesn't exist");
            }
        }
    }

    /**
     * Stop the InstallationServiceImpl
     */
    @LifeCycle(on=LifeCycleType.STOP)
    public void stop() {
        log.start();

        slLoadedCache = null;
        componentLoadedCache = null;

        log.end();
    }

    /**
     * Shutdown the InstallationService. All shared libraries and components are
     * uninstalled
     */
    public void shutdown() throws Exception {
        // Uninstall all components
        List<String> removeList = new ArrayList<String>(componentLoadedCache);
        for (String compName : removeList) {
            unloadInstaller(compName, true);
        }

        // Uninstall all shared libs
        removeList = new ArrayList<String>(slLoadedCache);
        for (String slName : removeList) {
            uninstallSharedLibrary(slName);
        }

    }

}
