/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: SystemStateImpl.java 154 26 avr. 2006 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.management.systemstate;

import java.util.List;

import org.objectweb.fractal.fraclet.annotation.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.Interface;
import org.objectweb.fractal.fraclet.annotation.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.LifeCycleType;
import org.objectweb.fractal.fraclet.annotation.Monolog;
import org.objectweb.fractal.fraclet.annotation.Provides;
import org.objectweb.fractal.fraclet.annotation.Requires;
import org.objectweb.petals.jbi.management.deployment.DeploymentService;
import org.objectweb.petals.jbi.management.service.InstallationService;
import org.objectweb.petals.jbi.management.service.LifeCycleManagerService;
import org.objectweb.petals.jbi.management.service.ManagementException;
import org.objectweb.petals.repository.RepositoryService;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * Allows to persist component, shared library and service assembly states and
 * restore them after petals system shutdown or crash.
 * 
 * @author Olivier Fabre - EBM Websourcing
 */
@FractalComponent
@Provides(interfaces=@Interface(name="service",signature=org.objectweb.petals.jbi.management.systemstate.SystemState.class))
public class SystemStateImpl implements SystemState {

    /**
     * JBI Container :: Deployment Service
     */
	@Requires(name="deployment",signature=org.objectweb.petals.jbi.management.deployment.DeploymentService.class)
    protected DeploymentService deploymentService;

    /**
     * JBI Container :: Installation Service
     */
    @Requires(name="installation",signature=org.objectweb.petals.jbi.management.service.InstallationService.class)
    protected InstallationService installationService;

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Monolog Logger instance.
     */
    @Monolog(name="logger")
    protected Logger logger;

    /**
     * JBI Container :: JMX Agent Admin
     */
    @Requires(name="lifecyclemanager",signature=org.objectweb.petals.jbi.management.service.LifeCycleManagerService.class)
    protected LifeCycleManagerService managerService;

    /**
     * Platform Component :: Repository Service
     */
    @Requires(name="repository",signature=org.objectweb.petals.repository.RepositoryService.class)
    protected RepositoryService repositorySrv;

    protected StateStorage stateStorage;

    protected SystemRecovery systemRecovery;

    public ComponentState createComponentStateHolder(String componentName,
            String installURL, String zipURL, String installState,
            String lifecycleState) throws Exception {
        return stateStorage.createComponentStateHolder(componentName,
                installURL, zipURL, installState, lifecycleState);
    }

    public ServiceAssemblyState createServiceAssemblyStateHolder(String saName,
            String installURL, String zipURL, String lifecycleState)
        throws Exception {
        return stateStorage.createServiceAssemblyStateHolder(saName,
                installURL, zipURL, lifecycleState);
    }

    public SharedLibraryState createSharedLibraryStateHolder(String slName,
            String installURL, String zipURL) throws Exception {
        return stateStorage.createSharedLibraryStateHolder(slName, installURL,
                zipURL);
    }

    public ComponentState deleteComponentStateHolder(String componentName)
        throws Exception {
        return stateStorage.deleteComponentStateHolder(componentName);
    }

    public ServiceAssemblyState deleteServiceAssemblyStateHolder(String saName)
        throws Exception {
        return stateStorage.deleteServiceAssemblyStateHolder(saName);
    }

    public SharedLibraryState deleteSharedLibraryStateHolder(String slName)
        throws Exception {
        return stateStorage.deleteSharedLibraryStateHolder(slName);
    }

    public boolean recoverAllComponent() throws ManagementException {
        return systemRecovery.recoverAllComponent();
    }

    public boolean recoverAllEntities() throws ManagementException {
        log.start();
        boolean result = false;

        recoverAllSharedLibrary();
        recoverAllComponent();
        recoverAllServiceAssembly();
        result = true;

        log.end();
        return result;
    }

    public boolean recoverAllServiceAssembly() throws ManagementException {
        return systemRecovery.recoverAllServiceAssembly();
    }

    public boolean recoverAllSharedLibrary() throws ManagementException {
        return systemRecovery.recoverAllSharedLibrary();
    }

    public List<ComponentState> retrieveAllComponentStates() {
        return stateStorage.retrieveAllComponentStates();
    }

    public List<ServiceAssemblyState> retrieveAllServiceAssemblyStates() {
        return stateStorage.retrieveAllServiceAssemblyStates();
    }

    // ----------------------------------------------------------------------
    // Fractal BindingController implementation
    // ----------------------------------------------------------------------

    public List<SharedLibraryState> retrieveAllSharedLibraryStates() {
        return stateStorage.retrieveAllSharedLibraryStates();
    }

    // ----------------------------------------------------------------------
    // Fractal LifecycleController implementation
    // ----------------------------------------------------------------------

    public void updateComponentInstallationState(String componentName,
            String installState) throws Exception {
        stateStorage.updateComponentInstallationState(componentName,
                installState);

    }

    public void updateComponentLifeCycleState(String componentName,
            String lifecycleState) throws Exception {
        stateStorage.updateComponentLifeCycleState(componentName,
                lifecycleState);

    }

    public void updateServiceAssemblyState(String saName, String lifecycleState)
        throws Exception {
        stateStorage.updateServiceAssemblyState(saName, lifecycleState);

    }

    @LifeCycle(on=LifeCycleType.START)
    protected void start() {
        log = new LoggingUtil(logger);
        log.start();

        stateStorage = new StateStorage();

        try {
            stateStorage.init();
        } catch (Exception e) {
            String msg = "SystemState component can't be correctly initiated : "
                    + "StateStorage init failed !";
            log.error(msg, e);
            throw new RuntimeException(msg);
        }

        systemRecovery = new SystemRecovery(installationService, logger,
                deploymentService, managerService, stateStorage, repositorySrv);

        log.end();
    }

    @LifeCycle(on=LifeCycleType.STOP)
    protected void stop() {
        log.start();

        logger = null;

        log.end();
    }
}
