/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: MessageExchangeImpl.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.messaging;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.net.URI;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.petals.jbi.registry.ConsumerEndpoint;

/**
 * @author Adrien LOUIS - EBM WebSourcing
 * @author Gael BLONDELLE - EBM WebSourcing
 * @author Anass OUAZZANI - EBM WebSourcing
 */
public class MessageExchangeImpl implements MessageExchange, Cloneable, Serializable {

    public static final String IN_MSG = "in";

    public static final URI IN_ONLY_PATTERN = URI
        .create("http://www.w3.org/2004/08/wsdl/in-only");

    public static final URI IN_OPTIONAL_OUT_PATTERN = URI
        .create("http://www.w3.org/2004/08/wsdl/in-opt-out");

    public static final URI IN_OUT_PATTERN = URI
        .create("http://www.w3.org/2004/08/wsdl/in-out");

    public static final String OUT_MSG = "out";

    public static final URI ROBUST_IN_ONLY_PATTERN = URI
        .create("http://www.w3.org/2004/08/wsdl/robust-in-only");

    private static final short SERIALIZE_ROLE_CONSUMER = 0;

    private static final short SERIALIZE_ROLE_PROVIDER = 1;

    private static final short SERIALIZE_STATUS_ACTIVE = 0;

    private static final short SERIALIZE_STATUS_DONE = 1;

    private static final short SERIALIZE_STATUS_ERROR = 2;

    /**
     * 
     */
    private static final long serialVersionUID = 1L;

    protected ConsumerEndpoint consumerEndpoint;

    protected ServiceEndpoint endpoint;

    protected Exception error;

    protected String exchangeId;

    protected Fault fault;

    // //////////////////////////////////////////////////////////
    //
    // implementation specific attributes
    //
    // //////////////////////////////////////////////////////////

    protected QName interfaceName;

    protected Map<String, NormalizedMessage> messages = new HashMap<String, NormalizedMessage>();

    protected QName operation;

    protected URI pattern;

    protected Map<String, Object> properties = new HashMap<String, Object>();

    protected transient Role role;

    protected QName service;

    protected transient ExchangeStatus status;

    protected boolean terminated;

    protected boolean transacted;

    /**
     * Create a MessageExchangeImpl This constructor is called by a consumer who
     * wants to initialize an exchange.
     * 
     * @param consumerDeliveryChannel
     */
    public MessageExchangeImpl(ConsumerEndpoint consumerEndpoint) {

        status = ExchangeStatus.ACTIVE;

        role = Role.CONSUMER;

        this.consumerEndpoint = consumerEndpoint;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#createFault()
     */
    public Fault createFault() throws MessagingException {
        return new FaultImpl();
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#createMessage()
     */
    public NormalizedMessage createMessage() throws MessagingException {
        return new NormalizedMessageImpl();
    }

    public ConsumerEndpoint getConsumerEndpoint() {
        return consumerEndpoint;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getEndpoint()
     */
    public ServiceEndpoint getEndpoint() {
        return endpoint;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getError()
     */
    public Exception getError() {
        return error;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getExchangeId()
     */
    public String getExchangeId() {
        return exchangeId;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getFault()
     */
    public Fault getFault() {
        return fault;
    }

    public QName getInterfaceName() {
        return interfaceName;
    }

    /**
     * get the message associated to the reference. It is <b>Not</b> Case
     * Sensitive.
     * 
     * @see javax.jbi.messaging.MessageExchange#getMessage(java.lang.String)
     */
    public NormalizedMessage getMessage(String name) {
        if (name != null) {
            return (NormalizedMessage) messages.get(name.toLowerCase());
        } else {
            return null;
        }
    }

    public Map<String, NormalizedMessage> getMessages() {
        return messages;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getOperation()
     */
    public QName getOperation() {
        return operation;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getPattern()
     */
    public URI getPattern() {
        return pattern;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getProperty(java.lang.String)
     */
    public Object getProperty(String name) {
        return properties.get(name);
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getPropertyNames()
     */
    public Set getPropertyNames() {
        return properties.keySet();
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.messaging.MessageExchange#getRole()
     */
    public Role getRole() {
        return role;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getService()
     */
    public QName getService() {
        return service;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getStatus()
     */
    public javax.jbi.messaging.ExchangeStatus getStatus() {
        return status;
    }

    public boolean isTerminated() {
        return terminated;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#isTransacted()
     */
    public boolean isTransacted() {
        return transacted;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#setEndpoint(java.lang.Exception)
     */
    public void setEndpoint(ServiceEndpoint endpoint) {
        this.endpoint = endpoint;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#setError(java.lang.Exception)
     */
    public void setError(Exception error) {
        this.error = error;

        this.status = ExchangeStatus.ERROR;
    }

    /**
     * @param exchangeId
     *            The exchangeId to set.
     */
    public void setExchangeId(String exchangeId) {
        this.exchangeId = exchangeId;
    }

    // //////////////////////////////////////////
    // specific methods
    // //////////////////////////////////////////

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.messaging.MessageExchange#setFault(javax.jbi.messaging.Fault)
     */
    public void setFault(Fault fault) throws MessagingException {
        checkNotTerminated();

        if (role == null) {
            throw new MessagingException("The Role is not defined.");
        }

        // check that a fault has not already been sent
        if (this.fault != null) {
            throw new MessagingException("A fault has already been set.");
        }

        // check that the status has not be positionned with DONEor EXCEPTION
        if (ExchangeStatus.DONE.equals(status)
            || ExchangeStatus.ERROR.equals(status)) {
            throw new MessagingException(
                "The MessageExchange state and the current Role do not allow this operation.");
        }

        // only the provider can send a fault, except in the case of a
        // InOptionnalOut,
        // where the consumer can set a fault after the provider sent a response

        if (MessageExchange.Role.CONSUMER.equals(role)
            && !messages.containsKey(OUT_MSG)) {
            throw new MessagingException(
                "The MessageExchange state and the current Role do not allow this operation.");
        }

        this.fault = fault;
    }

    public void setInterfaceName(QName interfaceName) {
        this.interfaceName = interfaceName;
    }

    /**
     * Set the specified message. Check the availability to set the message with
     * the following rules: <br>
     * <li>CONSUMER Role <-> in reference or PROVIDER Role <-> out reference</li>
     * <li>JBI predefined MEP <-> in or out reference</li>
     * <li>setMessage(XXX,reference) has not already been called. <br>
     * This is <b>not</b> case sensitive
     * 
     * @see javax.jbi.messaging.MessageExchange#setMessage(javax.jbi.messaging.NormalizedMessage,
     *      java.lang.String)
     */
    public void setMessage(NormalizedMessage msg, String name)
        throws MessagingException {
        checkNotTerminated();

        if (msg == null) {
            throw new MessagingException("NormalizedMessage must be non null.");
        }
        if (name == null || name.trim().length() == 0) {
            throw new MessagingException(
                "The message reference must be non null and non empty.");
        }

        // check if the pattern allows this reference
        checkPatternMatching(name);

        // check Role
        checkRoleMatching(name);

        // check that the message has not already been sent
        if (messages.containsKey(name.toLowerCase())) {
            throw new MessagingException(
                "A message has already been set with the '" + name
                    + "' reference.");
        }

        // no exception was thrown, set the message
        messages.put(name.toLowerCase(), msg);
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#setOperation(javax.xml.namespace.QName)
     */
    public void setOperation(QName name) {
        this.operation = name;
    }

    /**
     * @param pattern
     *            The pattern to set.
     */
    public void setPattern(URI pattern) {
        this.pattern = pattern;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#setProperty(java.lang.String,
     *      java.lang.Object)
     */
    public void setProperty(String name, Object obj) {
        properties.put(name, obj);
    }

    public void setRole(Role role) {
        this.role = role;
    }

    /**
     * set Destination ServiceName. A destinsation Endpointreference is created
     * with only "serviceName" attribute
     */
    public void setService(QName service) {
        this.service = service;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#setStatus(javax.jbi.messaging.ExchangeStatus)
     */
    public void setStatus(ExchangeStatus status) throws MessagingException {
        checkNotTerminated();

        // check that DONE can be set.
        if (ExchangeStatus.DONE.equals(status)) {
            if (Role.CONSUMER.equals(role)) {
                if (IN_ONLY_PATTERN.equals(pattern)) {
                    throw new MessagingException(
                        "The MessageExchange state does not allow this operation.");
                } else if (IN_OUT_PATTERN.equals(pattern) && fault == null
                    && getMessage(OUT_MSG) == null) {
                    throw new MessagingException(
                        "The MessageExchange state does not allow this operation.");
                } else if (IN_OPTIONAL_OUT_PATTERN.equals(pattern)
                    && fault == null && getMessage(OUT_MSG) == null) {
                    throw new MessagingException(
                        "The MessageExchange state does not allow this operation.");
                }
            } else {
                if (ROBUST_IN_ONLY_PATTERN.equals(pattern) && fault != null) {
                    throw new MessagingException(
                        "The MessageExchange state does not allow this operation.");
                } else if (IN_OUT_PATTERN.equals(pattern)) {
                    throw new MessagingException(
                        "The MessageExchange state does not allow this operation.");
                } else if (IN_OPTIONAL_OUT_PATTERN.equals(pattern)) {
                    if (fault != null && getMessage(OUT_MSG) == null) {
                        throw new MessagingException(
                            "The MessageExchange state does not allow this operation.");
                    } else if (fault == null && getMessage(OUT_MSG) != null) {
                        throw new MessagingException(
                            "The MessageExchange state does not allow this operation.");
                    }
                }
            }
        }
        this.status = status;
    }

    public void setTerminated(boolean terminated) {
        this.terminated = terminated;
    }

    /**
     * @param transacted
     *            The transacted to set.
     */
    public void setTransacted(boolean transacted) {
        this.transacted = transacted;
    }

    public String toString() {
        return "MessageExchange@" + getExchangeId();
    }

    /**
     * Check if the exchange is still valid.
     * 
     * @throws MessagingException
     */
    protected void checkNotTerminated() throws MessagingException {
        if (isTerminated()) {
            throw new MessagingException("The Exchange is terminated.");
        }
    }

    /**
     * Check that the exchange MEP allows the specified reference. This only
     * works for the JBI predefined MEPs. It check that reference is IN or OUT.
     * Other reference or MEP will cause an exception
     * 
     * @param name
     * @throws MessagingException
     *             the MEP does not allow the reference
     */
    protected void checkPatternMatching(String name) throws MessagingException {
        if (pattern == null) {
            throw new MessagingException("The MEP is not defined.");
        }

        if (name == null) {
            throw new MessagingException("The reference name is not defined.");
        }

        if (IN_MSG.equalsIgnoreCase(name)) {
            if (IN_ONLY_PATTERN.equals(pattern)
                || ROBUST_IN_ONLY_PATTERN.equals(pattern)
                || IN_OUT_PATTERN.equals(pattern)
                || IN_OPTIONAL_OUT_PATTERN.equals(pattern)) {
                return;
            }
        } else if (OUT_MSG.equalsIgnoreCase(name)
            && ((IN_OUT_PATTERN.equals(pattern) || IN_OPTIONAL_OUT_PATTERN
                .equals(pattern)))) {
            return;
        }
        throw new MessagingException(
            "the MessageExchange state does not allow this operation.");
    }

    /**
     * Check that the Role allows the specified reference. This only works for
     * the JBI predefined MEPs. It check that reference is IN or OUT and the
     * Role is CONSUMER or PROVIDER. Other reference will cause an exception
     * 
     * @param name
     * @throws MessagingException
     *             the Role does not allow the reference
     */
    protected void checkRoleMatching(String name) throws MessagingException {
        if (role == null) {
            throw new MessagingException("the Role is not defined.");
        }

        if (name == null) {
            throw new MessagingException("The reference name is not defined.");
        }

        if (MessageExchange.Role.CONSUMER.equals(role)) {
            if (IN_MSG.equalsIgnoreCase(name)) {
                return;
            }
        } else if (MessageExchange.Role.PROVIDER.equals(role)
            && OUT_MSG.equalsIgnoreCase(name)) {
            return;
        }
        throw new MessagingException("The Role does not allow this operation.");
    }
    
    /**
     * Depending on the state of the Exchange,
     * clean IN and/or OUT normalized messages
     * to reduce exchange size.
     * Status DONE, ERROR : all messages are cleared
     * Fault message : all other messages are cleared
     * OUT message set : IN message is cleared
     */
    public void cleanMessages(){
        // if status is DONE or ERROR, do not send 
        // Normalized messages
        if(! ExchangeStatus.ACTIVE.equals(status)){
            messages.clear();
            fault=null;
        }
        // if there is a fault, no other messages are kept
        if(fault != null){
            messages.clear();
        }
        // if OUT response, IN msg is not kept
        if(messages.containsKey(OUT_MSG) && messages.containsKey(IN_MSG)){
            messages.remove(IN_MSG);
        }
    }

    /**
     * @see MessageExchangeImpl#readObject(ObjectInputStream)
     */
    protected void readObjectDelegate(ObjectInputStream s) throws IOException {
        // Role deserialization
        switch (s.readShort()) {
            case 0:
                role = MessageExchange.Role.CONSUMER;
                break;
            case 1:
                role = MessageExchange.Role.PROVIDER;
                break;
            default:
                break;
        }

        // Status deserialization
        switch (s.readShort()) {
            case 0:
                status = ExchangeStatus.ACTIVE;
                break;
            case 1:
                status = ExchangeStatus.DONE;
                break;
            case 2:
                status = ExchangeStatus.ERROR;
                break;
            default:
                break;
        }

        try {
            s.defaultReadObject();
        } catch (ClassNotFoundException e) {
            throw new IOException(e.getClass() + ":" + e.getMessage());
        }
    }

    /**
     * @see MessageExchangeImpl#writeObject(ObjectOutputStream)
     */
    protected void writeObjectDelegate(ObjectOutputStream s) throws IOException {
        // Role serialization
        if (MessageExchange.Role.CONSUMER.equals(role)) {
            s.writeShort(SERIALIZE_ROLE_CONSUMER);
        } else {
            s.writeShort(SERIALIZE_ROLE_PROVIDER);
        }

        // Status serialization
        if (ExchangeStatus.ACTIVE.equals(status)) {
            s.writeShort(SERIALIZE_STATUS_ACTIVE);
        } else if (ExchangeStatus.DONE.equals(status)) {
            s.writeShort(SERIALIZE_STATUS_DONE);
        } else {
            s.writeShort(SERIALIZE_STATUS_ERROR);
        }
        s.defaultWriteObject();
    }

    /**
     * Used to deserialize Status and Role objects
     * 
     * @param s
     * @throws IOException
     */
    private void readObject(ObjectInputStream s) throws IOException {
        readObjectDelegate(s);
    }

    /**
     * Used to serialize Status and Role objects
     * 
     * @param s
     * @throws IOException
     */
    private void writeObject(ObjectOutputStream s) throws IOException {
        writeObjectDelegate(s);
    }
    
    /**
     * Used to clone all fields of MessageExchangeImpl except :
     * <br/> the NormalizedMessages contained in the Map of messages
     * <br/> the Fault object
     * <br/> qnames...
     */
    public synchronized MessageExchangeImpl clone(){
        MessageExchangeImpl msg = null;
        try{
            msg = (MessageExchangeImpl) super.clone();
            msg.consumerEndpoint = (ConsumerEndpoint)consumerEndpoint.clone();
        } catch (CloneNotSupportedException e) {
            msg = null;
        }
        return msg;        
    }
}
