/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: RouterImpl.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.routing;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;

import org.objectweb.fractal.fraclet.annotation.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.Interface;
import org.objectweb.fractal.fraclet.annotation.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.LifeCycleType;
import org.objectweb.fractal.fraclet.annotation.Monolog;
import org.objectweb.fractal.fraclet.annotation.Provides;
import org.objectweb.fractal.fraclet.annotation.Requires;
import org.objectweb.petals.jbi.component.context.ComponentContextImpl;
import org.objectweb.petals.jbi.component.lifecycle.ComponentLifeCycle;
import org.objectweb.petals.jbi.management.service.EndpointService;
import org.objectweb.petals.jbi.management.service.LifeCycleManagerService;
import org.objectweb.petals.jbi.messaging.MessageExchangeImpl;
import org.objectweb.petals.jbi.registry.AbstractEndpoint;
import org.objectweb.petals.jbi.registry.ConsumerEndpoint;
import org.objectweb.petals.jbi.registry.InternalEndpoint;
import org.objectweb.petals.jbi.transport.TransportException;
import org.objectweb.petals.jbi.transport.Transporter;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * Routes messages to their destinations. Performs an itinerary resolution,
 * adding for example a transformatino service before delivering the message to
 * its recipient, and perform an address resolution, finding the endpoint
 * satisfying the the destination information (finding an endpoint for a
 * specified service). Resolutions are based on defined rules.
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 */
@FractalComponent
@Provides(interfaces=@Interface(name="service",signature=org.objectweb.petals.jbi.routing.Router.class))
public class RouterImpl implements Router {

	@Requires(name="endpoint-service",signature=org.objectweb.petals.jbi.management.service.EndpointService.class)
	protected EndpointService endpointService;

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    @Monolog(name="logger")
    protected Logger logger;

    @Requires(name="lifecyclemanager",signature=org.objectweb.petals.jbi.management.service.LifeCycleManagerService.class)
    protected LifeCycleManagerService manager;

    @Requires(name="transporter",signature=org.objectweb.petals.jbi.transport.Transporter.class)
    protected Transporter transporter;

    protected AddressResolver addressResolver;

    // ----------------------------------------------------------------------
    // Fractal BindingController implementation
    // ----------------------------------------------------------------------

    /**
     * Sends the specified message. Depending on the Role of the sender, send
     * the message to the producer (request) or to the consumer (response)
     * 
     * @param source
     * @param exchange
     * @param timeOut
     * @throws RoutingException
     */
    public void send(ComponentContextImpl source, MessageExchangeImpl exchange,
        long timeOut) throws RoutingException {
        log.call();

        if (MessageExchange.Role.CONSUMER.equals(exchange.getRole())) {
            sendToProvider(source, exchange, timeOut);
        } else {
            sendToConsumer(source, exchange, timeOut);
        }
    }

    // ----------------------------------------------------------------------
    // Fractal LifecycleController implementation
    // ----------------------------------------------------------------------

    @LifeCycle(on=LifeCycleType.START)
    protected void start() {
        log = new LoggingUtil(logger);
        log.start();
        addressResolver = new AddressResolver(endpointService);
        log.end();
    }

    @LifeCycle(on=LifeCycleType.STOP)
    protected void stop() {
        log.start();

        logger = null;

        log.end();
    }

    /**
     * Sends the specified message, sent by the specified component. TODO send
     * directly to the DeliveryChannel of the component related to the recipient
     * endpoint.
     * 
     * @param source
     * @param exchange
     * @param timeOut
     * @throws RoutingException
     */
    protected void sendToConsumer(ComponentContextImpl source,
        MessageExchangeImpl exchange, long timeOut) throws RoutingException {
        log.start();

        try {
            AbstractEndpoint consumerEndpoint = exchange.getConsumerEndpoint();

            String containerName = consumerEndpoint.getContainerName();

            exchange.setRole(MessageExchange.Role.CONSUMER);

            transporter.send(exchange, containerName, timeOut);
        } catch (TransportException e) {
            // restore the Role if the message has not been sent
            exchange.setRole(MessageExchange.Role.PROVIDER);

            throw new RoutingException(e);
        }

        log.end();
    }

    /**
     * Sends the specified message to the provider of the service. TODO for the
     * moment, send directly to the DeliveryChannel of the component related to
     * the recipient endpoint.
     * 
     * @param source
     * @param exchange
     * @throws RoutingException
     */
    protected void sendToProvider(ComponentContextImpl source,
        MessageExchangeImpl exchange, long timeOut) throws RoutingException {
        log.start();

        // resolve the address (find endpoint)
        addressResolver.resolveAddress(source, exchange);

        try {
            AbstractEndpoint providerEndpoint = (AbstractEndpoint) exchange
                .getEndpoint();

            String containerName = providerEndpoint.getContainerName();

            exchange.setRole(MessageExchange.Role.PROVIDER);

            transporter.send(exchange, containerName, timeOut);
        } catch (TransportException e) {
            // restore the Role if the message has not been sent
            exchange.setRole(MessageExchange.Role.CONSUMER);

            // TODO set message to Error ?
            throw new RoutingException(e);
        }

        log.end();
    }

    /**
     * route the received exchange to the correct DeliveryChannel.
     */
    public void receive(MessageExchangeImpl exchange) throws RoutingException {
        log.start();

        // find the name of the component to deliver the Exchange.
        // Depends on the Role (consumer or provider)
        String componentname = null;

        if (MessageExchange.Role.PROVIDER.equals(exchange.getRole())) {
            componentname = ((InternalEndpoint) exchange.getEndpoint())
                .getComponentName();
        } else {
            componentname = ((ConsumerEndpoint) exchange.getConsumerEndpoint())
                .getComponentName();
        }

        // retrieve the ComponentLifeCycle
        ComponentLifeCycle comp = manager.getComponentByName(componentname);

        // push the exchange
        try {
            comp.getComponentContext().getDeliveryChannelImpl().push(exchange);
        } catch (JBIException e) {
            throw new RoutingException(e);
        }
        log.end();
    }
}
