/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: JoramTransporter.java,v 0.3 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.transport;

import javax.jbi.JBIException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.ObjectMessage;
import javax.naming.InitialContext;

import org.objectweb.fractal.fraclet.annotation.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.Interface;
import org.objectweb.fractal.fraclet.annotation.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.LifeCycleType;
import org.objectweb.fractal.fraclet.annotation.Provides;
import org.objectweb.fractal.fraclet.annotation.Requires;
import org.objectweb.petals.jbi.messaging.MessageExchangeImpl;
import org.objectweb.petals.jbi.routing.Router;
import org.objectweb.petals.jbi.routing.RoutingException;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.PropertyUtil;
import org.objectweb.petals.util.SystemUtil;

/**
 * Implementation of the <code>Transporter</code> with JORAM. <br>
 * This implementation uses the fractal JoramAgent. <br>
 * Each <code>JoramTransporter</code> registrers a <code>Topic</code> in the
 * Joram distributed JNDI with the following name :
 * <em>JoramConnection-[containerId]</em> <br>
 * The name of a container is the <em>ID</em> specified in the
 * <em>conf/server.properties</em> file. <br>
 * This <code>JoramTransporter</code> uses an <code>ObjectSerializer</code>
 * to create <em>jms.ObjectMessage</em> to send via JMS. So, the
 * <code>MessageExchange</code> to send is serialized.
 * 
 * @version $Rev: 477 $ $Date: 2006-05-29 17:18:07 +0200 (lun., 29 mai 2006) $
 * @since Petals 1.0
 * @author alouis
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.objectweb.petals.jbi.transport.Transporter.class))
public class JoramTransporter implements Transporter {

	protected String host;

	protected int id;

	protected JoramConnection joramConnection;

	protected JoramAgent joramAgent;

	protected Serializer serializer = new ObjectSerializer();

	protected LoggingUtil log;

	protected String pwd;

	@Requires(name = "router", signature = org.objectweb.petals.jbi.routing.Router.class)
	protected Router router;

	protected int tcp;

	protected String user;

	/**
	 * Send the message to the specified container. The send can be local or
	 * distant. If the "org.objectweb.petals.transport.nonpersistent" exchange
	 * property is set to true, the message is not persisted by the Transport
	 * layer, but it can be lost in case of problem.
	 */
	public void send(MessageExchangeImpl messageExchange, String containerName,
			long timeToLive) throws TransportException {

		if (id == Integer.parseInt(containerName)) {
			sendLocal(messageExchange);

		} else {
			// check if the message has to be persisted
			Object noPersist = messageExchange
					.getProperty(PROPERTY_NOPERSISTANCE);
			boolean persist = (noPersist == null || !noPersist.toString()
					.toLowerCase().equals("true"));
			sendDistant(messageExchange, containerName, timeToLive, persist);
		}
	}

	/**
	 * init variables with joram.properties file and create/bind the topic for
	 * this container. A listener is created and started to listen incoming
	 * messages
	 * 
	 * @throws Exception
	 */
	protected void startJMS() throws Exception {
		log.start();

		// create the Joram Server
		joramAgent = new JoramAgent(log);
		joramAgent.startServer();

		// id = Integer.parseInt(PropertyUtil.getProperty(
		// PropertyUtil.SERVER_PROPS, "joram.id"));
		id = Integer.parseInt(SystemUtil.getContainerName());

		tcp = Integer.parseInt(SystemUtil.getJoramTCPPort());
		user = SystemUtil.getJoramUser();
		pwd = SystemUtil.getJoramPassword();
		host = SystemUtil.getHost();

		// create initialContext
		InitialContext context = new javax.naming.InitialContext(PropertyUtil
				.retrieveJNDIProperties());

		joramConnection = new JoramConnection(id, tcp, user, pwd, host, context);

		log.end();
	}

	public void startListening() throws TransportException {
		log.start();

		MessageListener aML = new MessageListener() {
			public void onMessage(Message aMessage) {
				processIncomingMessage(aMessage);
			}
		};
		try {
			joramConnection.start(aML);
		} catch (Exception e) {
			throw new TransportException(e);
		}

		log.end();
	}

	/**
	 * TODO manage exceptions This method is called asynchronously when a
	 * message is received from Joram, via the topic listener.
	 * 
	 * The recevied message is transform into a <code>MessageExchangeImpl</code>
	 * by using the <code>Serializer</code>, and is sent to the
	 * <code>Router</code> by calling <code>receive()</code>.
	 * 
	 * @param msg
	 *            the jms message TODO processIncomingMessage :manage the
	 *            reception of unknown message type. PB between the calculation
	 *            of the next recipient and notion of service provider/consumer
	 */
	protected void processIncomingMessage(Message msg) {
        log.start();
        if (msg instanceof ObjectMessage) {
            try {
                MessageExchangeImpl jbiMsg = serializer.jms2jbi(msg);

                // transfer the message to the router
                router.receive(jbiMsg);
                
                // ack the message
                msg.acknowledge();
            } catch (Exception e) {
                log.error(e.getMessage(), e);
            }
        } else {
            log.error("the received JMS message is not an objectmessage");
        }
        log.end();
    }

	/**
	 * Send the message to the specified distant container, by serializing the
	 * exhange with a <code>Serializer</code>. We retrieve the related
	 * distant <code>Topic</code> with the namle of the container, transform
	 * the <code>MessageExchangeImpl</code> into a <code>jms.Message</code>,
	 * and send this message onto the distant topic.
	 * 
	 * @param messageExchange
	 *            must be non null
	 * @param containerName
	 *            must be non null
	 * @param timeToLive
	 * @param persist
	 * @throws TransportException
	 *             container not found, invalid message.
	 */
	protected void sendDistant(MessageExchangeImpl messageExchange,
			String containerName, long timeToLive, boolean persist)
			throws TransportException {
		log.start();
		try {
			joramConnection.sendTo(containerName, messageExchange, timeToLive,
					persist);
		} catch (TransportException e) {
			log.error(e.getMessage(), e);
			throw e;
		}
		log.end();
	}

	/**
	 * init variables with joram.properties file and create/bind the topic for
	 * this container. A listener is created and started to listen incoming
	 * messages
	 * 
	 * @throws Exception
	 */
	@LifeCycle(on = LifeCycleType.START)
	protected void start() throws Exception {
		log = new LoggingUtil(null);
		log.start();

		// create the Joram Server
		joramAgent = new JoramAgent(log);
		joramAgent.startServer();

		// id = Integer.parseInt(PropertyUtil.getProperty(
		// PropertyUtil.SERVER_PROPS, "joram.id"));
		id = Integer.parseInt(SystemUtil.getContainerName());

		tcp = Integer.parseInt(SystemUtil.getJoramTCPPort());
		user = SystemUtil.getJoramUser();
		pwd = SystemUtil.getJoramPassword();
		host = SystemUtil.getHost();

		// create initialContext
		InitialContext context = new javax.naming.InitialContext(PropertyUtil
				.retrieveJNDIProperties());

		joramConnection = new JoramConnection(id, tcp, user, pwd, host, context);

		log.end();
	}

	/**
	 * Send the message locally. No serialization of the message
	 * 
	 * @param messageExchange
	 *            a messageExchange, can be null
	 * @throws TransportException
	 *             problem with the <code>Router</code>
	 */
	protected void sendLocal(MessageExchangeImpl messageExchange)
			throws TransportException {
		log.start();
		try {
			router.receive(messageExchange);
		} catch (RoutingException e) {
			String msg = "Can not send locally the MessageExchange to the local Router.";
			log.error(msg, e);
			throw new TransportException(msg, e);
		}
		log.end();
	}

	/**
	 * close the JMS connection (but does not close the joram server)
	 * 
	 * @throws JBIException
	 */
	@LifeCycle(on = LifeCycleType.STOP)
	public void stop() throws TransportException {
		log.start();
		try {
			joramConnection.stop();
			joramAgent.stopServer();
		} catch (Exception e) {
			String msg = "Can not stop the JMS connection.";
			log.error(msg, e);
			throw new TransportException(msg, e);
		}
		log.end();
	}

	/**
	 * Shutdown the Transporter. The related Topic is removed from Joram and
	 * from the distributed Jndi server
	 */
	public void shutdown() throws TransportException {
		try {
			joramConnection.unsubscribe();
		} catch (Exception e) {
			String msg = "Can not unreference the Transporter from the Registry.";
			log.error(msg, e);
			throw new TransportException(msg, e);
		}
		try {
			joramAgent.shutdown();
		} catch (Exception e) {
			String msg = "Can not shutdown correctly the Joram server.";
			log.error(msg, e);
			throw new TransportException(msg, e);
		}

	}
}