/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: DistributedJMXServerManager.java 16:34:49 ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.kernel.admin;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.naming.NotContextException;

import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.Interface;
import org.objectweb.fractal.fraclet.annotation.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.LifeCycleType;
import org.objectweb.fractal.fraclet.annotation.Provides;
import org.objectweb.petals.PetalsException;
import org.objectweb.petals.util.JNDIUtil;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.PropertyUtil;

/**
 * This class is a manager of JMX server. It is used to get a connection to a
 * JMX Server on a distant container.
 * 
 * @author ddesjardins - eBMWebsourcing
 */
@FractalComponent
@Provides(interfaces=@Interface(name="service",signature=org.objectweb.petals.kernel.admin.DistributedJMXServerFactory.class))
public class DistributedJMXServerFactoryImpl 
	implements DistributedJMXServerFactory {

    /**
     * Name of the container context
     */
    private static final String CONTAINERS_REF = "containers";

    /**
     * Information of the current container
     */
    protected ContainerInformation containerInformation;

    /**
     * Container context
     */
    protected Context containersContext;

    /**
     * Logger
     */
    protected LoggingUtil log;

    /**
     * Root context
     */
    protected InitialContext rootContext;

    /**
     * Create a
     * 
     * @param containerName
     * @return
     * @throws PetalsException
     */
    public DistributedJMXServer createDistributedJMXServer(String containerName)
        throws PetalsException {
        log.start();
        DistributedJMXServer distributedJMXServer = null;
        try {
            ContainerInformation containerInformation = (ContainerInformation) containersContext
                .lookup(containerName);
            if (containerInformation != null) {
                JMXConnector connector = null;
                Properties env = System.getProperties();
                env.put("java.naming.factory.initial",
                    "com.sun.jndi.rmi.registry.RegistryContextFactory");
                Map<String, Object> args = new HashMap<String, Object>();
                String username = containerInformation.getJmxLogin();
                String password = containerInformation.getJmxPassword();
                if (!"".equals(username)) {
                    if (password == null) {
                        password = "";
                    }
                    String[] credentials = new String[] {username, password};
                    args.put("jmx.remote.credentials", credentials);
                }
                String urlStr = "service:jmx:rmi:///jndi/rmi://"
                    + containerInformation.getHost() + ":"
                    + containerInformation.getJmxPort()
                    + "/management/rmi-jmx-connector";
                JMXServiceURL url;

                url = new JMXServiceURL(urlStr);
                connector = JMXConnectorFactory.newJMXConnector(url, args);

                distributedJMXServer = new DistributedJMXServer(connector);
            }
        } catch (Exception e) {
            throw new PetalsException(
                "Error while trying to connect to JMX server", e);
        }
        log.end();
        return distributedJMXServer;
    }

    /**
     * @see DistributedJMXServerFactory#isContainerStarted(String)
     */
    public boolean isContainerStarted(String containerName) {
        log.start();
        if (containerName == null) {
            throw new IllegalArgumentException(
                "The container name must be non null");
        }
        boolean isStarted = false;
        if (JNDIUtil.isBound(rootContext, CONTAINERS_REF, containerName)) {
            try {
                ContainerInformation containerInformation = (ContainerInformation) containersContext
                    .lookup(containerName);
                isStarted = containerInformation.isStarted();
            } catch (NamingException e) {
                // Do nothing
            }
        }
        log.end();
        return isStarted;
    }

    @LifeCycle(on=LifeCycleType.START)
    public void start() throws IllegalLifeCycleException {
    	log=new LoggingUtil(null);
        log.start();
        try {
            initContext();
            
        } catch (PetalsException e) {
            log.error("Error starting JMX server", e);
        }
        log.end();
    }

    @LifeCycle(on=LifeCycleType.STOP)
    public void stop() throws IllegalLifeCycleException {
        log.call();
    }

    /**
     * Initialize the containers context
     * 
     * @throws PetalsException
     */
    protected void initContext() throws PetalsException {
        try {
            rootContext = new InitialContext(PropertyUtil
                .retrieveJNDIProperties());
            if (!JNDIUtil.isBound(rootContext, "/", CONTAINERS_REF)) {
                containersContext = rootContext
                    .createSubcontext(CONTAINERS_REF);
            } else {
                containersContext = (Context) rootContext
                    .lookup(CONTAINERS_REF);
            }
            if (containersContext == null) {
                throw new NotContextException();
            }
        } catch (Exception e1) {
            throw new PetalsException(
                "Problem while trying to initialize the container context", e1);
        }
    }

}
