/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: PetalsAdminServiceMBean.java 16:40:52 alouis $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.kernel.admin.service;

import java.util.Collection;
import java.util.List;
import java.util.Map;


/**
 * This class is a manager of JMX server. It is used to get a connection to a
 * JMX Server on a distant container.
 * 
 * @author ddesjardins - eBMWebsourcing
 */
public interface PetalsAdminServiceMBean {

    public static final String CONF_HOST="host";
    public static final String CONF_HTML_PORT="htmlPort";
    public static final String CONF_JMX_LOGIN="jmxLogin";
    public static final String CONF_JMX_PASSWORD="jmxPassword";
    public static final String CONF_JMX_PORT="jmxPort";
    public static final String CONF_JNDI_FACTORY="jndiFactory";
    public static final String CONF_JNDI_PORT="jndiPort";
    public static final String CONF_JORAM_DOMAIN="joramDomain";
    public static final String CONF_JORAM_DOMAIN_PORT="joramDomainPort";
    public static final String CONF_JORAM_ID="joramId";
    public static final String CONF_JORAM_LOGIN="joramLogin";
    public static final String CONF_JORAM_PASSWORD="joramPassword";
    public static final String CONF_JORAM_TCP_PORT="joramTCPPort";
    public static final String CONF_NAME="name";
    public static final String CONF_STATE="state";
    public static final String CONF_UID="uid";
    
    /**
     * Unload the container from the Petals global environment.
     * All ServiceAssemblies, Components and Share Libraries are
     * stopped, shutdown, and uninstalled.
     * All registered Endpoints are deactivated.
     * Then the container is removed from the domain.
     * @throws Exception
     */
    void shutdownContainer() throws Exception;
    
    /**
     * Stop the container.
     * The container is still visible in the Petals global environment, 
     * but marked as 'stopped'.
     * The registered Endpoints are still visible,too. 
     * @throws Exception
     */
    void stopContainer() throws Exception;
   
    /**
     * Retrieve the collection of all the Started servers configuration from this
     * domain. keys are :<br>
     * - "host"<br>
     * - "htmlPort"<br>
     * - "jmxLogin"<br>
     * - "jmxPassword"<br>
     * - "jmxPort"<br>
     * - "joramDomain"<br>
     * - "joramDomainPort"<br>
     * - "joramTCPPort"<br>
     * - "joramId"<br>
     * - "joramLogin"<br>
     * - "joramPassword"<br>
     * - "name"<br>
     * - "state"<br>
     * - "uid"<br>
     * 
     * @return Collection of started server information (Map key;value)
     */
    Collection<Map<String,String>> getAllStartedServersConfiguration();
    
    /**
     * Retrieve the collection of all the servers configuration from this
     * domain. keys are :<br>
     * - "host"<br>
     * - "htmlPort"<br>
     * - "jmxLogin"<br>
     * - "jmxPassword"<br>
     * - "jmxPort"<br>
     * - "joramDomain"<br>
     * - "joramDomainPort"<br>
     * - "joramTCPPort"<br>
     * - "joramId"<br>
     * - "joramLogin"<br>
     * - "joramPassword"<br>
     * - "name"<br>
     * - "state"<br>
     * - "uid"<br>
     * 
     * @return Collection of server information (Map key;value)
     */
    Collection<Map<String,String>> getAllServersConfiguration();

    /**
     * Remove the specified server from the network configuration. <br>
     * The server is unbound, but not shutdown. All ServiceEndpoints activated
     * on this server are unbound. <br>
     * <b>Be very careful when using this operation, the network configuration
     * may become unstable.</b><br>
     * TODO JORAM configuration not updated
     * @param serverName
     * @throws Exception
     *             The server is not found in the network configuration, some
     *             Endpoints have not been removed, or problem while accessing
     *             the network configuration
     */
   void removeServerFromNetwork(String serverName) throws Exception;

   /**
    * shows if the monitoring is started or stopped.
    * @return
    */
   boolean getMonitoring();
   
   /**
    * start or stop the monitoring.
    * the monitoring information is reset when state is set to FALSE
    * @param state
    */
   void setMonitoring(boolean activate) throws Exception;
   
   /**
    * Return information about messages exchanges since the specified date.<br>
    * Information are sort by exchange-start-date.
    * Information is returned in a bulk mode, with no specific Java Object. <br>
    * - Each element of the list represents an exchange report.<br>
    * &nbsp;&nbsp;- an exchange report is a <em>List(Map(String,Object))</em> of the different captures of the exchange<br>
    * &nbsp;&nbsp;&nbsp;&nbsp;- a capture is a Map of <em>Object</em> objects, following the structure:<br>
    * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;- "id" <em>String</em> the id of the exchange<br>
    * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;- "time" <em>long</em> the date of the capture<br>
    * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;- "component" <em>String</em> the component that interact on this exchange<br>
    * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;- "role" <em>String</em> the role of the component : [consumer|provider]<br>
    * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;- "type" <em>String</em> the message type : [in|out|fault|ANY]<br>
    * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;- "status"<em>String</em> the message status : [active|done|error]<br>
    *            
    * @param sinceTime <em>long</em> representation of the date, in the standard <em>long<em> format
    * @return the raw reports. not null, may be empty
    */
   List<List<Map<String,Object>>> bulkData(long sinceTime);
   }
