/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: ConfigurationDiscovery.java 2:54:36 PM ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.kernel.configuration;

import java.util.ArrayList;
import java.util.List;

import org.objectweb.petals.PetalsException;
import org.objectweb.petals.kernel.admin.ContainerInformation;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.PropertyUtil;
import org.objectweb.petals.util.SystemUtil;

/**
 * This class is used to discover other Petals container and set the
 * configuration of this container
 * 
 * @author ddesjardins - eBMWebsourcing
 */
public class ConfigurationService {

    /**
     * Logger
     */
    protected LoggingUtil log;

    protected List<ContainerInformation> containers = new ArrayList<ContainerInformation>();

    /**
     * Container UID
     */
    protected long containerUID = SystemUtil.getContainerUID();

    /**
     * Joram domain port
     */
    protected String domainPort = SystemUtil.getJoramDomainPort();

    /**
     * Host of the container
     */
    protected String host = SystemUtil.getHost();

    /**
     * Html adaptor port
     */
    protected String htmlPort = SystemUtil.getHtmlPort();

    /**
     * Joram id
     */
    protected String id = SystemUtil.getJoramId();

    /**
     * JMX server port
     */
    protected String jmxPort = SystemUtil.getJmxPort();

    /**
     * JNDI server port
     */
    protected String jndiPort = SystemUtil.getJndiPort();

    /**
     * Joram tcp port
     */
    protected String tcpPort = SystemUtil.getJoramTCPPort();

    protected boolean isInitialized;

    public ConfigurationService(LoggingUtil log) {
        this.log = log;
    }

    /**
     * Set up the container configuration
     * 
     * In standalone mode:
     * <ol>
     * <li> set up only the container UID</li>
     * </ol>
     * 
     * In distributed mode:
     * <ol>
     * <li>scan the network for other JNDI servers</li>
     * <li>connect to a JNDI server to retrieve the list of
     * ContainerInformation</li>
     * <li>set up the properties of this container (change the ports if they
     * are already taken)</li>
     * <li>write the new configuration on the harddrive</li>
     * </ol>
     */
    public void setUpConfiguration() {
        log.start();
        List<String> ids = new ArrayList<String>();
        List<String> takenPorts = new ArrayList<String>();

        // Retreive the containers informations
        for (ContainerInformation information : containers) {
            if (information.getUid() != containerUID) {
                ids.add(information.getJoramId());
                if (host.equals(information.getHost())) {
                    takenPorts.add(information.getJndiPort());
                    takenPorts.add(information.getJoramTCPPort());
                    takenPorts.add(information.getHtmlPort());
                    takenPorts.add(information.getJmxPort());
                    takenPorts.add(information.getJoramDomainPort());
                }
            }
        }

        // Find not taken values
        findNotTakenValues(ids, takenPorts);

        if (containerUID == -1) {
            containerUID = System.currentTimeMillis();
            SystemUtil.setContainerUID(containerUID);
        }

        // Update the container properties
        try {
            PropertyUtil.updateContainerProperties(id, domainPort, tcpPort,
                    jmxPort, htmlPort, jndiPort, containerUID);
        } catch (PetalsException e) {
            log.error("Error durring Petals setup.", e);
        }
        isInitialized = true;
        log.end();
    }

    /**
     * Find a not taken value in a range
     * 
     * @param takens
     *            List of already taken values
     * @param startValue
     *            start value
     * @param range
     *            range of values
     * @return a not taken value in the range from the start value
     */
    protected String findNotTaken(List<String> takens, int startValue, int range) {
        int i = startValue;
        while (i < (startValue + range) && takens.contains("" + i)) {
            i++;
        }
        return "" + i;
    }

    /**
     * Find not taken values in the id list and in the port list
     * 
     * @param takenIds
     *            taken ids
     * @param takenPorts
     *            taken ports
     */
    protected void findNotTakenValues(List<String> takenIds,
            List<String> takenPorts) {
        // Find new id
        String newId = findNotTaken(takenIds, Integer.parseInt(id), 255);
        if (!newId.equals(takenIds)) {
            log.debug("A new id has been set for the container : " + newId);
            id = newId;
            SystemUtil.setContainerName(newId);
            SystemUtil.setJoramId(newId);
        }
        // Find new joram tcp port
        String newTCPPort = findNotTaken(takenPorts, Integer.parseInt(tcpPort),
                255);
        if (!newTCPPort.equals(tcpPort)) {
            takenPorts.add(newTCPPort);
            log.debug("A new Joram tcp port has been set for the container : "
                    + newTCPPort);
            tcpPort = newTCPPort;
            SystemUtil.setJoramTCPPort(newTCPPort);
        }
        // Find new html port
        String newHTMLPort = findNotTaken(takenPorts, Integer
                .parseInt(htmlPort), 255);
        if (!newHTMLPort.equals(htmlPort)) {
            takenPorts.add(newHTMLPort);
            log.debug("A new HTML port has been set for the container : "
                    + newHTMLPort);
            htmlPort = newHTMLPort;
            SystemUtil.setHtmlPort(newHTMLPort);
        }
        // Find new jndi port
        String newJNDIPort = findNotTaken(takenPorts, Integer
                .parseInt(jndiPort), 255);
        if (!newJNDIPort.equals(jndiPort)) {
            takenPorts.add(newJNDIPort);
            log.debug("A new JNDI port has been set for the container : "
                    + newJNDIPort);
            jndiPort = newJNDIPort;
            SystemUtil.setJndiPort(newJNDIPort);
        }
        // Find new jmx port
        String newJMXPort = findNotTaken(takenPorts, Integer.parseInt(jmxPort),
                255);
        if (!newJMXPort.equals(jmxPort)) {
            takenPorts.add(newJMXPort);
            log.debug("A new JMX port has been set for the container : "
                    + newJMXPort);
            jmxPort = newJMXPort;
            SystemUtil.setJmxPort(newJMXPort);
        }
        // Find new domain port
        String newDomainPort = findNotTaken(takenPorts, Integer
                .parseInt(domainPort), 255);
        if (!newDomainPort.equals(domainPort)) {
            takenPorts.add(newDomainPort);
            domainPort = newDomainPort;
            log.debug("A new Joram domain port has been set for the "
                    + " container : " + newDomainPort);
            SystemUtil.setJoramDomainPort(newDomainPort);
        }
    }

    public void setContainers(List<ContainerInformation> containers) {
        this.containers = containers;
    }

    public boolean isInitialized() {
        return isInitialized;
    }

}
