/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: NamingContextImpl.java 7:13:54 PM ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.kernel.registry.jndi;

import java.io.Serializable;
import java.util.Hashtable;

import javax.naming.Binding;
import javax.naming.CompositeName;
import javax.naming.Context;
import javax.naming.Name;
import javax.naming.NameClassPair;
import javax.naming.NameParser;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;

import org.objectweb.petals.kernel.registry.msg.request.BindRequest;
import org.objectweb.petals.kernel.registry.msg.request.CreateSubcontextRequest;
import org.objectweb.petals.kernel.registry.msg.request.DestroySubcontextRequest;
import org.objectweb.petals.kernel.registry.msg.request.ListBindingsRequest;
import org.objectweb.petals.kernel.registry.msg.request.ListRequest;
import org.objectweb.petals.kernel.registry.msg.request.LookupLinkRequest;
import org.objectweb.petals.kernel.registry.msg.request.LookupRequest;
import org.objectweb.petals.kernel.registry.msg.request.RebindRequest;
import org.objectweb.petals.kernel.registry.msg.request.RenameRequest;
import org.objectweb.petals.kernel.registry.msg.request.UnbindRequest;

/**
 * JNDI context
 * 
 * @author ddesjardins - eBMWebsourcing
 */
public class NamingContextImpl implements Context, Serializable {

    private static final long serialVersionUID = 1L;

    /**
     * Name of the context
     */
    protected String contextName;

    /**
     * Context environment
     */
    protected Hashtable<String, Object> env;

    /**
     * Connection to the RegistryServer
     */
    protected JNDIConnection jndiConnection;

    public NamingContextImpl(Hashtable<String, Object> env,
        JNDIConnection jndiConnection, String contextName) {
        super();
        this.env = env;
        this.jndiConnection = jndiConnection;
        this.contextName = contextName;
    }

    /**
     * @see Context#addToEnvironment(String, Object)
     */
    public Object addToEnvironment(String propName, Object propVal)
        throws NamingException {
        Object out = env.get(propName);
        env.put(propName, propVal);
        return out;
    }

    /**
     * @see Context#bind(Name, Object)
     */
    public void bind(Name name, Object obj) throws NamingException {
        bind(name.toString(), obj);
    }

    /**
     * @see Context#bind(String, Object)
     */
    public void bind(String name, Object obj) throws NamingException {
        jndiConnection.send(new BindRequest(contextName, name, obj, 0, 0));
    }

    /**
     * @see Context#close()
     */
    public void close() throws NamingException {
        env = null;
        jndiConnection = null;
        contextName = null;
    }

    /**
     * @see Context#composeName(Name, Name)
     */
    public Name composeName(Name name, Name prefix) throws NamingException {
        if (name == null) {
            throw new NamingException("Name must be non null");
        }
        if (prefix == null) {
            throw new NamingException("Prefix must be non null");
        }
        return new CompositeName(prefix.toString() + name.toString());
    }

    /**
     * @see Context#composeName(String, String)
     */
    public String composeName(String name, String prefix)
        throws NamingException {
        if (name == null) {
            throw new NamingException("Name must be non null");
        }
        if (prefix == null) {
            throw new NamingException("Prefix must be non null");
        }
        return prefix + name;
    }

    /**
     * @see Context#createSubcontext(Name)
     */
    public Context createSubcontext(Name name) throws NamingException {
        return createSubcontext(name.toString());
    }

    /**
     * @see Context#createSubcontext(String)
     */
    public Context createSubcontext(String name) throws NamingException {
        Object newName = jndiConnection.send(new CreateSubcontextRequest(
            contextName, name, 0, 0));
        NamingContextImpl contextImpl = new NamingContextImpl(env,
            jndiConnection, (String) newName);
        return contextImpl;
    }

    /**
     * @see Context#destroySubcontext(Name)
     */
    public void destroySubcontext(Name name) throws NamingException {
        destroySubcontext(name.toString());
    }

    /**
     * @see Context#destroySubcontext(String)
     */
    public void destroySubcontext(String name) throws NamingException {
        jndiConnection.send(new DestroySubcontextRequest(contextName, name, 0,
            0));
    }

    /**
     * @see Context#getEnvironment()
     */
    public Hashtable<?, ?> getEnvironment() throws NamingException {
        return env;
    }

    /**
     * @see Context#getNameInNamespace()
     */
    public String getNameInNamespace() throws NamingException {
        return contextName;
    }

    /**
     * @see Context#getNameParser(Name)
     */
    public NameParser getNameParser(Name name) throws NamingException {
        return new JNDINameParser();
    }

    /**
     * @see Context#getNameParser(String)
     */
    public NameParser getNameParser(String name) throws NamingException {
        return new JNDINameParser();
    }

    /**
     * @see Context#list(Name)
     */
    public NamingEnumeration<NameClassPair> list(Name name)
        throws NamingException {
        return list(name.toString());
    }

    /**
     * @see Context#list(String)
     */
    public NamingEnumeration<NameClassPair> list(String name)
        throws NamingException {
        NameClassPair[] names = (NameClassPair[]) jndiConnection
            .send(new ListRequest(contextName, name, 0, 0));
        return new NamingEnumerationImpl<NameClassPair>(names);
    }

    /**
     * @see Context#listBindings(Name)
     */
    public NamingEnumeration<Binding> listBindings(Name name)
        throws NamingException {
        return listBindings(name.toString());
    }

    /**
     * @see Context#listBindings(String)
     */
    public NamingEnumeration<Binding> listBindings(String name)
        throws NamingException {
        Binding[] bindings = (Binding[]) jndiConnection
            .send(new ListBindingsRequest(contextName, name, 0, 0));
        return new NamingEnumerationImpl<Binding>(bindings);
    }

    /**
     * @see Context#lookup(Name)
     */
    public Object lookup(Name name) throws NamingException {
        return lookup(name.toString());
    }

    /**
     * @see Context#lookup(String)
     */
    public Object lookup(String name) throws NamingException {
        return jndiConnection.send(new LookupRequest(contextName, name, 0, 0));
    }

    /**
     * @see Context#lookupLink(Name)
     */
    public Object lookupLink(Name name) throws NamingException {
        return lookupLink(name.toString());
    }

    /**
     * @see Context#lookupLink(String)
     */
    public Object lookupLink(String name) throws NamingException {
        return jndiConnection.send(new LookupLinkRequest(contextName, name, 0,
            0));
    }

    /**
     * @see Context#rebind(Name, Object)
     */
    public void rebind(Name name, Object obj) throws NamingException {
        rebind(name.toString(), obj);
    }

    /**
     * @see Context#rebind(String, Object)
     */
    public void rebind(String name, Object obj) throws NamingException {
        jndiConnection.send(new RebindRequest(contextName, name, obj, 0, 0));
    }

    /**
     * @see Context#removeFromEnvironment(String)
     */
    public Object removeFromEnvironment(String propName) throws NamingException {
        Object out = env.get(propName);
        if (env.containsKey(propName)) {
            env.remove(propName);
        }
        return out;
    }

    /**
     * @see Context#rename(Name, Name)
     */
    public void rename(Name oldName, Name newName) throws NamingException {
        rename(oldName.toString(), newName.toString());
    }

    /**
     * @see Context#rename(String, String)
     */
    public void rename(String oldName, String newName) throws NamingException {
        jndiConnection.send(new RenameRequest(contextName, oldName, newName, 0,
            0));
    }

    /**
     * @see Context#unbind(Name)
     */
    public void unbind(Name name) throws NamingException {
        unbind(name.toString());
    }

    /**
     * @see Context#unbind(String)
     */
    public void unbind(String name) throws NamingException {
        jndiConnection.send(new UnbindRequest(contextName, name, 0, 0));
    }

}
