/*
 * PETALS: PETALS Services Platform
 * Copyright (C) 2007 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): Adrien LOUIS
 * --------------------------------------------------------------------------
 * $Id: R
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.monitoring;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;

import org.objectweb.fractal.fraclet.annotation.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.Interface;
import org.objectweb.fractal.fraclet.annotation.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.LifeCycleType;
import org.objectweb.fractal.fraclet.annotation.Monolog;
import org.objectweb.fractal.fraclet.annotation.Provides;
import org.objectweb.fractal.fraclet.annotation.Requires;
import org.objectweb.util.monolog.api.Logger;

import org.objectweb.petals.jbi.component.context.ComponentContextImpl;
import org.objectweb.petals.jbi.messaging.MessageExchangeImpl;
import org.objectweb.petals.jbi.routing.Router;
import org.objectweb.petals.jbi.routing.RoutingException;
import org.objectweb.petals.util.LoggingUtil;

/**
 * @version $Rev: 617 $ $Date: 2006-06-19 17:28:41 +0200 (lun, 19 jun 2006) $
 * @since Petals 1.1
 * @author alouis
 */
@FractalComponent
@Provides(interfaces = {
    @Interface(name = "service", signature = org.objectweb.petals.jbi.routing.Router.class),
    @Interface(name = "monitoring-service", signature = org.objectweb.petals.monitoring.RouterMonitor.class)})
public class RouterMonitorImpl implements Router, RouterMonitor {

    private boolean monitoring = false;

    @Requires(name = "router", signature = org.objectweb.petals.jbi.routing.Router.class)
    private Router router;

    @Monolog(name="logger")
    protected Logger logger;
    
    protected LoggingUtil log;
    
    private HashMap<String, ExchangeStateReport> exchangeReports = new HashMap<String, ExchangeStateReport>();

    private TimeComparator timeComparator = new TimeComparator();

    ////////////////////////////////////////////
    // Router implementation
    ////////////////////////////////////////////
    
    public void send(ComponentContextImpl source, MessageExchangeImpl exchange,
        long timeOut) throws RoutingException {

        if (monitoring) {
            report(exchange);
        }

        router.send(source, exchange, timeOut);
    }

    public void receive(MessageExchangeImpl exchange) throws RoutingException {

        if (monitoring) {
            report(exchange);
        }

        router.receive(exchange);
    }

    private void report(MessageExchangeImpl exchange) {
        ExchangeStateReport report = exchangeReports.get(exchange
            .getExchangeId());
        if (report == null) {
            report = new ExchangeStateReport(exchange);
            exchangeReports.put(exchange.getExchangeId(), report);
        }
        report.reportState(exchange);

    }

    ////////////////////////////////////////////
    // Monitor implementation
    ////////////////////////////////////////////
    
    public boolean isMonitoring() {
        return monitoring;
    }

    /**
     * if set to false, the reports are reset.
     */
    public void setMonitoring(boolean monitoring) {
        this.monitoring = monitoring;
        if (! monitoring) {
            exchangeReports = new HashMap<String, ExchangeStateReport>();
        }
    }
    
    
    /**
     * Return the reports for all exchange since the specified date, sorted by
     * date
     * 
     * @param sinceDate
     *            date (in 'time' format)
     * @return a map of reports. not null, may be empty
     */
    public List<ExchangeStateReport> getReports(long sinceDate) {
        log.start();
        // 1. get reports since the specified date
        List<ExchangeStateReport> list = new ArrayList<ExchangeStateReport>();

        for (ExchangeStateReport report : exchangeReports.values()) {
            if (report.getReportList().size() == 0
                || report.getReportList().get(0).time >= sinceDate) {
                list.add(report);
            }
        }
        // 2. sort the list by date
        Collections.sort(list, timeComparator);

        // 3. return
        log.end();
        return list;
    }

    class TimeComparator implements Comparator<ExchangeStateReport> {

        /**
         * Compare the two exchange-reports by comparing the date of their
         * respective first state-capture.
         */
        public int compare(ExchangeStateReport o1, ExchangeStateReport o2) {
            List<ExchangeStateCapture> l1 = o1.getReportList();
            List<ExchangeStateCapture> l2 = o2.getReportList();

            int comp = 0;
            if (l1.size() == 0 || l2.size() == 0) {
                // one of the exchange has no capture, so it is more recent than
                // the other exchange
                comp = l1.size() - l2.size();
            } else {
                comp = (int) (l2.get(0).time - l1.get(0).time);
            }

            return comp;
        }

    }
    
    
    
//  ----------------------------------------------------------------------
    // Fractal LifecycleController implementation
    // ----------------------------------------------------------------------

    @LifeCycle(on=LifeCycleType.START)
    protected void start() {
        log = new LoggingUtil(logger);
    }
}
