/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005-2006 EBM WebSourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.tools.jbicommon.descriptor;

import javax.xml.namespace.QName;

import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;
import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * The consumes element is used to declare a service consumed by a service unit.
 * A consumed service must be define with the following required attribute:
 * <ul>
 * <li><b>interface-name.</b> This qualified name indicates the type of
 * service to be consumed.</li>
 * </ul>
 * Optionally a fully qualified name of the endpoint the SU will use to consume
 * the service can be specified. If the service endpoint name is declared, an
 * optional attribute can be supplied:
 * <ul>
 * <li><b>link-type.</b> This attribute is used to indicate the expected use
 * of this consumer with connection links. The possible values of this
 * attributes are:</li>
 * <ul>
 * <li><b>standard.</b> This is the default value. This indicates that the
 * provided service endpoint must be routed according to standard normalized
 * message routing rules.</li>
 * <li><b>hard.</b>This indicates that the provided service endpoint name must
 * match a service provider's service endpoint name; indirect connections are
 * not allowed.</li>
 * <li><b>soft.</b>This indicates that the provided service endpoint must
 * <i>not</i> match a service provider's service endpoint name; rather, it must
 * match an indirect connection name.</li>
 * </ul>
 * </ul>
 * 
 * @version $Rev: 477 $ $Date: 2006-05-29 15:18:07Z $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 */
public class Consumes extends ExtensibleJbiElement {

    /**
     * Hard Link Type.
     */
    public static final String HARD_LINK_TYPE = "hard";

    /**
     * Soft Link Type.
     */
    public static final String SOFT_LINK_TYPE = "soft";

    /**
     * Standard Link Type.
     */
    public static final String STD_LINK_TYPE = "standard";

    /**
     * Endpoint name to be provided.
     */
    private String endpointName;

    /**
     * Qualified interface name.
     */
    private QName interfaceName;

    /**
     * The optional link type.
     */
    private String linkType;

    /**
     * Service name to be provided.
     */
    private QName serviceName;

    /**
     * Default constructor.
     */
    public Consumes() { // NOPMD by gblondelle
        super();
    }

    @Override
    public boolean equals(final Object other) {
        if (!(other instanceof Consumes)) {
            return false; // NOPMD by gblondelle
        }
        Consumes castOther = (Consumes) other;
        return new EqualsBuilder().append(endpointName, castOther.endpointName)
                .append(interfaceName, castOther.interfaceName).append(
                        linkType, castOther.linkType).append(serviceName,
                        castOther.serviceName).isEquals();
    }

    /**
     * Returns the optional fully qualified endpoint which the service unit will
     * use to consume the service.
     * 
     * @return <code>String</code> the endpoint name.
     */
    public String getEndpointName() {
        return endpointName;
    }

    /**
     * Return the qualified name indicating the type of service to be consumed.
     * 
     * @return <code>QName</code> the interface name.
     */
    public QName getInterfaceName() {
        return interfaceName;
    }

    /**
     * Returns the optional link type associated to a fully qualified endpoint
     * the SU will use to consume the service.
     * 
     * @return <code>String</code> the link type: <code>"standard"</code>,
     *         <code>"hard"</code> or <code>"soft"</code>.
     */
    public String getLinkType() {
        return linkType;
    }

    //
    // Getters and Setters
    //

    /**
     * Returns the optional fully qualified service name which the service unit
     * will use to consume the service.
     * 
     * @return <code>QName</code> the service name.
     */
    public QName getServiceName() {
        return serviceName;
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().append(endpointName).append(interfaceName)
                .append(linkType).append(serviceName).toHashCode();
    }

    /**
     * Returns <code>true</code> if the hard link type is used.
     */
    public boolean isHardLink() {
        return (hasQualifiedEndpoint())
                && (linkType != null && linkType.equals(HARD_LINK_TYPE));
    }

    /**
     * Returns <code>true</code> if the soft link type is used.
     */
    public boolean isSoftLink() {
        return (hasQualifiedEndpoint())
                && (linkType != null && linkType.equals(SOFT_LINK_TYPE));
    }

    /**
     * Returns <code>true</code> if the standard link type is used.
     */
    public boolean isStandardLink() {
        return (hasQualifiedEndpoint())
                && (linkType == null || linkType.equals(STD_LINK_TYPE));
    }

    @Override
    public String toString() {
        return new ToStringBuilder(this).append("endpointName", endpointName)
                .append("interfaceName", interfaceName).append("linkType",
                        linkType).append("serviceName", serviceName).toString();
    }

    /**
     * Returns <code>true</code> if the optional fully qualified endpoint name
     * have been specified. Otherwise, returns <code>false</code>.
     */
    boolean hasQualifiedEndpoint() {
        return serviceName != null && endpointName != null;
    }

    //
    // implementation methods
    //

    /**
     * Set the optional fully qualified endpoint name which the service unit
     * will consume.
     * 
     * @param endpointName
     *            the endpoint name.
     */
    void setEndpointName(final String endpointName) {
        this.endpointName = endpointName;
    }

    /**
     * Set the interface name indicating the type of service to be consumed.
     * 
     * @param interfaceName
     *            the interface name.
     */
    void setInterfaceName(final QName interfaceName) {
        this.interfaceName = interfaceName;
    }

    /**
     * Set the optional link type.
     * 
     * @param type
     *            the link type.
     */
    void setLinkType(final String type) {
        this.linkType = type;
    }

    /**
     * Set the optional fully qualified service name which the service unit will
     * consume.
     * 
     * @param serviceName
     *            the service name.
     */
    void setServiceName(final QName serviceName) {
        this.serviceName = serviceName;
    }
}
