/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: JBIDescriptorBuilder.java 565 2006-06-08 10:03:27Z ofabre $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.tools.jbicommon.descriptor; // NOPMD by ofabre

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.objectweb.petals.tools.jbicommon.util.FileUtil;
import org.objectweb.petals.tools.jbicommon.util.XMLUtil;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * The JBIDescriptorBuilder helps to populate a <code>JBIDescriptor</code>
 * instance with values loaded from a specified <i>jbi.xml</i>.
 * 
 * @version $Rev: 565 $ $Date: 2006-06-08 10:03:27Z $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 */
public class JBIDescriptorBuilder {
    /**
     * This class is used to handle errors during creating XML objects, parsing
     * XML and validating a XML.
     */
    private class JbiXmlErrorHandler implements ErrorHandler {

        /**
         * 
         */
        public JbiXmlErrorHandler() { // NOPMD by ofabre
            super();
        }

        /**
         * @see org.xml.sax.ErrorHandler#error(org.xml.sax.SAXParseException)
         */
        public void error(final SAXParseException se) throws SAXException {
            log.log(Level.SEVERE, se.getMessage(), se);
            throw se;
        }

        /**
         * @see org.xml.sax.ErrorHandler#fatalError(org.xml.sax.SAXParseException)
         */
        public void fatalError(final SAXParseException se) throws SAXException {
            log.log(Level.SEVERE, se.getMessage(), se);
            throw se;
        }

        /**
         * @see org.xml.sax.ErrorHandler#warning(org.xml.sax.SAXParseException)
         */
        public void warning(final SAXParseException e) throws SAXException {
            log.log(Level.WARNING, e.getMessage(), e);
        }

    }

    public static final String JBI_XSD = "jbi.xsd";

    /**
     * Standard location to the JBI Descriptor resource.
     */
    public static final String JBI_DESCRIPTOR_RESOURCE = "META-INF"
            + File.separator + "jbi.xml";

    /**
     * The JBI Descriptor XML Schema Resource Location ("schema/jbi.xsd").
     */
    public static final String JBI_SCHEMA_RESOURCE = "schema" + File.separator
            + JBI_XSD;

    private static final String BOOTSTRAP_CLASS_NAME_ELEM_NAME = "bootstrap-class-name";

    private static final String BOOTSTRAP_CLASS_PATH_ELEM_NAME = "bootstrap-class-path";

    private static final String COMPONENT_CLASS_NAME_ELEM_NAME = "component-class-name";

    private static final String COMPONENT_CLASS_PATH_ELEM_NAME = "component-class-path";

    private static final String ENDPOINT_ATTR_NAME = "endpoint-name";

    private static final String IDENTIFICATION_ELEM_NAME = "identification";

    private static final String INTERFACE_ATTR_NAME = "interface-name";

    private static final String INVALID_EXTENSION_NAMESPACE = "Invalid namespace for jbi extensions ";

    /**
     * Schema file
     */
    private static Schema schema = null;

    private static final String SERVICE_ATTR_NAME = "service-name";

    private static final String SHARED_LIBRARY_ELEM_NAME = "shared-library";

    // private static final String SHARED_LIBRARY_LIST_ELEM_NAME =
    // "shared-library-list";

    private static final String VERSION_ATTR_NAME = "version";

    /**
     * URI to the jbi.xml file.
     */
    private URI jbiXmlUri;

    /**
     * Logger wrapper
     */
    private Logger log; // NOPMD by ofabre

    /**
     * The source XML document file.
     */
    private Document source;

    private final JbiXmlErrorHandler xmlErrorHandler = new JbiXmlErrorHandler();

    /**
     * JBI Descriptor Builder constructor using the source xml file URI from
     * which the document will be loaded.
     */
    public JBIDescriptorBuilder(final URI sourceUri, final Logger logger) {
        if (sourceUri == null) {
            throw new IllegalArgumentException(
                    "DescriptorBuilder: invalid source URI argument.");
        }

        this.jbiXmlUri = sourceUri;
        log = logger;
    }

    /**
     * Load, Parse and Validate jbi.xml document for building (populating) the
     * <code>Descriptor</code> object.
     * 
     * @return a JBIDescriptor object mapping the XML jbi descriptor
     * 
     * @throws JBIDescriptorException
     *             if an error occured during JBIDescriptor creation
     */
    public JBIDescriptor build() throws JBIDescriptorException {

        // do not repeat load, parse and validation steps
        if (source == null) {

            // validate the jbi.xml file against the schema
            if (!validateJbiXml(jbiXmlUri)) {
                throw new JBIDescriptorException(
                        "Can't validate jbi descriptor :"
                                + jbiXmlUri.toString());
            }

            // parse XML after validation
            source = parseJbiXml(jbiXmlUri);

        }
        return buildDescriptor(source);
    }

    /**
     * Returns a String representation of a JBI descriptor file
     * 
     * @param descriptorFile
     *            the JBI Descriptor <code>File</code>
     * @return a <code>String</code> representation of the JBI Descriptor
     */
    public String getJbiDescriptorAsString() {
        File descriptorFile = new File(jbiXmlUri);
        String ret = null;
        if (descriptorFile.isFile()) {
            ByteArrayOutputStream os = null;
            InputStream is = null;
            try {
                os = new ByteArrayOutputStream();
                is = new FileInputStream(descriptorFile);
                FileUtil.copyInputStream(is, os);
                ret = os.toString();
            } catch (IOException e) {
                log.log(Level.SEVERE, "Error reading jbi descritor: "
                        + descriptorFile, e);
            }
        }
        return ret;
    }

    /**
     * @return Returns the source.
     */
    public Document getSource() {
        return source;
    }

    /**
     * @return Returns the jbiXmlUri.
     */
    public URI getURI() {
        return jbiXmlUri;
    }

    /**
     * 
     * @param sourceURI
     * @return
     */
    public Document parseJbiXml(final URI sourceURI)
            throws JBIDescriptorException {

        DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory
                .newInstance();
        docBuilderFactory.setNamespaceAware(true);

        DocumentBuilder docBuilder = null;

        try {
            docBuilder = docBuilderFactory.newDocumentBuilder();
        } catch (ParserConfigurationException pce) {
            throw new JBIDescriptorException("Bad XML parser configuration",
                    pce);
        }

        // set up xml parsing error handler
        docBuilder.setErrorHandler(xmlErrorHandler);

        // parse the xml document URI to return
        String jbiXmlURI = sourceURI.toString();

        Document document = null;

        try {
            document = docBuilder.parse(jbiXmlURI);
        } catch (SAXException saxe) {
            throw new JBIDescriptorException(
                    "JBI descriptor is not well formed XML", saxe);
        } catch (IOException ioe) {
            throw new JBIDescriptorException("Can't read JBI descriptor", ioe);
        }
        return document;
    }

    /**
     * 
     * @param sourceURI
     * @return
     * @throws SAXException
     * @throws IOException
     */
    public boolean validateJbiXml(final URI sourceURI) {
        boolean res = true;
        // instance for xml error handling

        Schema jbiSchema = getJBISchema();
        // create a Validator instance which is be used to validate a
        // document
        if (jbiSchema == null) {
            log.log(Level.WARNING, "Can't check jbi descriptor validity "
                    + "because jbi schema is unavailable");
        } else {
            Validator validator = schema.newValidator();
            validator.setErrorHandler(xmlErrorHandler);

            try {
                // validate the DOM tree
                validator.validate(new StreamSource(new File(sourceURI)));
            } catch (SAXException e) {
                log.log(Level.SEVERE, "Bad JBI descriptor for " + sourceURI, e);
                res = false;
            } catch (IOException e) {
                log.log(Level.SEVERE, "JBI descriptor not found " + sourceURI,
                        e);
                res = false;
            }
        }
        return res;
    }

    protected URI getJbiXmlUri() {
        return jbiXmlUri;
    }

    protected void setJbiXmlUri(final URI jbiXmlUri) {
        this.jbiXmlUri = jbiXmlUri;
    }

    /**
     * FIXME : Needed by unit tests. Unit tests should initialize it by a
     * constructor.
     * 
     * @param log
     */
    protected void setLog(final Logger log) {
        this.log = log;
    }

    /**
     * Parse the XML document and load values to the Descriptor object
     * representation.
     * 
     * @param document
     * @return a populated Descriptor instance with values loaded from the XML
     *         file. May return <code>null</code> if an error occur while
     *         navigating on XML tree.
     * @throws JBIDescriptorException
     */
    private JBIDescriptor buildDescriptor(final Document document)
            throws JBIDescriptorException {

        JBIDescriptor descriptor = new JBIDescriptor();

        // get root element
        Element jbi = document.getDocumentElement();

        // extract JBI version attribute from the root element
        String versionAttr = XMLUtil.getRequiredAttributeValue(jbi,
                VERSION_ATTR_NAME);
        descriptor.setVersion(Double.parseDouble(versionAttr));

        List<Node> children = XMLUtil.getNodeChildren(jbi);
        for (Node child : children) {
            if ("component".equals(child.getNodeName())) {
                descriptor.setComponent(loadComponentXml(child));
            } else if (SHARED_LIBRARY_ELEM_NAME.equals(child.getNodeName())) {
                descriptor.setSharedLibrary(loadSharedLibraryXml(child));
            } else if ("service-assembly".equals(child.getNodeName())) {
                descriptor.setServiceAssembly(loadServiceAssemblyXml(child));
            } else if ("services".equals(child.getNodeName())) {
                descriptor.setServices(loadServicesXml(child));
            }
        }
        return descriptor;
    }

    private synchronized Schema getJBISchema() { // NOPMD by gblondelle
        if (schema == null) {

            // create a SchemaFactory capable of understanding WXS schemas
            SchemaFactory factory = SchemaFactory
                    .newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);

            factory.setErrorHandler(xmlErrorHandler);

            // load a WXS schema, represented by a Schema instance
            try {
                schema = factory
                        .newSchema(new StreamSource(openJbiSchemaFile()));
            } catch (SAXException saxe) {
                log.log(Level.SEVERE, saxe.getMessage(), saxe);
            }
        }
        return schema;
    }

    /**
     * @param element
     * @throws JBIDescriptorException
     */
    private ComponentDescription loadComponentXml(final Node production)
            throws JBIDescriptorException {
        ComponentDescription component = new ComponentDescription();

        // component attributes
        component
                .setType(XMLUtil.getRequiredAttributeValue(production, "type"));
        component.setComponentClassLoaderDelegation(XMLUtil.getAttributeValue(
                production, "component-class-loader-delegation"));
        component.setBootstrapClassLoaderDelegation(XMLUtil.getAttributeValue(
                production, "bootstrap-class-loader-delegation"));

        Extensions extensions = new Extensions();
        List<SharedLibraryList> slList = new ArrayList<SharedLibraryList>();
        // component sub nodes
        List<Node> children = XMLUtil.getNodeChildren(production);
        for (Node child : children) {
            if (IDENTIFICATION_ELEM_NAME.equals(child.getNodeName())) {
                component.setIdentification(loadIdentificationXml(child));
            } else if (COMPONENT_CLASS_NAME_ELEM_NAME.equals(child
                    .getNodeName())) {
                component.setComponentClassName(XMLUtil.getTextContent(child));
            } else if (COMPONENT_CLASS_PATH_ELEM_NAME.equals(child
                    .getNodeName())) {
                component.setComponentClassPath(XMLUtil.getTextContents(child
                        .getChildNodes()));
            } else if (BOOTSTRAP_CLASS_NAME_ELEM_NAME.equals(child
                    .getNodeName())) {
                component.setBootstrapClassName(XMLUtil.getTextContent(child));
            } else if (BOOTSTRAP_CLASS_PATH_ELEM_NAME.equals(child
                    .getNodeName())) {
                component.setBootstrapClassPath(XMLUtil.getTextContents(child
                        .getChildNodes()));
            } else if (SHARED_LIBRARY_ELEM_NAME.equals(child.getNodeName())) {
                loadSharedLibraryListXml(slList, child);
            } else {
                loadExtensions(extensions, child);
            }
        }
        // set extensions
        component.setExtensions(extensions);
        // set shared libs
        component.setSharedLibraryList(slList);

        return component;
    }

    /**
     * @param connectionsTree
     * @return
     * @throws JBIDescriptorException
     */
    private List<Connection> loadConnectionsXml(final Node connectionsTree)
            throws JBIDescriptorException {

        List<Connection> connections = new ArrayList<Connection>();

        if (connectionsTree != null) {
            // get "connections/connection" Node List
            List<Node> children = XMLUtil.getNodeChildren(connectionsTree);
            for (Node child : children) {
                if ("connection".equals(child.getNodeName())) {
                    connections.add(loadConnectionXml(child));
                }
            }
        }
        return connections;
    }

    private Connection loadConnectionXml(final Node node)
            throws JBIDescriptorException {

        // connection object
        Connection newItem = new Connection();

        List<Node> children = XMLUtil.getNodeChildren(node);
        for (Node child : children) {
            if ("consumer".equals(child.getNodeName())) {
                /*
                 * TODO deal with cases where there is only the interface name
                 * and no endpoint name and service name
                 * newItem.setConsumerInterfaceName(extractXmlAttributeQName(consumer,
                 * "interface-name"));
                 */

                newItem.setConsumerServiceName(XMLUtil
                        .extractXmlAttributeQName(child, SERVICE_ATTR_NAME));

                newItem.setConsumerEndpointName(XMLUtil.getAttributeValue(
                        child, ENDPOINT_ATTR_NAME));
            } else if ("provider".equals(child.getNodeName())) {
                newItem.setProviderServiceName(XMLUtil
                        .extractRequiredXmlAttributeQName(child,
                                SERVICE_ATTR_NAME));

                newItem.setProviderEndpointName(XMLUtil
                        .getRequiredAttributeValue(child, ENDPOINT_ATTR_NAME));
            }
        }
        return newItem;
    }

    /**
     * @param list
     * @return
     * @throws JBIDescriptorException
     */
    private Consumes loadConsumesXml(final Node node)
            throws JBIDescriptorException {
        Consumes newItem = new Consumes();

        // consumes attributes
        newItem.setInterfaceName(XMLUtil.extractRequiredXmlAttributeQName(node,
                INTERFACE_ATTR_NAME));

        newItem.setServiceName(XMLUtil.extractXmlAttributeQName(node,
                SERVICE_ATTR_NAME));

        newItem.setEndpointName(XMLUtil.getAttributeValue(node,
                ENDPOINT_ATTR_NAME));

        newItem.setLinkType(XMLUtil.getAttributeValue(node, "link-type"));

        // consumes extensions
        Extensions extensions = new Extensions();
        List<Node> children = XMLUtil.getNodeChildren(node);
        for (Node child : children) {
            loadExtensions(extensions, child);
        }
        newItem.setExtensions(extensions);

        return newItem;
    }

    /**
     * Extension nodes are nodes which names are different from those specified
     * by JBI. This method load these extensions.
     * 
     * @param node
     *            an extension xml node
     */
    private void loadExtensions(final Extensions extensions, final Node node) {

        // Add the extension node to the extensions documentfragment containing
        // all extensions
        DocumentFragment fragment = extensions.getDocumentFragment();
        if (fragment == null) {
            fragment = node.getOwnerDocument().createDocumentFragment();
            extensions.setDocumentFragment(fragment);
        }
        fragment.appendChild(node);

        // Load extension information
        String uriString = node.lookupNamespaceURI(node.getPrefix());
        if (uriString != null) {
            URI uri = null;
            try {
                uri = new URI(uriString);
            } catch (URISyntaxException e) {
                log.warning(INVALID_EXTENSION_NAMESPACE + node.getNodeName());
            }

            // Check if the node uri is a PEtALS uri
            if (PetalsExtensionsUris.PETALS_EXTENSIONS
                    .equals(PetalsExtensionsUris.valueOf(uri))) {
                loadPetalsExtensions(extensions, node);
            } else {
                loadUnknownExtension(extensions, node);
            }
        }
    }

    private Identification loadIdentificationXml(final Node identificationNode) {
        Identification identification = new Identification();

        Extensions extensions = new Extensions();
        // identification sub nodes
        List<Node> children = XMLUtil.getNodeChildren(identificationNode);
        for (Node child : children) {
            if ("name".equals(child.getNodeName())) {
                identification.setName(XMLUtil.getTextContent(child));
            } else if ("description".equals(child.getNodeName())) {
                identification.setDescription(XMLUtil.getTextContent(child));
            } else {
                loadExtensions(extensions, child);
            }
        }
        // set extensions
        identification.setExtensions(extensions);

        return identification;
    }

    private void loadPetalsExtensions(final Extensions extensions,
            final Node node) {
        List<Node> children = XMLUtil.getNodeChildren(node);
        for (Node child : children) {
            String childURIString = child.lookupNamespaceURI(child.getPrefix());
            if (childURIString != null) {
                URI childURI = null;
                try {
                    childURI = new URI(childURIString); // NOPMD by gblondelle
                } catch (URISyntaxException e) {
                    log.warning(INVALID_EXTENSION_NAMESPACE
                            + child.getNodeName());
                }

                if (PetalsExtensionsUris.PETALS_EXTENSION_KEY_VALUE
                        .equals(PetalsExtensionsUris.valueOf(childURI))) {
                    loadPetalsKeyValueExtension(extensions, child);
                } else {
                    loadUnknownExtension(extensions, child);
                }
            }
        }
    }

    private void loadPetalsKeyValueExtension(final Extensions extensions,
            final Node child) {
        PetalsExtension petalsExtension = new PetalsExtension();
        petalsExtension
                .setExtensionURI(PetalsExtensionsUris.PETALS_EXTENSION_KEY_VALUE
                        .value());
        // treat key value nodes
        Map<String, String> keyValues = new HashMap<String, String>();
        List<Node> children = XMLUtil.getNodeChildren(child);
        for (Node extChild : children) {
            keyValues.put(extChild.getNodeName(), XMLUtil
                    .getTextContent(extChild));
        }
        petalsExtension.setExtensionObject(keyValues);
        extensions.addPetalsExtension(
                PetalsExtensionsUris.PETALS_EXTENSION_KEY_VALUE.value(),
                petalsExtension);
    }

    /**
     * @param list
     * @return
     * @throws JBIDescriptorException
     */
    private Provides loadProvidesXml(final Node node)
            throws JBIDescriptorException {
        Provides newItem = new Provides();

        // Provides attributes
        newItem.setInterfaceName(XMLUtil.extractRequiredXmlAttributeQName(node,
                INTERFACE_ATTR_NAME));

        newItem.setServiceName(XMLUtil.extractRequiredXmlAttributeQName(node,
                SERVICE_ATTR_NAME));

        newItem.setEndpointName(XMLUtil.getRequiredAttributeValue(node,
                ENDPOINT_ATTR_NAME));

        // provides extensions
        Extensions extensions = new Extensions();
        List<Node> children = XMLUtil.getNodeChildren(node);
        for (Node child : children) {
            loadExtensions(extensions, child);
        }
        newItem.setExtensions(extensions);

        return newItem;
    }

    /**
     * @param element
     * @throws JBIDescriptorException
     */
    private ServiceAssembly loadServiceAssemblyXml(final Node node)
            throws JBIDescriptorException {
        ServiceAssembly serviceAssembly = new ServiceAssembly();

        Extensions extensions = new Extensions();
        // Services assembly sub nodes
        List<Node> children = XMLUtil.getNodeChildren(node);
        for (Node child : children) {
            if (IDENTIFICATION_ELEM_NAME.equals(child.getNodeName())) {
                serviceAssembly.setIdentification(loadIdentificationXml(child));
            } else if ("service-unit".equals(child.getNodeName())) {
                serviceAssembly.addServiceUnit(loadServiceUnitXml(child));
            } else if ("connections".equals(child.getNodeName())) {
                serviceAssembly.setConnections(loadConnectionsXml(child));
            } else {
                loadExtensions(extensions, child);
            }
        }
        // set extensions
        serviceAssembly.setExtensions(extensions);

        return serviceAssembly;
    }

    /**
     * @param element
     * @throws JBIDescriptorException
     */
    private Services loadServicesXml(final Node production)
            throws JBIDescriptorException {

        Services services = new Services();

        // services attributes
        // is binding component (relative to service unit)
        String isBindingComponent = XMLUtil.getRequiredAttributeValue(
                production, "binding-component");
        services.setBindingComponent(Boolean.parseBoolean(isBindingComponent));

        Extensions extensions = new Extensions();
        // Services sub nodes
        List<Node> children = XMLUtil.getNodeChildren(production);
        for (Node child : children) {
            if ("provides".equals(child.getNodeName())) {
                services.addProvides(loadProvidesXml(child));
            } else if ("consumes".equals(child.getNodeName())) {
                services.addConsumes(loadConsumesXml(child));
            } else {
                loadExtensions(extensions, child);
            }
        }
        // set extensions
        services.setExtensions(extensions);

        return services;
    }

    /**
     * @param list
     * @return
     */
    private ServiceUnit loadServiceUnitXml(final Node node) {

        ServiceUnit result = new ServiceUnit();

        Extensions extensions = new Extensions();
        // Service unit sub nodes
        List<Node> children = XMLUtil.getNodeChildren(node);
        for (Node child : children) {
            if (IDENTIFICATION_ELEM_NAME.equals(child.getNodeName())) {
                result.setIdentification(loadIdentificationXml(child));
            } else if ("target".equals(child.getNodeName())) {
                List<Node> childChildren = XMLUtil.getNodeChildren(child);
                for (Node targetChild : childChildren) {
                    if ("artifacts-zip".equals(targetChild.getNodeName())) {
                        result.setTargetArtifactsZip(XMLUtil
                                .getTextContent(targetChild));
                    } else if ("component-name".equals(targetChild
                            .getNodeName())) {
                        result.setTargetComponentName(XMLUtil
                                .getTextContent(targetChild));
                    }
                }
            } else {
                loadExtensions(extensions, child);
            }
        }
        // set extensions
        result.setExtensions(extensions);

        return result;
    }

    /**
     * @param list
     * @return
     */
    private void loadSharedLibraryListXml(final List<SharedLibraryList> slList,
            final Node node) {
        SharedLibraryList sharedLibrary = new SharedLibraryList();
        sharedLibrary.setName(XMLUtil.getTextContent(node));
        sharedLibrary.setVersion(XMLUtil.getAttributeValue(node,
                VERSION_ATTR_NAME));
        slList.add(sharedLibrary);
    }

    /**
     * @param element
     */
    private SharedLibrary loadSharedLibraryXml(final Node production) {
        SharedLibrary sharedLibrary = new SharedLibrary();

        // shared library attributes
        sharedLibrary.setClassLoaderDelegation(XMLUtil.getAttributeValue(
                production, "class-loader-delegation"));
        sharedLibrary.setVersion(XMLUtil.getAttributeValue(production,
                VERSION_ATTR_NAME));

        List<Node> children = XMLUtil.getNodeChildren(production);
        for (Node child : children) {
            if (IDENTIFICATION_ELEM_NAME.equals(child.getNodeName())) {
                sharedLibrary.setIdentification(loadIdentificationXml(child));
            } else if ("shared-library-class-path".equals(child.getNodeName())) {
                sharedLibrary.setSharedLibraryClassPath(XMLUtil
                        .getTextContents(child.getChildNodes()));
            }
        }
        return sharedLibrary;
    }

    private void loadUnknownExtension(final Extensions extensions,
            final Node node) {
        String namespaceURI = node.lookupNamespaceURI(node.getPrefix());
        UnknownExtension unknownExtension = new UnknownExtension();
        URI unknownURI = null;
        try {
            unknownURI = new URI(namespaceURI);
        } catch (URISyntaxException e) {
            log.warning(INVALID_EXTENSION_NAMESPACE + node.getNodeName());
        }
        if (unknownURI != null) {
            unknownExtension.setExtensionURI(unknownURI);
            unknownExtension.setNode(node);
            extensions.addUnknownExtension(unknownURI, unknownExtension);
        }
    }

    /**
     * NOTE: The first location to search for JBI schema files is within the
     * Petals installation root directory. If a schema file is not found within
     * root directory, the alternative is use the packaged schema files flatten
     * somewhere which can be found in class path. It won't work if the schema
     * are compressed in a Zip or Jar archive.
     * 
     * @return
     */
    private InputStream openJbiSchemaFile() {
        InputStream theJbiSchemaResource = null;

        // find through class path
        // obtain class loader
        ClassLoader cl = Thread.currentThread().getContextClassLoader();
        // FIXME : is this piece of code necessary?
        // if (cl == null) {
        // cl = JBIDescriptorBuilder.class.getClassLoader();
        // }

        // load jbi.xsd schema resource file
        URL schemaURL = cl.getResource(JBI_XSD);
        if (schemaURL != null) {
            try {
                theJbiSchemaResource = schemaURL.openStream();
            } catch (IOException e) {
                log.log(Level.SEVERE, e.getMessage(), e);
            }
        }

        if (theJbiSchemaResource == null) {
            // find in petals installation root
            String petalsRoot = System.getProperty("petals.home");
            if (petalsRoot != null) {
                try {
                    theJbiSchemaResource = new FileInputStream(new File(
                            petalsRoot + JBI_SCHEMA_RESOURCE));
                } catch (FileNotFoundException e) {
                    log.log(Level.SEVERE, e.getMessage(), e);
                }
            }
        }

        return theJbiSchemaResource;
    }

}
