/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005-2006 EBM WebSourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.tools.jbicommon.descriptor;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;
import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * The services element is used to declare the static services provided and
 * consumed by a service unit. The term "static" in this context means those
 * services that are provided or consumed due to the deployment of the service
 * unit, and that are known at design- and/or deployment-time.
 * 
 * @version $Rev: 438 $ $Date: 2006-05-19 11:12:26Z $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 */
public class Services extends ExtensibleJbiElement {

    /**
     * Used to distinguish between service engine and binding component.
     */
    private boolean bindingComponent;

    /**
     * Services provided by a service unit.
     */
    private List<Provides> provides;

    /**
     * Services consumed by a service unit.
     */
    private List<Consumes> consumes;

    /**
     * Default constructor.
     */
    public Services() { 
        super();
        consumes = new ArrayList<Consumes>();
        provides = new ArrayList<Provides>();
    }

    @Override
    public boolean equals(final Object other) {
        if (!(other instanceof Services)) {
            return false; // NOPMD by gblondelle
        }
        Services castOther = (Services) other;
        return new EqualsBuilder().append(bindingComponent,
                castOther.bindingComponent)
                .append(provides, castOther.provides).append(consumes,
                        castOther.consumes).isEquals();
    }

    //
    // Getters and Setters
    //

    @Override
    public int hashCode() {
        return new HashCodeBuilder().append(bindingComponent).append(provides)
                .append(consumes).toHashCode();
    }

    /**
     * Attribute used to distinguish if the services declared is related to a
     * service engine or a binding component.
     * 
     * @return <code>boolean</code> indicates if it's a binding component.
     */
    public boolean isBindingComponent() {
        return bindingComponent;
    }

    @Override
    public String toString() {
        return new ToStringBuilder(this).append("bindingComponent",
                bindingComponent).append("provides", provides).append(
                "consumes", consumes).toString();
    }

    protected void addConsumes(final Consumes consumesElem) {
        consumes.add(consumesElem);
    }

    protected void addProvides(final Provides providesElem) {
        provides.add(providesElem);
    }

    /**
     * Set if it's relative to a binding component.
     * 
     * @param bindingComponent
     *            <code>true</code> to indicate it's a binding component.
     */
    protected void setBindingComponent(final boolean bindingComponent) {
        this.bindingComponent = bindingComponent;
    }

    /**
     * Set the list of services consumed by a service unit deployed.
     * 
     * @param consumes
     *            the list of consumed services.
     */
    protected void setConsumes(final List<Consumes> consumes) {
        this.consumes = consumes;
    }

    /**
     * Returns the list of services consumed by a service unit deployed.
     * @return consumes the list of consumed services.
     */
    public List<Consumes> getConsumes() {
        return consumes;
    }

    /**
     * Set the list of services provided with the deployment of a service unit.
     * 
     * @param provides
     *            the list of provided services.
     */
    void setProvides(final List<Provides> provides) {
        this.provides = provides;
    }

    /**
     * Returns the list of services provided with the deployment of a service unit.
     * @return provides the list of provided services.
     */
    public List<Provides> getProvides() {
        return provides;
    }

}
