/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: XMLUtil.java 98 2006-02-24 16:18:48Z alouis $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.tools.jbicommon.util;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptorException;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.sun.org.apache.xpath.internal.XPathAPI;

/**
 * Contains utilities methods for XML operations.
 * 
 * @version $Rev: 98 $ $Date: 2006-02-24 17:18:48 +0100 (ven., 24 f??vr. 2006) $
 * @since Petals 1.0
 * @author Adrien LOUIS
 * @author ddesjardins - eBMWebsourcing
 * @author gblondelle - eBMWebsourcing
 */
public final class XMLUtil {

	private static final Logger LOGGER = Logger.getLogger(XMLUtil.class
			.getName());

	private XMLUtil() {
		// Do nothing
	}

	/**
	 * Create an attribute node with the specified value
	 * 
	 * @param document
	 *            XML document
	 * @param att
	 *            attribute name
	 * @param value
	 *            attribute value
	 * @return an attribute, null if the document is null
	 */
	public static Node createAttribute(final Document document,
			final String att, final String value) {
		Node element = null;
		if (document != null) {
			element = document.createAttribute(att);
			element.setNodeValue(value);
		}
		return element;
	}

	/**
	 * Create a node with the specified name, attributes and values. attVal is a
	 * suite : att1,val1,att2,val2,...
	 * 
	 * @param document
	 *            xlm document
	 * @param nodeName
	 *            node name
	 * @param attVal
	 *            attribute values
	 * @return the node with its attribute, null if the array is not correct or
	 *         the document is null.
	 */
	public static Node createNode(final Document document,
			final String nodeName, final String... attVal) {
		Node element = null;
		if (document != null && attVal.length % 2 == 0) {
			element = document.createElement(nodeName);
			for (int i = 0; i < attVal.length; i = i + 2) {
				element.getAttributes().setNamedItem(
						createAttribute(document, attVal[i], attVal[i + 1]));
			}
		}
		return element;
	}

	/**
	 * Search for a child with the given nodeName. If recursive, search in all
	 * the child of first level, then if not found, search in the 2nd level of
	 * the first child, ...
	 * 
	 * @param node
	 *            parent node
	 * @param nodeName
	 *            node name
	 * @param recursive
	 *            boolean to know if we got through the xml tree
	 * @return a node
	 */
	public static Node findChild(final Node node, final String nodeName,
			final boolean recursive) {
		node.normalize();
		Node result = null;
		NodeList nl = node.getChildNodes();
		if (node != null && nodeName != null) {
			result = lookupNodeInNodeList(nodeName, nl);
			// now, search recursively if required
			if (result == null && recursive) {
				for (int i = 0; i < nl.getLength() && result == null; i++) {
					result = findChild(nl.item(i), nodeName, true);
				}
			}
		}
		return result;
	}

	private static Node lookupNodeInNodeList(final String nodeName,
			final NodeList nl) {
		Node result = null;
		for (int i = 0; i < nl.getLength() && result == null; i++) {
			if (nodeName.equals(nl.item(i).getNodeName())) {
				result = nl.item(i);
			}
		}
		return result;
	}

	/**
	 * Search for a child with the given nodeName. If recursive, search in all
	 * the child of firdt level, then if not found, search in the 2nd level of
	 * the first child, ...
	 * 
	 * @param node
	 *            parent node
	 * @param namespaceURI
	 *            The namespaceURI of the node
	 * @param nodeName
	 *            node name
	 * @param recursive
	 *            boolean to know if we got through the xml tree
	 * @return a node
	 */
	public static Node findChild(final Node node, final String namespaceURI,
			final String nodeName, final boolean recursive) {
		node.normalize();
		String prefix = node.lookupPrefix(namespaceURI);
		return findChild(node, prefix + nodeName, recursive);
	}

	/**
	 * Transform an XML {@link Node} into a String
	 * 
	 * @param node
	 *            the XML {@link Node} to parse
	 * @return the resulting String, null if node is null
	 * @throws TransformerException
	 *             if errors occured during transformation
	 */
	public static String parseToString(final Node node)
			throws TransformerException {
		String result = null;
		if (node != null) {
			node.normalize();
			Transformer transformer = TransformerFactory.newInstance()
					.newTransformer();
			StringWriter stringWriter = new StringWriter();
			transformer.transform(new DOMSource(node), new StreamResult(
					stringWriter));
			StringBuffer buffer = stringWriter.getBuffer();
			result = buffer.toString();
		}
		return result;
	}

	/**
	 * Return the value of the attribute in the node
	 * 
	 * @param n
	 *            the node
	 * @param attName
	 *            the name of the attribute
	 * @return the value of the attribute (can be empty), null if not found
	 */
	public static String getAttributeValue(final Node n, final String attName) {
		String ret = null;
		if (n != null) {
			NamedNodeMap atts = n.getAttributes();
			Node att = atts.getNamedItem(attName);
			if (att != null) {
				ret = att.getNodeValue();
			}
		}
		return ret;
	}

	/**
	 * Return the value of the attribute in the node. Throws an exception if
	 * missing attribute.
	 * 
	 * @param n
	 *            the node
	 * @param attName
	 *            the name of the attribute
	 * @return the value of the attribute (can be empty).
	 * @throws JBIDescriptorException
	 *             if missing the required attribute
	 */
	public static String getRequiredAttributeValue(final Node n,
			final String attName) throws JBIDescriptorException {
		String ret = null;
		if (n != null) {
			NamedNodeMap atts = n.getAttributes();
			Node att = atts.getNamedItem(attName);
			if (att != null) {
				String attValue = att.getNodeValue();
				if (!StringHelper.isNullOrEmpty(attName)) {
					ret = attValue;
				} else {
					throw new JBIDescriptorException(
							"Required attribute can't be empty: " + attName);
				}
			} else {
				throw new JBIDescriptorException(
						"Missing a required attribute: " + attName);
			}
		}
		return ret;
	}

	/**
	 * Return the first child of a node, regardless <i>text</i> node
	 * 
	 * @param node
	 * @return
	 */
	public static Node getFirstChild(final Node node) {
		Node result = node.getFirstChild();
		while (result != null && result.getNodeType() == Node.TEXT_NODE) {
			result = result.getNextSibling();
		}
		return result;
	}

	/**
	 * Return the next sibling of a node, regardless <i>text</i> node
	 * 
	 * @param node
	 * @return
	 */
	public static Node getNextSibling(final Node node) {
		Node result = node.getNextSibling();
		if (result != null) {
			while (result.getNodeType() == Node.TEXT_NODE) {
				result = result.getNextSibling();
			}
		}
		return result;
	}

	// /**
	// * return the node corresponding to the given path for the Document.
	// *
	// * @param path
	// * @param document
	// * @return the first matching node, null otherwise, or if document/path is
	// * null
	// */
	// public static Node getNode(final Document document, final String path) {
	// Node result = null;
	// if (document != null && path != null) {
	// try {
	// result = XPathAPI.selectSingleNode(document, path);
	// } catch (TransformerException e) {
	// LOGGER.log(Level.SEVERE, "Unexpected XPATH problem", e);
	// }
	// }
	// return result;
	// }

	/**
	 * Return the <i>text</i> element of a node
	 * 
	 * @param node
	 * @return
	 */
	public static String getTextContent(final Node node) {
		String result = null;
		NodeList list = node.getChildNodes();
		for (int i = 0; i < list.getLength(); i++) {
			if (list.item(i).getNodeType() == Node.TEXT_NODE) {
				result = list.item(i).getNodeValue();
				result = result.replace('\t', ' ').replace('\n', ' ').trim();
				break;
			}
		}
		return result;
	}

	/**
	 * Create a String result from a DOM document
	 * 
	 * @param document
	 *            the DOM Document
	 * @return a String representation of the DOM Document
	 * @throws TransformerException
	 */
	public static String createStringFromDOMDocument(final Node document)
			throws TransformerException {
		return createStringFromDOMNode(document, false);
	}

	/**
	 * Create a String result from a DOM Node
	 * 
	 * @param node
	 *            the DOM Node
	 * @return a String representation of the DOM Document
	 * @throws TransformerException
	 */
	public static String createStringFromDOMNode(final Node node)
			throws TransformerException {
		return createStringFromDOMNode(node, true);
	}

	/**
	 * Create a String result from a DOM Node
	 * 
	 * @param node
	 *            the DOM Node
	 * @return a String representation of the DOM Document
	 * @throws TransformerException
	 */
	public static String createStringFromDOMNode(final Node node,
			final boolean omitDeclaration) throws TransformerException {
		node.normalize();
		Source source = new DOMSource(node);
		StringWriter out = new StringWriter();
		Result resultStream = new StreamResult(out);
		TransformerFactory tFactory = TransformerFactory.newInstance();
		Transformer transformer;
		transformer = tFactory.newTransformer();
		if (omitDeclaration) {
			transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION,
					"yes");
		}
		transformer.transform(source, resultStream);
		return out.toString();
	}

	/**
	 * Create a document from a String
	 * 
	 * @param xml
	 *            an xml string
	 * @return a {@link Document} representing the document, null if an error
	 *         occured
	 */
	public static Document createDocumentFromString(final String xml) {
		Document doc = null;
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		factory.setNamespaceAware(true);
		DocumentBuilder builder;
		try {
			builder = factory.newDocumentBuilder();
			InputStream in = new ByteArrayInputStream(xml.getBytes());
			InputSource inputSource = new InputSource(in);
			doc = builder.parse(inputSource);
			doc.normalize();
		} catch (ParserConfigurationException e) {
			LOGGER.log(Level.SEVERE, "Bad XML parser configuration", e);
		} catch (SAXException e) {
			LOGGER.log(Level.SEVERE,
					"Bad XML fragment can't be transformed to a DOM tree.", e);
		} catch (IOException e) {
			LOGGER.log(Level.SEVERE, "Unexpected Error", e);
		}

		return doc;
	}

	/**
	 * Return an array of String representing each Text element of the nodes
	 * which are in the list.
	 * 
	 * @param list
	 * @return
	 */
	public static List<String> getTextContents(final NodeList list) {
		List<String> result = null;
		if (list != null) {
			result = new ArrayList<String>();
			for (int i = 0; i < list.getLength(); i++) {
				Node pathElement = list.item(i);
				if (pathElement.getNodeType() != Node.TEXT_NODE) {
					result.add(getTextContent(pathElement));
				}
			}
		}
		return result;
	}

	/**
	 * @param node
	 * @param attrName
	 * @return
	 */
	public static QName extractXmlAttributeQName(final Node node,
			final String attrName) {
		String attr = XMLUtil.getAttributeValue(node, attrName);

		QName qName = null;

		// qname like "xmlns:name"
		if ((attr.indexOf(':') > -1) && (attr.charAt(0) != '{')) {
			String ns = attr.substring(0, attr.indexOf(':'));

			String namespace = node.lookupNamespaceURI(ns);
			/*
			 * String namespace = source.getDocumentElement().getAttribute(
			 * "xmlns:" + ns);
			 */

			qName = new QName(namespace, attr.substring(attr.indexOf(':') + 1),
					ns);
		} else {
			// qname like "{ns}name" or "name"
			qName = QName.valueOf(attr);
		}
		return qName;
	}

	/**
	 * @param node
	 * @param attrName
	 * @return
	 * @throws JBIDescriptorException
	 */
	public static QName extractRequiredXmlAttributeQName(final Node node,
			final String attrName) throws JBIDescriptorException {
		String attr = XMLUtil.getRequiredAttributeValue(node, attrName);

		QName qName = null;

		// qname like "xmlns:name"
		if ((attr.indexOf(':') > -1) && (attr.charAt(0) != '{')) {
			String ns = attr.substring(0, attr.indexOf(':'));

			String namespace = node.lookupNamespaceURI(ns);
			/*
			 * String namespace = source.getDocumentElement().getAttribute(
			 * "xmlns:" + ns);
			 */

			qName = new QName(namespace, attr.substring(attr.indexOf(':') + 1),
					ns);
		} else {
			// qname like "{ns}name" or "name"
			qName = QName.valueOf(attr);
		}
		return qName;
	}

	/**
	 * A List of Node that contains all children of this node; nodes must be of
	 * type Node.ELEMENT_NODE to be returned. If there are no children, this is
	 * a NodeList containing no nodes.
	 * 
	 * @param node
	 *            the node from which children will be extracted
	 * @return a List of child Node of type Node.ELEMENT_NODE
	 */
	public static List<Node> getNodeChildren(final Node node) {
		List<Node> children = new ArrayList<Node>();
		if (node != null) {
			NodeList childrenList = node.getChildNodes();
			for (int i = 0; i < childrenList.getLength(); i++) {
				Node child = childrenList.item(i);
				if (child.getNodeType() == Node.ELEMENT_NODE) {
					children.add(child);
				}
			}
		}

		return children;
	}

	/**
	 * FIXME Change this method to use DOM instead of XPATH. return the node
	 * corresponding to the given path for the Document.
	 * 
	 * @param path
	 * @param document
	 * @return the first matching node, null otherwise, or if document/path is
	 *         null
	 */
	public static Node getNode(Document document, String path) {
		Node result = null;
		if (document != null && path != null) {
			try {
				result = XPathAPI.selectSingleNode(document, path);
			} catch (TransformerException e) {
				return null;
			}
		}
		return result;
	}

}
