/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: PackageAssemblyMojo.java 16:56:43 dutoo $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.tools.packageplugin;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.List;
import java.util.jar.Attributes;
import java.util.jar.JarEntry;
import java.util.jar.JarOutputStream;
import java.util.jar.Manifest;

import org.apache.commons.io.FileUtils;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.artifact.resolver.ArtifactResolver;
import org.apache.maven.model.Dependency;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;

import org.objectweb.petals.tools.packageplugin.utils.CopyFileSetUtil;

/**
 * Assemble the Petals distribution
 * 
 * @goal assembly
 * @description Package the Petals distribution
 * @aggregator
 * @requiresDependencyResolution runtime
 * 
 * @author ddesjardins - eBMWebsourcing
 * @author Marc Dutoo - Open Wide
 */
public class PackageAssemblyMojo extends AbstractMojo {

    /**
     * @component
     */
    protected ArtifactResolver artifactResolver;

    /**
     * Local maven repository.
     * 
     * @parameter expression="${localRepository}"
     * @required
     * @readonly
     */
    protected ArtifactRepository localRepository;

    /**
     * The Maven Project.
     * 
     * @parameter expression="${project}"
     * @required
     * @readonly
     */
    protected MavenProject project;

    /**
     * Base directory
     * 
     * @parameter expression="${basedir}"
     * @required
     * @readonly
     */
    protected File baseDir;

    /**
     * Directory of the petals assembly
     * 
     * @parameter
     * @required
     */
    protected File petalsAssemblyDirectory;

    
    /**
     * Execute the plugin
     */
    public void execute() throws MojoExecutionException {
        try {
            // Prepare the assembly directory
            prepare();

            // Copy the ressources
            copyRessources();

            // Copy the dependencies
            copyLib();

            // Jar the server
            makeServerJar();
 
            // Make extras dir
            fillExtrasDir();
        } catch (IOException e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        }
    }

    /**
     * Copy petals-component-common lib and all its dependencies into extras dir
     * 
     */
    protected void fillExtrasDir() throws IOException {
        // Component common project file
        File componentCommonProjectDir = new File(baseDir + File.separator
                + ".." + File.separator + ".." + File.separator + "components"
                + File.separator + "common");

        File extrasDir = new File(petalsAssemblyDirectory + File.separator
                + "extras");
        File componentCommonExtraDir = new File(extrasDir, "component-common");

        // Copy component common lib
        CopyFileSetUtil.copyFileSetToDirectory(new File(
                componentCommonProjectDir, "target"), componentCommonExtraDir,
                "*.jar", ".svn", true);

        // Copy component common dependencies
        CopyFileSetUtil.copyFileSetToDirectory(new File(
                componentCommonProjectDir, "target" + File.separator
                        + "dependencies"), componentCommonExtraDir, "*.jar",
                ".svn", true);

    }

    
    /**
     * Copy the libraries
     * 
     * @throws IOException
     */
    protected void copyLib() throws IOException {
        List dependencies = project.getDependencies();
        for (Object object : dependencies) {
            Dependency dependency = (Dependency) object;
            if (dependency.getScope() == null
                    || (dependency.getScope() != null && !dependency.getScope()
                            .equals("test"))) {
                String jarName = dependency.getArtifactId() + "-"
                        + dependency.getVersion() + ".jar";
                FileUtils.copyFileToDirectory(new File(localRepository
                        .getBasedir()
                        + File.separator
                        + dependency.getGroupId().replace(".", File.separator)
                        + File.separator
                        + dependency.getArtifactId()
                        + File.separator
                        + dependency.getVersion()
                        + File.separator + jarName), new File(
                        petalsAssemblyDirectory + File.separator + "lib"));
            }
        }
    }

    /**
     * Copy the ressources
     * 
     * @throws IOException
     * 
     */
    protected void copyRessources() throws IOException {
        // Copy the scripts to the bin dir
        CopyFileSetUtil.copyFileSetToDirectory(new File(baseDir
            + File.separator + "src" + File.separator + "scripts"), new File(
            petalsAssemblyDirectory + File.separator + "bin"), "**"
            + File.separator + "*.*", ".svn", true);
        // Copy configuration files into /conf
        CopyFileSetUtil.copyFileSetToDirectory(new File(baseDir
                + File.separator + "src" + File.separator + "conf"), new File(
                petalsAssemblyDirectory + File.separator + "conf"), "**"
                + File.separator + "*.*", ".svn", true);

        // Copy fractal architecture definitions to the conf/adl
        CopyFileSetUtil.copyFileSetToDirectory(new File(baseDir
                + File.separator + "src" + File.separator + "fractal"),
                new File(petalsAssemblyDirectory + File.separator + "conf"
                        + File.separator + "adl"), "**" + File.separator
                        + "*.*", ".svn", true);

        // Copy XML Schemas files into /schemas directory
        CopyFileSetUtil.copyFileSetToDirectory(new File(baseDir
                + File.separator + "src" + File.separator + "schema"),
                new File(petalsAssemblyDirectory + File.separator + "schema"),
                "**" + File.separator + "*.*", ".svn", true);

        // Copy samples components files into /samples directory
        CopyFileSetUtil.copyFileSetToDirectory(new File(baseDir
                + File.separator + ".." + File.separator + ".."
                + File.separator + "components"), new File(
                petalsAssemblyDirectory + File.separator + "components"), "**"
                + File.separator + "target" + File.separator + "*.zip", false);
        // Add all txt files to the binary
        CopyFileSetUtil.copyFileSetToDirectory(baseDir,
                petalsAssemblyDirectory, "*.txt", true);

        // Copy Demos files from every demo directory's subfolders
        File rootDemoDir = new File(baseDir + File.separator + ".."
                + File.separator + ".." + File.separator + "demos");
        for (File demoDir : rootDemoDir.listFiles()) {
            if (!demoDir.isDirectory() || demoDir.getName().startsWith(".")) {
                // filtering out .svn, pom.xml etc.
                continue;
            }
            String demoAssemblyDirName = demoDir.getName();
            CopyFileSetUtil.copyFileSetToDirectory(demoDir, new File(
                    petalsAssemblyDirectory + File.separator + "demos"
                            + File.separator + demoAssemblyDirName), "**"
                    + File.separator + "target" + File.separator + "*.zip",
                    false);
            CopyFileSetUtil.copyFileSetToDirectory(demoDir, new File(
                    petalsAssemblyDirectory + File.separator + "demos"
                            + File.separator + demoAssemblyDirName), "**"
                    + File.separator + "target" + File.separator + "*.war",
                    false);
        }

        // Copy ant files
        try {
            File antProjectDir = new File(baseDir + File.separator + ".."
                    + File.separator + ".." + File.separator + "tools"
                    + File.separator + "ant");
            File antDistribDir = new File(petalsAssemblyDirectory
                    + File.separator + "ant");

            CopyFileSetUtil.copyFileSetToDirectory(new File(antProjectDir,
                    "target"), new File(antDistribDir, "lib"), "*.jar", true);
            FileUtils.copyFileToDirectory(new File(antProjectDir, "src"
                    + File.separator + "distribution" + File.separator
                    + "README"), antDistribDir);
            FileUtils.copyFileToDirectory(
                    new File(antProjectDir, "src" + File.separator
                            + "distribution" + File.separator + "TODO"),
                    antDistribDir);
            FileUtils.copyFileToDirectory(new File(antProjectDir, "src"
                    + File.separator + "distribution" + File.separator
                    + "ant-sample.xml"), antDistribDir);
            FileUtils.copyFileToDirectory(new File(antProjectDir, "src"
                    + File.separator + "distribution" + File.separator
                    + "taskdef.xml"), antDistribDir);
            CopyFileSetUtil.copyFileSetToDirectory(new File(antProjectDir,
                    "target"), new File(antDistribDir, "lib"), "*.jar", true);

            // Copy mortgage demo auto-install ant file
            FileUtils.copyFileToDirectory(new File(baseDir + File.separator
                    + ".." + File.separator + ".." + File.separator + "demos"
                    + File.separator + "mortgage" + File.separator + "workflow"
                    + File.separator + "auto-install.xml"), new File(
                    petalsAssemblyDirectory + File.separator + "demos"
                            + File.separator + "mortgage"));

            // Copy mortgage demo auto-install ant file
            FileUtils.copyFileToDirectory(new File(baseDir + File.separator
                    + ".." + File.separator + ".." + File.separator + "demos"
                    + File.separator + "mortgage" + File.separator + "workflow"
                    + File.separator + "auto-install.xml"), new File(
                    petalsAssemblyDirectory + File.separator + "demos"
                            + File.separator + "mortgage"));

        } catch (Exception e) {
            // Do nothing
            // The platform is assembled without the ant tool
        }
    }

    /**
     * Make the server jar
     * 
     * @throws IOException
     */
    protected void makeServerJar() throws IOException {
        Manifest manifest = new Manifest();
        Attributes attributes = manifest.getMainAttributes();
        attributes.putValue(Attributes.Name.MANIFEST_VERSION.toString(), "1.0");
        attributes.putValue("Created-By", "Petals Assembly");
        attributes.putValue("Built-By", System.getProperty("user.name"));
        attributes.putValue("Package", "org.objectweb.petals");
        attributes.putValue("Build-Jdk", System.getProperty("java.version"));

        String classpath = "../conf/ ../conf/adl/ ../lib-ext/";
        for (File file : new File(petalsAssemblyDirectory + File.separator
                + "lib").listFiles()) {
            // classpath += " .." + File.separator + "common" + File.separator
            // + file.getName();
            classpath += " ../lib/" + file.getName();
        }
        attributes.putValue("Class-Path", classpath);

        attributes.putValue("Extension-Name", "petals");
        attributes.putValue("Specification-Title", "Petals Services Platform");
        attributes.putValue("Specification-Vendor", "ObjectWeb Consortium");
        attributes.putValue("Implementation-Title", "org.objectweb.petals");
        attributes.putValue("Implementation-Vendor", "ObjectWeb Consortium");

        String version = petalsAssemblyDirectory.getName();
        if (version.indexOf("-") > -1) {
            version = version.substring(version.indexOf("-") + 1);
        }
        attributes.putValue("Implementation-Version", version);
        attributes.putValue("Main-Class",
                "org.objectweb.petals.system.launch.Launcher");

        JarOutputStream jarOutputStream = new JarOutputStream(
                new FileOutputStream(petalsAssemblyDirectory + File.separator
                        + "bin" + File.separator + "server.jar"), manifest);

        // insert the license
        JarEntry license = new JarEntry("META-INF" + File.separator
                + "LICENSE.txt");
        jarOutputStream.putNextEntry(license);
        DataInputStream dis = new DataInputStream(new FileInputStream(baseDir
                + File.separator + "LICENSE.txt"));
        byte[] content = new byte[dis.available()];
        dis.readFully(content);
        jarOutputStream.write(content);
        jarOutputStream.closeEntry();

        jarOutputStream.flush();
        jarOutputStream.close();
    }

    /**
     * Prepare the assembly
     * 
     * @throws IOException
     * 
     */
    protected void prepare() throws IOException {
        FileUtils.deleteDirectory(petalsAssemblyDirectory);
        petalsAssemblyDirectory.mkdirs();

        // Log directory
        new File(petalsAssemblyDirectory + File.separator + "logs").mkdirs();
        // Bin directory
        new File(petalsAssemblyDirectory + File.separator + "bin").mkdirs();
        // Conf directory
        new File(petalsAssemblyDirectory + File.separator + "conf").mkdirs();
        // Fractal directory
        new File(petalsAssemblyDirectory + File.separator + "conf"
                + File.separator + "adl").mkdirs();
        // Schema directory
        new File(petalsAssemblyDirectory + File.separator + "schema").mkdirs();
        // Server lib directory
        new File(petalsAssemblyDirectory + File.separator + "lib").mkdirs();
       
        // Samples directory
        new File(petalsAssemblyDirectory + File.separator + "components")
                .mkdirs();

        // Demos directory
        new File(petalsAssemblyDirectory + File.separator + "demos").mkdirs();

       
        // Ant directory
        new File(petalsAssemblyDirectory + File.separator + "ant").mkdirs();
        new File(petalsAssemblyDirectory + File.separator + "ant"
                + File.separator + "lib").mkdirs();

        // Joram directory
        new File(petalsAssemblyDirectory + File.separator + "joram").mkdirs();

        // Extras directories
        new File(petalsAssemblyDirectory + File.separator + "extras"
                + File.separator + "component-common").mkdirs();

        // Installs directories
        new File(petalsAssemblyDirectory + File.separator + "install").mkdirs();
        new File(petalsAssemblyDirectory + File.separator + "installed")
                .mkdirs();
        new File(petalsAssemblyDirectory + File.separator + "uninstalled")
                .mkdirs();
        new File(petalsAssemblyDirectory + File.separator + "work").mkdirs();
        new File(petalsAssemblyDirectory + File.separator + "lost+found")
                .mkdirs();
        String repositoryDirectory = petalsAssemblyDirectory + File.separator
                + "repository";
        new File(repositoryDirectory).mkdirs();
        new File(repositoryDirectory + File.separator + "components").mkdirs();
        new File(repositoryDirectory + File.separator + "shared-libraries").mkdirs();
        new File(repositoryDirectory + File.separator + "service-assemblies")
                .mkdirs();

        
    }

}
