package org.objectweb.petals.tools.petalsadmin.utils;

import java.io.IOException;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import javax.management.MBeanServerConnection;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;

/**
 * Class used to get a connector to the JMX server
 * 
 * @author ddesjardins - eBMWebsourcing
 */
public final class JBIJMXConnectorUtil {
    
    private final static String PETALS_DOMAIN = "Petals";
    
    private final static String TYPE = "type";
    
    private final static String NAME = "name";
    
    private final static String SERVICE_TYPE = "service";
    
    private final static String ADMIN_NAME = "Admin";
    
    private final static String INSTALLATION_NAME = "Installation";
    
    private final static String DEPLOYMENT_NAME = "Deployment";

    private JBIJMXConnectorUtil() {
        // Do nothing
    }

    /**
     * Retreive the object name of the admin service
     * 
     * @param mBeanServer
     *            mbean server
     * @return object name
     * @throws IOException
     * @throws NullPointerException
     * @throws MalformedObjectNameException
     */
    public static ObjectName getAdminServiceMBeanName(
            MBeanServerConnection mBeanServer) throws IOException,
        MalformedObjectNameException, NullPointerException {
        ObjectName result = null;
        Set objNames;
        Hashtable<String, String> attributes = new Hashtable<String, String>();
        attributes.put(NAME, ADMIN_NAME);
        attributes.put(TYPE, SERVICE_TYPE);
        ObjectName objName = new ObjectName(PETALS_DOMAIN, attributes);
        objNames = mBeanServer.queryNames(objName, null);
        if (objNames != null && objNames.size() == 1) {
            result = (ObjectName) objNames.iterator().next();
        }
        return result;
    }

    /**
     * Retreive the object name of a component
     * 
     * @param mBeanServer
     *            mbean server
     * @return object name
     * @throws IOException
     * @throws NullPointerException
     * @throws MalformedObjectNameException
     */
    public static ObjectName getComponentMBeanName(
            MBeanServerConnection mBeanServer, String name) throws IOException,
        MalformedObjectNameException, NullPointerException {
        ObjectName result = null;
        String pattern = "*:type=engine,name=" + name;
        Set objNames;
        ObjectName objName = new ObjectName(pattern);
        objNames = mBeanServer.queryNames(objName, null);
        if (objNames != null && objNames.size() == 1) {
            result = (ObjectName) objNames.iterator().next();
        }
        if (result == null) {
            pattern = "*:type=binding,name=" + name;
            objName = new ObjectName(pattern);
            objNames = mBeanServer.queryNames(objName, null);
            if (objNames != null && objNames.size() == 1) {
                result = (ObjectName) objNames.iterator().next();
            }
        }
        return result;
    }

    /**
     * Get a JMX connector
     * 
     * @param host
     *            host
     * @param port
     *            port
     * @param username
     *            username
     * @param password
     *            password
     * @return jmx connector
     * @throws IOException
     */
    public static JMXConnector getConnection(String host, String port,
            String username, String password) throws IOException {
        JMXConnector connector = null;
        Properties env = System.getProperties();
        env.put("java.naming.factory.initial",
                "com.sun.jndi.rmi.registry.RegistryContextFactory");
        Map<String, Object> args = new HashMap<String, Object>();
        if (!"".equals(username)) {
            if (password == null) {
                password = "";
            }
            String[] credentials = new String[] {username, password};
            args.put("jmx.remote.credentials", credentials);
        }
        String urlStr = "service:jmx:rmi:///jndi/rmi://" + host + ":" + port
                + "/management/rmi-jmx-connector";
        JMXServiceURL url = new JMXServiceURL(urlStr);
        connector = JMXConnectorFactory.connect(url, args);
        return connector;
    }

    /**
     * Retreive the object name of the deployment service
     * 
     * @param mBeanServer
     *            mbean server
     * @return object name
     * @throws IOException
     * @throws NullPointerException
     * @throws MalformedObjectNameException
     */
    public static ObjectName getDeploymentServiceMBeanName(
            MBeanServerConnection mBeanServer) throws IOException,
        MalformedObjectNameException, NullPointerException {
        ObjectName result = null;
        Set objNames;
        Hashtable<String, String> attributes = new Hashtable<String, String>();
        attributes.put(NAME, DEPLOYMENT_NAME);
        attributes.put(TYPE, SERVICE_TYPE);
        ObjectName objName = new ObjectName(PETALS_DOMAIN, attributes);
        objNames = mBeanServer.queryNames(objName, null);
        if (objNames != null && objNames.size() == 1) {
            result = (ObjectName) objNames.iterator().next();
        }
        return result;
    }

    /**
     * Retreive the object name of the installation service
     * 
     * @param mBeanServer
     *            mbean server
     * @return object name
     * @throws IOException
     * @throws NullPointerException
     * @throws MalformedObjectNameException
     */
    public static ObjectName getInstallationServiceMBeanName(
            MBeanServerConnection mBeanServer) throws IOException,
        MalformedObjectNameException, NullPointerException {
        ObjectName result = null;
        Set objNames;
        Hashtable<String, String> attributes = new Hashtable<String, String>();
        attributes.put(NAME, INSTALLATION_NAME);
        attributes.put(TYPE, SERVICE_TYPE);
        ObjectName objName = new ObjectName(PETALS_DOMAIN, attributes);
        objNames = mBeanServer.queryNames(objName, null);
        if (objNames != null && objNames.size() == 1) {
            result = (ObjectName) objNames.iterator().next();
        }
        return result;
    }

}
